// vincia24.cc is a part of the VINCIA plugin to PYTHIA 8
// This is a simple test program to illustrate how several HepMC
// files can be used for uncertainty bands
// Copyright (C) 2019 Peter Skands.

// PYTHIA and VINCIA include files
#include "Vincia/Vincia.h"
#include "Pythia8Plugins/HepMC2.h"

// Use PYTHIA and VINCIA namespaces
using namespace Pythia8;
using namespace Vincia;

// Main Program.
int main( int argc, char* argv[] ) {

  // Check that correct number of command-line arguments
  if (argc != 3) {
    cout << endl
         << " Unexpected number of command-line arguments." << endl << endl
         << " Use the program like this: ./vincia24 vincia24.cmnd HepMCprefix"
         << endl << endl
         << " The program will create the HepMC file"
         << " HepMCprefix, HepMCprefix0 for the central weight, and" << endl
         << " HepMCprefix{1..n} for the uncertainy variations."
         << endl << endl
         << " For passing events directly to Rivet with fifo pipes do: " << endl
         << "   mkfifo HepMCprefix{0..n} " << endl
         << "   ./vincia24 vincia24.cmnd HepMCprefix &" << endl
         << "   rivet -H YODA_NAME.yoda -a ANALYSIS_NAME HepMCprefixX " << endl
         << " to analyse the Xth uncertainy variation." << endl
         << " Note that you have to execute the rivet command on all fifos."
         << endl << endl;
    return 1;
  }

  // Generator.
  Pythia pythia;
  VinciaPlugin vincia(&pythia);

  // Read user settings from file.
  pythia.readFile(argv[1]);

  // Initialize Vincia & Pythia.
  vincia.init();

  // Settings pointer
  Settings& settings = pythia.settings;

  // Check that uncertainty bands are switched on
  if (!settings.flag("Vincia:uncertaintyBands")) {
    cout << endl
         << " Uncertainty bands are not switched on. Exiting."
         << endl;
    return 1;
  }

  // Number of variations
  int nWeights = vincia.nWeights();

  // HepMC Interface.
  string hepmcFile = string(argv[2]);  
  HepMC::Pythia8ToHepMC ToHepMC;
  // Switch off warnings for parton-level events.
  ToHepMC.set_print_inconsistency(false);
  // Vector of events, one for each vector of weights
  vector< HepMC::IO_GenEvent* > HepMCevent;
  HepMCevent.resize(nWeights+1);
  for (int iVar=0; iVar<nWeights; iVar++) {
    string name = hepmcFile + static_cast<ostringstream*>( &(ostringstream() << 
      iVar) )->str();
    HepMCevent[iVar] = new HepMC::IO_GenEvent(name, std::ios::out);
  }
  HepMCevent[nWeights] = new HepMC::IO_GenEvent(hepmcFile, std::ios::out);

  // Number of events.
  int nEvent = settings.mode("Main:numberOfEvents");

  // Begin event loop.
  for (int iEvent = 1; iEvent <= nEvent; ++iEvent) {

    // Generate next event.
    if (!pythia.next()) continue;

    // Central weight, includes info.weight() and reweighting for MC
    // violations, if Vincia:allowReweighting is switched on).
    double weight = vincia.weight();

    // Create vector with weights.
    vector<double> weights;
    weights.push_back(weight);
    for (int iVar=1; iVar<nWeights; iVar++)
      weights.push_back(vincia.weight(iVar));

    // Fill HepMC files.
    HepMC::GenEvent* HepMCeventNow = new HepMC::GenEvent();
    ToHepMC.fill_next_event( pythia, HepMCeventNow );
    // Set the weights.
    for (int iVar=0; iVar<nWeights; iVar++) {
      HepMCeventNow->weights()[0] = weights[iVar];
      *HepMCevent[iVar] << HepMCeventNow;
    }
    HepMCeventNow->weights() = weights;
    *HepMCevent[nWeights] << HepMCeventNow;
    delete HepMCeventNow;

  } // End of event loop.

  // Print infos.
  vincia.printInfo();
  pythia.stat();

  // Done.
  return 0;
}
