// MatrixElementCorrections.h is a part of the VINCIA plug-in to the 
// PYTHIA 8 generator.
// Copyright (C) 2019 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

#ifndef Vincia_MatrixElementCorrections_H
#define Vincia_MatrixElementCorrections_H

#include "Pythia8/Basics.h"
#include "Pythia8/Event.h"
#include "Vincia/VinciaCommon.h"
#include "Vincia/TimeAntShower.h"
#include "Vincia/SpaceAntShower.h"
#include "Vincia/AntennaFunctions.h"
#include "Vincia/Resolution.h"
#include "Vincia/SubtractionTerms.h"
#include "Vincia/Tools.h"
#include "Vincia/MHV.h"
#include "Vincia/MG4interface.h"
#include "Vincia/MG5interface.h"

namespace Vincia {

  // Precision to check the scale for strong ordering
  const double Q2DIFFMAX = 0.0; // 1e-6;

  // Process classes for which we can do MECs
  const int procClassMin         = 1;
  const int procClassMax         = 4;
  const int singleBosonDecay     = 1;
  const int singleBosonProd      = 2;  
  const int pureQCD              = 3;
  const int EWtwoToN             = 4;

  // Max number of particles for which we can do MECs
  const int nMax[] = { 0, 7, 6, 7, 9 };
  // 0 for unknown process
  // 7 for single boson decays, up to X -> 6
  // 6 for single boson production, up to PP -> X+3
  // 7 for QCD, up to PP -> 5 (4 for non-MHV)
  // 9 for electroweak 2->n processes (8 for non-MHV)

  // Min number of particles for which we can do MECs
  const int nMin[] = { 0, 4, 4, 5, 5 };

  // Forward declarations on the classes in this file
  class MatrixElementCorrections;

  //***********************************************************************
  // Helper functions
  //***********************************************************************

  // Extract particles from event
  vector<Particle> convertEventToParticles(Event& event, int MEclass,
                                           map<int,int>& convMap);
  vector<Particle> convertEventToParticles(Event& event, int MEclass);

  // Functions to count specific partons
  int countNrQuarks(vector<Particle> parts, int emi1, int emi2);
  int countNrGluons(vector<Particle> parts);
  int countNrPartons(vector<Particle> parts);
  int countNrLeptons(vector<Particle> parts);

  // Extract colour chains from input
  vector< vector<int> > getColChains(vector<Particle> parts);
  
  // Function to check if a q qbar pair can be clustered
  bool canClustQQB(vector<Particle> parts, vector< vector<int> > cChains,
    int iJ, int iPQ, int nQLmin);

  //***********************************************************************
  // class clustCombi
  //***********************************************************************

  class clustCombi {
  public:
    clustCombi(int iRadIn, int iEmiIn, int iRecIn) {
      iRadSav = iRadIn;
      iEmiSav = iEmiIn;
      iRecSav = iRecIn;
    }
    clustCombi() {
      iRadSav = 0;
      iEmiSav = 0;
      iRecSav = 0;
    }
    ~clustCombi() {};
    int iRadSav, iEmiSav, iRecSav;
  };

  //***********************************************************************
  // class MEChist
  //***********************************************************************

  class MEChist {

  public:

    // Constructor.
    MEChist( int depth,
             MatrixElementCorrections* MECptr,
             TimeAntShower* timeShowerPtr,
             SpaceAntShower* spaceShowerPtr,
             Event& stateIn,
             map<string, double> stateVarsIn,
             double probIn,
             bool orderedIn,
             int nQLmin,
             clustCombi clusterIn,
             MEChist* motherIn );

    // The destructor deletes each child.
    ~MEChist() {
      for (int i=0; i<(int)childrenSav.size(); i++)
        delete childrenSav[i];
    }

    // The state.
    Event stateSav;
    map<string, double> stateVarsSav;
    double scaleSav;
    double probSav;
    double MEvalSav;
    double MECcontribSav;
    double asEffNsav;
    bool orderedSav;

    // The clustering that lead to this state.
    clustCombi clusterSav;

    // The previous step from which this step has been clustered. If
    // null, this is the initial step.
    MEChist* motherSav;

    // The different steps corresponding to possible clusterings of this
    // state. If empty, this is the final step to Born state.
    vector<MEChist*> childrenSav;

  };

  // Helper functions with histories as input.
  bool hasChildren( vector<MEChist*> histories );
  vector<MEChist*> getLeaves( MEChist* history );
  vector< vector<MEChist*> > getLayers( MEChist* history, int depth );

  //***********************************************************************
  // class MatrixElementCorrections
  //***********************************************************************

  class MatrixElementCorrections {

  public:

    // Constructor.
    MatrixElementCorrections() : isInitPtr(false), isInit(false) {};

    // Destructor.
    virtual ~MatrixElementCorrections() {};

    // Initialize Pointers
    void initPtr(Settings* settingsPtrIn, Info* infoPtrIn, Rndm* rndmPtrIn,
      MG4interface* mg4interfacePtrIn, MG5interface* mg5interfacePtrIn,
      MHVamplitudes* mhvPtrIn, VinciaCommon* vinComPtrIn) {
      settingsPtr     = settingsPtrIn;
      infoPtr         = infoPtrIn;
      rndmPtr         = rndmPtrIn;
      mg4interfacePtr = mg4interfacePtrIn;
      mg5interfacePtr = mg5interfacePtrIn;
      mhvPtr          = mhvPtrIn;
      vinComPtr       = vinComPtrIn;
      isInitPtr       = true;
    }

    // Initialize
    void init();

    // Wrapper function to return a specific antenna function
    AntennaFSR* getAntFSR(int iAnt) { return antSetFSR->getAnt(iAnt); }
    AntennaISR* getAntISR(int iAnt) { return antSetISR->getAnt(iAnt); }

    // Initialize Pointers to Antenna sets
    void initAntPtr(AntennaSetFSR* antFSRusr, AntennaSetISR* antISRusr) {
      antSetFSR = antFSRusr;
      antSetISR = antISRusr;
    }

    // Can we do the process
    bool canDoMEC(vector<Particle> parts, bool hardProcess);

    // Some helpers for process types
    bool decayME() { return (procClass == singleBosonDecay); }
    bool prodME()  { return (procClass == singleBosonProd ); }
    bool twoToN()  { return (procClass == singleBosonProd || 
      procClass == pureQCD || procClass == EWtwoToN); }
      
    void setUseMHV(bool useMHVIn) { useMHV = useMHVIn; }
    int  getHelAmpUsed() { return helAmpUsed; }

    // Added by A Lifson
    // What type of amplitude are we calculating
    // -2  : unpolarised
    // -1  : unphysical
    //  0  : MHV
    //  1  : NMHV (for future code)
    //  2  : NNMHV (for future code)
    //  3  : > NNMHV
    // Note: helicity conservation along a fermion line is handled by
    // MHVamplitudes class and by MG4interface separately of this check.
    int helAmpType(vector<Particle> parts);

    // Added by A Lifson
    // Polarize particles
    bool polarize(int nIn, vector<Particle>& parts);

    // Set and get process class
    void setProcClass(int procClassIn) { procClass = procClassIn; }
    int  getProcClass()                { return procClass; }

    // Set minimal number of quark lines and
    // Higgs-g for Higgs-q coupling off?
    // 0 = interference on
    // 1 = interference off and qqb->H as basic process
    // 2 = interference off and  gg->H as basic process
    void setClustProps(int nQLminIn, int HinterIn) { 
      nQLmin = nQLminIn;
           if (HinterIn == 1) mg4interfacePtr->setHinterferenceOff(true);
      else if (HinterIn == 2) mg4interfacePtr->setHinterferenceOff(false);
    }

    // Call both functions
    void prepMECs(int procClassIn, int nQLminIn, int HinterIn) {
      procClass = procClassIn;
      nQLmin    = nQLminIn;
           if (HinterIn == 1) mg4interfacePtr->setHinterferenceOff(true);
      else if (HinterIn == 2) mg4interfacePtr->setHinterferenceOff(false);
    }

    // Extract MEclass and clustering probability for input event.
    void prepMECs(const Event& eventIn, int incomeType);

    // Main Function: Construct all possible histories and
    // return MEn+1 / sum ant*MEn
    double performMECs(const Event& eventIn, TimeAntShower* timeShowerPtr, 
      SpaceAntShower* spaceShowerPtr, int nStep=-1);

    // Return the contribution from an unordered event included with merging.
    double unorderedContribution(MEChist* history, int nJet, TimeAntShower*
      timeShowerPtr, SpaceAntShower* spaceShowerPtr);

    // Return the contribution from an incomplete path included with merging.
    double incompleteContribution(MEChist* history);

    // Return scales and values in case of only unordered shower paths.
    map<string, double> getUnorderedInfos( const Event& eventIn,
      TimeAntShower* timeShowerPtr, SpaceAntShower* spaceShowerPtr);

    // Get the order of the n partons for evaluating the matrix elements
    vector< vector<int> > orderForME(vector<Particle> parts);

    // Evaluate a matrix element for the defined process classes
    double evalME(vector<Particle> parts);

    void setVerbose(int verboseIn) { verbose = verboseIn; }

    // Information about merging
    bool   doMerging, showerPathsOnly, nonShowerPathsOnly;
    int    nJetMax, incompleteScalePrescrip;
    double muFSave, unorderedScaleVariation;

    // Verbose level
    int verbose;

  private:

    // Is initialized?
    bool isInitPtr, isInit;

    // Pointer to PYTHIA 8 Settings database, info, and random-number generator
    Settings* settingsPtr;
    Info*     infoPtr;
    Rndm*     rndmPtr;
    // MadGraph interfaces
    MG4interface* mg4interfacePtr;
    MG5interface* mg5interfacePtr;
    // MHV interface
    MHVamplitudes* mhvPtr;
    // Pointer to VINCIA common function classes
    VinciaCommon* vinComPtr;

    // Antenna Sets
    AntennaSetFSR* antSetFSR;
    AntennaSetISR* antSetISR;

    // Matching settings
    bool meCorrections, matchingFullColor;
    int  matchingLO;

    // Process class
    int procClass;

    // Do we need at least N quark lines left?
    int nQLmin;

    // For helicity dependent corrections
    bool useMHV, isMHV, helicityShower;

    // Which polarisation was chosen?
    int helAmpUsed;

  };

}

#endif
