// TrialGeneratorsFSR.h is a part of the VINCIA plug-in to the PYTHIA 8 
// generator.
// Copyright (C) 2019 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the VinciaTrialGenerator and 
// derived classes.

#ifndef VINCIATrialGeneratorsFSR_H
#define VINCIATrialGeneratorsFSR_H

#include "Vincia/VinciaCommon.h"

namespace Vincia {

  // Instructions to create a new trial function:
  // See the classes below for examples and the VINCIA authors' compendium 
  // for existing definitions and derivations.
  // 1) Create a class inheriting from TrialGenerator
  // 2) Supply the basic trial function, aTrial()
  // 3) Choose a zeta definition for each predefined QE choice
  //       Encode the relevant zeta integrals, getIz()
  //       Encode the relevant zeta limits, getZmin() and getZmax()
  //       Encode the phase-space inversions, getYij() and getYjk()
  // 4) Encode the two getQ2 functions, for constant and running alphaS
  // 5) If needed, revise the init() function as well to suit your needs
  // 6) IMPORTANT! Add your definitions and derivations to the relevant 
  //    parts of the VINCIA authors' compendium, so that it is kept up to date.
  
  // Instructions to create a new evolution variable:
  // See the classes below for examples and the VINCIA authors' compendium
  // for existing definitions and derivations.
  // 1) For each existing TrialGenerator (i.e., each trial function that
  //    VINCIA already contains), select a zeta definition to use with 
  //    your new evolution variable.
  //       Encode the relevant zeta integrals, by adding to getIz()
  //       Encode the relevant zeta limits, by adding to getZmin() and getZmax()
  //       Encode the phase-space inversions, getYij() and getYjk()
  //    Note: the only function that it is really mandatory to add your 
  //    evolution variable to is the default soft-eikonal one encoded in the
  //    base class. The finite one may or may not be relevant, and the 
  //    additional collinear trial functions are only used for sector showers
  //    and for gluon splittings, which you may not be interested in modifying.
  //    The best is of course to ensure maximal compatibility by adding the
  //    relevant expressions to all trial functions.
  // 2) Add the solutions to your evolution integrals to the two getQ() 
  //    functions, for constant and running alphaS
  // 3) Revise the init() function to increase the number of recognized
  //    evolution variables
  // 4) IMPORTANT! Add your definitions and derivations to the relevant 
  //    parts of the VINCIA authors' compendium, so that it is kept up to date.

  // *****
  // Base class TrialGenerator
  // Note: base class is coded for a soft-eikonal trial function

  class TrialGeneratorFSR {
    
  public: 
    
    // -----------------------------------------------------------------
    // Constructor and pointer initialization (non-virtual)
    // -----------------------------------------------------------------

    // Constructor
    TrialGeneratorFSR() : evTypeSav(1), evNormFacSav(1.), 
      isInitPtr(false), isInit(false) {};
    
    // Init Pointers
    void initPtr(Rndm* rndmPtrIn, Settings* settingsPtrIn);

    // -----------------------------------------------------------------
    // Virtual functions 
    // Defining the particular kind of trial generator
    // -----------------------------------------------------------------

    // Name of trial generator
    virtual string name() {
      return "TrialGeneratorFSR";
    }

    // Initialize
    virtual void init(int evTypeIn, double evNormFacIn=1.0);

    // The trial antenna function, optionally in dimensionful form
    // Note: does not include color and headroom factors 
    virtual double aTrial(double yij, double yjk, double sIK=1.0);

    // Generate a new Q value, with first-order running alphaS
    virtual double genQ2run(double q2old, double sIK, double zMin, double zMax, 
                         double colFac, 
                         double b0, double kR, double Lambda,
                         double headroomFac=1.0, double biasFac=1.0);

    // Generate a new Q value, with constant trial alphaS
    virtual double genQ2(double q2old, double sIK, double zMin, double zMax, 
                         double colFac, double alphaSvalue, 
                         double headroomFac=1.0, double biasFac=1.0);

    // Generate a new zeta value in [zMin,zMax]
    virtual double genZ(double zMin, double zMax);
    
    // The zeta integral
    virtual double getIz(double zMin, double zMax);

    // The zeta boundaries, for a given value of the evolution variable
    // Note: for Type 2, zeta boundaries are for the branch yij < yjk
    // Note: for Type 3, zeta boundaries are nominally [0,1] (returned here)
    //       and should instead be bounded by the value used for the 
    //       hadronization cutoff and/or evolution window cutoffs
    virtual double getZmin(double yE);
    virtual double getZmax(double yE);
    
    // Inverse transforms
    // To obtain yij and yjk from yE and zeta
    virtual double getYij(double yE, double zeta);
    virtual double getYjk(double yE, double zeta);

    // -----------------------------------------------------------------
    // Non-Virtual functions
    // -----------------------------------------------------------------

    bool checkInit();

  protected:

    // Pointers
    Rndm* rndmPtr;
    Settings* settingsPtr;

    // Evolution Type and Normalization factor
    int evTypeSav;
    double evNormFacSav;

  private:

    // Status
    bool isInitPtr, isInit;

    // Verbose level
    int verbose;
  };


  // ===================================================================
  // TrialSoft
  // Soft-eikonal trial function (clone of base class)
  // ===================================================================

  class TrialSoft : public TrialGeneratorFSR {

  public: 

    // Name of trial generator
    virtual string name() {
      return "TrialSoft";
    }

  };

  // ===================================================================
  // TrialCollI
  // I-collinear trial function (for sector showers)
  // ===================================================================

  class TrialCollI : public TrialGeneratorFSR {

  public: 

    // Name of trial generator
    virtual string name() {
      return "TrialCollI";
    }

    // Trial function
    virtual double aTrial(double yij, double yjk, double sIK=1.0);

    // Generate a new Q value, with first-order running alphaS
    virtual double genQ2run(double q2old, double sIK, double zMin, double zMax, 
			    double colFac, double b0, double kR, double Lambda,
			    double headroomFac=1.0, double biasFac=1.0);

    // Generate a new Q value, with constant trial alphaS
    virtual double genQ2(double q2old, double sIK, double zMin, double zMax, 
                         double colFac, double alphaSvalue, 
                         double headroomFac=1.0, double biasFac=1.0);

    // Generate a new zeta value in [zMin,zMax]
    virtual double genZ(double zMin, double zMax);
    
    // The zeta integral
    virtual double getIz(double zMin, double zMax);

    // The zeta boundaries, for a given value of the evolution variable
    // Note: for Type 2, zeta boundaries are for the branch yij < yjk
    // Note: for Type 3, zeta boundaries are nominally [0,1] (returned here)
    //       and should instead be bounded by the value used for the 
    //       hadronization cutoff and/or evolution window cutoffs
    virtual double getZmin(double yE);
    virtual double getZmax(double yE);
    
    // Inverse transforms
    // To obtain yij and yjk from yE and zeta
    virtual double getYij(double yE, double zeta);
    virtual double getYjk(double yE, double zeta);

  };

  // ===================================================================
  // TrialCollK
  // K-collinear trial function (for sector showers)
  // Obtained by swapping invariants in TrialCollI
  // ===================================================================

  class TrialCollK : public TrialCollI {

  public: 

    // Name of trial generator
    virtual string name() {
      return "TrialCollK";
    }

    // Trial function
    virtual double aTrial(double yij, double yjk, double sIK=1.0) {
      return TrialCollI::aTrial(yjk, yij, sIK);
    }
    
    // Inverse transforms
    // To obtain yij and yjk from yE and zeta
    virtual double getYij(double yE, double zeta) {
      return TrialCollI::getYjk(yE, zeta);
    }
    virtual double getYjk(double yE, double zeta) {
      return TrialCollI::getYij(yE, zeta);
    }

  };

  // ===================================================================
  // TrialSplitK
  // K-collinear trial g->qq splitting function
  // ===================================================================

  class TrialSplitK : public TrialGeneratorFSR {
    
  public:
    
    // Name of trial generator
    virtual string name() {
      return "TrialSplitK";
    }
    
    // Trial function
    virtual double aTrial(double yij, double yjk, double sIK=1.0);

    // Generate a new Q value, with first-order running alphaS
    virtual double genQ2run(double q2old, double sIK, double zMin, double zMax, 
			    double colFac, 
			    double b0, double kR, double Lambda,
			    double headroomFac=1.0, double biasFac=1.0);

    // Generate a new Q value, with constant trial alphaS
    virtual double genQ2(double q2old, double sIK, double zMin, double zMax, 
                         double colFac, double alphaSvalue, 
                         double headroomFac=1.0, double biasFac=1.0);

    // Generate a new zeta value in [zMin,zMax]
    virtual double genZ(double zMin, double zMax);
    
    // The zeta integral
    virtual double getIz(double zMin, double zMax);

    // The zeta boundaries, for a given value of the evolution variable
    // Note: for Type 2, zeta boundaries are for the branch yij < yjk
    // Note: for Type 3, zeta boundaries are nominally [0,1] (returned here)
    //       and should instead be bounded by the value used for the 
    //       hadronization cutoff and/or evolution window cutoffs
    virtual double getZmin(double yE);
    virtual double getZmax(double yE);
    
    // Inverse transforms
    // To obtain yij and yjk from yE and zeta
    virtual double getYij(double yE, double zeta);
    virtual double getYjk(double yE, double zeta);

  };

  // ===================================================================
  // TrialSplitI
  // I-collinear trial g->qq splitting function
  // Obtained by swapping invariants in TrialSplitK
  // ===================================================================

  class TrialSplitI : public TrialSplitK {
    
  public: 
    
    // Name of trial generator
    virtual string name() {
      return "TrialSplitI";
    }

    // Trial function
    virtual double aTrial(double yij, double yjk, double sIK=1.0) {
      return TrialSplitK::aTrial(yjk, yij, sIK);
    }
    
    // Inverse transforms
    // To obtain yij and yjk from yE and zeta
    virtual double getYij(double yE, double zeta) {
      return TrialSplitK::getYjk(yE, zeta);
    }
    virtual double getYjk(double yE, double zeta) {
      return TrialSplitK::getYij(yE, zeta);
    }

  };

} // End namespace Vincia

#endif
