// File: extractColMatrix.cc
// This program should be used to extract an MG4 colour matrix and
// convert it to next_perm standard used by MHV amplitudes.
// Copyright (C) 2019 Peter Skands / Andrew Lifson
   
//**************************************************************************
 
// Stdlib header files for character manipulation.
#include <cctype>
 
// Stdlib header files for input/output.
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <map>
#include <vector>
#include <algorithm>
 
// Generic utilities.
using std::tolower;
using std::swap;
 
// Standard containers.
using std::string;
using std::map;
using std::vector;
using std::iterator;
 
// Input/output streams.
using std::cin;
using std::cout;
using std::cerr;
using std::ios;
using std::setw;
using std::replace;
using std::istream;
using std::ostream;
using std::ifstream;
using std::ofstream;
using std::istringstream;
using std::ostringstream;
 
// Input/output formatting.
using std::endl;
   
//**************************************************************************

// Generic facilities.

//*********

// Convert string to lowercase for case-insensitive comparisons.
// Also remove initial and trailing blanks, if any.

string toLower(const string& name) { 

  // Copy string without initial and trailing blanks.
  if (name.find_first_not_of(" ") == string::npos) return "";
  int firstChar = name.find_first_not_of(" ");
  int lastChar  = name.find_last_not_of(" ");
  string temp   = name.substr( firstChar, lastChar + 1 - firstChar);

  // Convert to lowercase letter by letter.
  for (int i = 0; i < int(temp.length()); ++i) 
    temp[i] = std::tolower(temp[i]); 
  return temp; 

}

//**************************************************************************
 
// Extract MG4 colour matrix from .f file
int main(int argc, char* argv[]) {
  
 // Check if HepMC file name is provided.
  if (argc <= 1) {
    cout << endl;
    cout << "Usage: ./extractColMatrix mg4matrix.f [multNum] [multDenom]" << endl;
    cout << "Please specify MG4 matrix.f file (e.g. ug2uggg.f)." << endl;
    cout << endl;
    return 0;
  }

  // Extract filename
  string fileName(argv[1]);
  // Remove 2 and .f to just get list of partons
  int loc = fileName.find("2");
  fileName.erase(loc,1);
  loc = fileName.find(".");
  fileName.erase(loc,string::npos);
  cout<<" Parton order is : "<<fileName<<endl;
  // Find indices of gluons
  vector<int> iG;
  map<int,bool> isGluon;
  // Map from MG gluons to 1,2,3,4,...
  map<int,int> normOrdG;
  loc = 0;
  while (loc != string::npos) {
    loc = fileName.find("g",loc);
    if (loc != string::npos) {
      iG.push_back(++loc);
      isGluon[loc]=true;
      normOrdG[loc]=iG.size();
    }
  }
  if (iG.size() > 0) {
    cout<<" Found gluons in positions: ";
    for (int i=0; i<(int)iG.size(); ++i) cout<<" "<<iG[i];
    cout<<" (with first parton having index 1)"<<endl;
  }

  // String for number of gluons
  string nGluonString = "N";
  if (iG.size() == 0) nGluonString = "Zero";
  if (iG.size() == 1) nGluonString = "One";
  if (iG.size() == 2) nGluonString = "Two";
  if (iG.size() == 3) nGluonString = "Three";
  if (iG.size() == 4) nGluonString = "Four";
  if (iG.size() == 5) nGluonString = "Five";
  if (iG.size() == 6) nGluonString = "Six";
  if (iG.size() == 7) nGluonString = "Seven";
  if (iG.size() == 8) nGluonString = "Eight";
  if (iG.size() == 9) nGluonString = "Nine";
  
  // Get user-supplied multipler (default unity)
  double multNum = 1.;
  double multDenom = 1.;
  if (argc >= 3) {
    istringstream stream(argv[2]);
    stream>>multNum;
    cout<<" MultNum = "<<multNum;
    if (argc >= 4) {
      istringstream stream2(argv[3]);
      stream2>>multDenom;
      cout<<" MultDenom = "<<multDenom;
    }      
    cout<<endl;
  }

  // Process file first time
  // Extract Denominators and find size of matrix
  ifstream is(argv[1]);
  string line;
  vector<int> denom;
  while ( getline(is, line) ) {
    // Extract MG denominator
    if (line.find("Denom") != string::npos) {
      int dBeg = line.find("/") + 1;
      int dEnd = line.find("/",dBeg);
      string dString = line.substr(dBeg,dEnd-dBeg);
      istringstream stream(dString);
      int d;
      stream>>d;
      denom.push_back(d);
    }
  }
  // Define size of matrix
  int size = denom.size();

  // Define MG matrix 
  int mgMatrix[size][size];

  // Process file second time to extract Matrix entries 
  is.clear();
  is.seekg(0);
  int rowNum = -1;
  int colNum = -1;
  while ( getline(is, line) ) {
    // Count up row number each time we pass a Denom statement
    if (line.find("DATA") != string::npos && line.find("Denom") != string::npos) {
      ++rowNum;
      colNum = -1;
    }
    // Extract MG denominator
    if (line.find("DATA") != string::npos && line.find("CF") != string::npos) {
      int dBeg = line.find("/") + 1;
      int dEnd = line.find("/",dBeg);
      string dString = line.substr(dBeg,dEnd-dBeg);
      // Remove commas
      std::replace( dString.begin(), dString.end(), ',', ' ');
      istringstream stream(dString);
      while(stream){        
        ++colNum;
        int c;
        stream>>c;        
        if (rowNum < size && colNum < size)
          mgMatrix[rowNum][colNum] = c;
      }      
      // The last read fails, so is not a correct entry.
      colNum--;
    }
  }

  // Process file third time to extract colour orderings
  is.clear();
  is.seekg(0);
  map<int,int> mapMGtoMHV;
  map<int,int> mapMHVtoMG;
  int iPermMG = -1;
  vector<string> mgOrderings;
  cout<<" MadGraph colour orderings : "<<endl;
  while ( getline(is, line) ) {
    // Find lines containing T[ ] orderings    
    if (line.find("T[") != string::npos && line.find("C ") != string::npos) {
      ++iPermMG;
      int dBeg = line.find("T") + 2;
      // Find last ]
      int dTmp = line.find("]",dBeg);
      int dEnd = dTmp;
      while (dTmp != string::npos) {
        dTmp = line.find("]",dEnd+1);
        if (dTmp != string::npos) dEnd = dTmp;
      }
      string dString(line.substr(dBeg,dEnd-dBeg));
      mgOrderings.push_back(dString);
      // Remove commas and any additional T symbols (for multiple quark chains)
      std::replace( dString.begin(), dString.end(), ',', ' ');
      std::replace( dString.begin(), dString.end(), 'T', ' ');
      // A multi-quark chain will now look like:
      //    4  6  5] [ 2  3
      // which can be broken down further (to do)
      //cout<<"  "<<dString<<endl;
      // Extract order of gluons only for this perm, in MG indexing
      vector<int> ordGandrew;
      istringstream stream(dString);
      int iPartonOld = 0;
      do { 
        int iParton;
        stream>>iParton;
        // Avoid repeating last entry twice
        if (iParton == iPartonOld) continue;
        iPartonOld = iParton;
        // Check ordering of gluons in Andrews notation (gluons numbered 1,2,3,...)
        if (isGluon[iParton]) {
          //cout<<" i "<<iParton<<" maps to gluon "<<normOrdG[iParton]<<endl;        
          ordGandrew.push_back(normOrdG[iParton]);
        }
      } while(stream);
      cout<<" MG     perm "<<iPermMG<<" is";
      for (int i=0; i<(int)ordGandrew.size(); ++i)
        cout<<" "<<ordGandrew[i];
      cout<<endl;
      
      // normOrdG now contains the ordering as Andrew would have labeled them.
      // Find out which permutation number Andrew would have given this ordering
      int nG = ordGandrew.size();
      int testList[nG];
      for (int i = 0; i < nG ; ++i) {
        testList[i] = i + 1;
      }
      int iPerm = 0;      
      do {
        // Check if this is the right permutation        
        bool isRight = true;
        for (int i = 0; i < nG; i++) {
          if (testList[i] != ordGandrew[i]) isRight=false;
        }
        if (isRight) {
          mapMGtoMHV[iPermMG]=iPerm;
          mapMHVtoMG[iPerm]=iPermMG;
          cout << " Andrew perm "<<iPerm<<" is ";
          for (int i = 0; i < nG; i++) {
            cout << testList[i] << " ";
          }
          cout << endl;
          break;
        }
        // If not yet found right permutation, do next one
        ++iPerm;
      }
      while (std::next_permutation(testList, testList + nG) );
      
    }
  }

  // Print permutation map
  cout<<" Permutation map is: "<<endl;
  for (int i=0; i<size; ++i) {
    cout<<" iMG = "<<i<<" : iMHV = "<<mapMGtoMHV[i]<<endl;     
  }
  
  // Construct MHV colour matrix
  int mhvMatrix[size][size];
  int mhvDenom[size];
  for (int iMG=0; iMG<size; ++iMG) {
    int iMHV = mapMGtoMHV[iMG];
    mhvDenom[iMHV]=denom[iMG];
    for (int jMG=0; jMG<size; ++jMG) {
      int jMHV = mapMGtoMHV[jMG];
      mhvMatrix[iMHV][jMHV] = mgMatrix[iMG][jMG];
    }
  }
  
  // Close file and print matrix
  is.close();
  cout<<" MG colour matrix : "<<endl;
  for (int i=0; i<size; ++i) {
    cout<<" i "<<setw(5)<<i<<"   ";
    for (int j=0; j<size; ++j) {
      // Put six entries on each line for nice formatting
      if (j%6 == 0 && j >= 1) cout<<endl<<"   "<<"     "<<"   ";
      cout<<" "<<setw(6)<<mgMatrix[i][j]<<"/"<<denom[i]<<".";
    }
    cout<<endl;
  }
  string outName = "Quark"+nGluonString+"GlueColour"; 
  string outFile = outName+".h";
  ofstream os(outFile);
  os<<"// Orderings correspond to gluon permutations generated with std::next_permutation()"<<endl;
  os<<"// (remapped from MadGraph colour matrix, including MHV prefactors)"<<endl;
  os<<"  const double "<<outName<<" ["<<size<<"]["<<size<<"] = {"<<endl;
  for (int i=0; i<size; ++i) {
    os<<"// Row "<<setw(5)<<i<<" : MG ordering  = ["<<mgOrderings[mapMHVtoMG[i]]<<" ]"<<endl;
    os<<"   {  ";
    for (int j=0; j<size; ++j) {
      // Put six entries on each line for nice formatting
      if (j%6 == 0 && j >= 1) os<<endl<<"      ";
      os<<" "<<setw(6)<<multNum*mhvMatrix[i][j]<<"/"<<multDenom*mhvDenom[i]<<".";
      if (j != size -1) os<<",";
      else os<<"}";
    }
    if (i != size-1) os<<",";
    else os<<" };";
    os<<endl;
  }
  os.close();

  // Print saved file
  ifstream test(outFile);
  while ( getline(test, line) ) {
    cout<<line<<endl;
  }
  test.close();
  
  cout<<" Saved remapped matrix to file : "<<outFile<<endl;

  // Exit
  return 0;
}
