// MG5interface.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2019 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Wrappers for MadGraph 5 FORTRAN matrix elements

#include "Vincia/MG5interface.h"

namespace Vincia {

  //==============================================================================

  // Wrappers for the MG5 FORTRAN common blocks
  
  extern "C" {

    //----------------------------------------------------------------------------

    // Common block for helicity selection

    extern struct {
      int             userhel;
    } heluserchoice_;
    //----------------------------------------------------------------------------
    // Common blocks converted from coupl.inc.
    // Expand/modify to agree with current coupl.inc structure
    // (Note: could be done by a script in future)

    extern struct {
      double          G;   
    } strong_;

    //extern struct {
    //  complex         gal[2];        
    //} weak_;

    extern struct {
      double          mu_t;
    } rscale_;
    
    extern struct {
      double          mdl_mt,mdl_mw,mdl_mz,mdl_mta,mdl_mb,mdl_mh;      
    } masses_;

    extern struct {
      double          mdl_wz,mdl_wh,mdl_ww,mdl_wt;
    } widths_;
    
    extern struct {
      std::complex<double>
      GC_1, GC_2, GC_3, GC_4, GC_5, GC_6, GC_7, GC_8, GC_9,
      GC_10, GC_11, GC_12,
      GC_15, GC_21, GC_27, GC_30, GC_31, GC_32, GC_33, GC_34, GC_35, GC_36,
      GC_37, GC_38, GC_39, GC_50, GC_51, GC_52, GC_53, GC_54, GC_55, GC_56,
      GC_57, GC_58, GC_59, GC_60, GC_61, GC_62, GC_63, GC_64, GC_65, GC_66,
      GC_67, GC_68, GC_69, GC_70, GC_71, GC_72, GC_73, GC_74, GC_75, GC_76,
      GC_77, GC_78, GC_79, GC_80, GC_81, GC_82, GC_83, GC_94, GC_95, GC_96,
      GC_97, GC_98, GC_99, GC_100;      
    } couplings_;
          
  } // End extern "C"

//==============================================================================

// The MG5 FORTRAN interface 
  //------------------------------------------------------------------------------

  // Main routine to initialise the MadGraph coupling common blocks, required
  // by the MG matrix elements. Adapted from MG5 : SOURCE/MODEL/couplings.f
  // Also set global information to be used in calls to MG5 matrix elements.

  bool MG5interface::init() {
    
    // Set verbose level
    verbose = settingsPtr->mode("Vincia:verbose");

    // Verbose output
    if (verbose >= 2) printOut("MG5interface::init","...");

    // Check if pointers initialized
    if (! isInitPtr) {
      printOut("MG5interface::init","Cannot initialize, pointers not set.");
      return false;      
    }
    
    // Innocent until proven guilty
    isInit = true;

    // Set color depth
    colorDepth = (settingsPtr->flag("Vincia:matchingFullColor")) ? 1 : 0;

    // Verbose output
    if (verbose >= 3) printOut("MG5interface::init","   setting MG masses...");

    // Initialise userhel
    heluserchoice_.userhel = -1;
    
    // Verbose output
    if (verbose >= 3) printOut("MG5interface::init","   setting MG C++ masses, widths, couplings...");

    // Initialise MG5 C++ parameters (same as for MG5 PYTHIA8 interface)
    pars.setIndependentParameters(particleDataPtr, couplingsPtr, slhaPtr); 
    pars.setIndependentCouplings(); 
    // In the VINCIA context, alphaS_MGME = 1/4pi (- > gS = 1; we
    // control the running separately). Thus, even the MG5 "dependent"
    // parameters only need to be set once.

    // Alternatively, we could evaluate the QCD coupling at MZ but should
    // then use a coupling definition from a Vincia parameter list rather
    // than PYTHIA's couplings.
    //    double muDefME  = 91.2;
    //    double alpS = couplingsPtr->alphaS(pow2(muDefME));

    double alpS = 1.0 / ( 4 * M_PI );
    pars.setDependentParameters(particleDataPtr, couplingsPtr, slhaPtr, alpS); 
    pars.setDependentCouplings();
    
    // Now "pars" contains all the MG5 SM parameters and couplings, in C++
    // format. Extract and hand to the equivalent FORTRAN common blocks.
    // Verbose output
    if (verbose >= 3) printOut("MG5interface::init","   setting MG FORTRAN masses, widths, couplings...");

    // Set masses and widths (initialized through Pythia)
    masses_.mdl_mb   = pars.mdl_MB;
    masses_.mdl_mt   = pars.mdl_MT;
    masses_.mdl_mta  = pars.mdl_MTA;
    masses_.mdl_mz   = pars.mdl_MZ;
    masses_.mdl_mw   = pars.mdl_MW;
    masses_.mdl_mh   = pars.mdl_MH;
    widths_.mdl_wt   = pars.mdl_WT;
    widths_.mdl_wz   = pars.mdl_WZ;
    widths_.mdl_ww   = pars.mdl_WW;
    widths_.mdl_wh   = pars.mdl_WH;

    // Strong coupling
    strong_.G        = pars.G;
    
    // Set couplings
    couplings_.GC_1  = pars.GC_1;
    couplings_.GC_2  = pars.GC_2;
    couplings_.GC_3  = pars.GC_3;
    couplings_.GC_4  = pars.GC_4;
    couplings_.GC_5  = pars.GC_5;
    couplings_.GC_6  = pars.GC_6;
    couplings_.GC_7  = pars.GC_7;
    couplings_.GC_8  = pars.GC_8;
    couplings_.GC_9  = pars.GC_9;

    couplings_.GC_10 = pars.GC_10;
    couplings_.GC_11 = pars.GC_11;
    couplings_.GC_12 = pars.GC_12;

    couplings_.GC_15 = pars.GC_15;

    couplings_.GC_21 = pars.GC_21;
    couplings_.GC_27 = pars.GC_27;
    
    couplings_.GC_30 = pars.GC_30;
    couplings_.GC_31 = pars.GC_31;
    couplings_.GC_32 = pars.GC_32;
    couplings_.GC_33 = pars.GC_33;
    couplings_.GC_34 = pars.GC_34;
    couplings_.GC_35 = pars.GC_35;
    couplings_.GC_36 = pars.GC_36;
    couplings_.GC_37 = pars.GC_37;
    couplings_.GC_38 = pars.GC_38;
    couplings_.GC_39 = pars.GC_39;
    
    couplings_.GC_50 = pars.GC_50;
    couplings_.GC_51 = pars.GC_51;
    couplings_.GC_52 = pars.GC_52;
    couplings_.GC_53 = pars.GC_53;
    couplings_.GC_54 = pars.GC_54;
    couplings_.GC_55 = pars.GC_55;
    couplings_.GC_56 = pars.GC_56;
    couplings_.GC_57 = pars.GC_57;
    couplings_.GC_58 = pars.GC_58;
    couplings_.GC_59 = pars.GC_59;
    couplings_.GC_60 = pars.GC_60;
    couplings_.GC_61 = pars.GC_61;
    couplings_.GC_62 = pars.GC_62;
    couplings_.GC_63 = pars.GC_63;
    couplings_.GC_64 = pars.GC_64;
    couplings_.GC_65 = pars.GC_65;
    couplings_.GC_66 = pars.GC_66;
    couplings_.GC_67 = pars.GC_67;
    couplings_.GC_68 = pars.GC_68;
    couplings_.GC_69 = pars.GC_69;
    couplings_.GC_70 = pars.GC_70;
    couplings_.GC_71 = pars.GC_71;
    couplings_.GC_72 = pars.GC_72;
    couplings_.GC_73 = pars.GC_73;
    couplings_.GC_74 = pars.GC_74;
    couplings_.GC_75 = pars.GC_75;
    couplings_.GC_76 = pars.GC_76;
    couplings_.GC_77 = pars.GC_77;
    couplings_.GC_78 = pars.GC_78;
    couplings_.GC_79 = pars.GC_79;
    couplings_.GC_80 = pars.GC_80;
    couplings_.GC_81 = pars.GC_81;
    couplings_.GC_82 = pars.GC_82;
    couplings_.GC_83 = pars.GC_83;
    
    couplings_.GC_94 = pars.GC_94;
    couplings_.GC_95 = pars.GC_95;
    couplings_.GC_96 = pars.GC_96;
    couplings_.GC_97 = pars.GC_97;
    couplings_.GC_98 = pars.GC_98;
    couplings_.GC_99 = pars.GC_99;
    couplings_.GC_100= pars.GC_100;

    // Verbose output
    if (verbose >= 3) printOut("MG5interface::init","   setting Color Depth...");

    // Set color depth
    colorDepth = (settingsPtr->flag("Vincia:matchingFullColor")) ? 1 : 0;

    // Create map containing the MG5 amplitudes
    initAmplitudes();
    
    return true;
  }

//--------------------------------------------------------------------------

  void MG5interface::initAmplitudes() {

    // Clear amplitude map and list of internal matrix elements
    ampMap.clear();
    for (int i=0; i<(int)MG5matrixList.size(); ++i) {
      delete MG5matrixList[i];
    }
    MG5matrixList.resize(0);
    
    // Make the list of internal MG5 matrix elements (hardcoded)
    MG5matrixList.push_back(new MG5dd_dd());
    MG5matrixList.push_back(new MG5dd_ddg());
    MG5matrixList.push_back(new MG5dd_ddgg());
    MG5matrixList.push_back(new MG5dd_ddggg());
    MG5matrixList.push_back(new MG5dd_dxddd());
    MG5matrixList.push_back(new MG5dd_dxdddg());
    MG5matrixList.push_back(new MG5dd_uxddu());
    MG5matrixList.push_back(new MG5dd_uxddug());
    MG5matrixList.push_back(new MG5dg_dg());
    MG5matrixList.push_back(new MG5dg_dgg());
    MG5matrixList.push_back(new MG5dg_dggg());
    MG5matrixList.push_back(new MG5dg_dgggg());
    MG5matrixList.push_back(new MG5dg_dxdd());
    MG5matrixList.push_back(new MG5dg_dxddg());
    MG5matrixList.push_back(new MG5dg_dxddgg());
    MG5matrixList.push_back(new MG5dg_dxdxddd());
    MG5matrixList.push_back(new MG5dg_uxdu());
    MG5matrixList.push_back(new MG5dg_uxdug());
    MG5matrixList.push_back(new MG5dg_uxdugg());
    MG5matrixList.push_back(new MG5dg_uxdxddu());
    MG5matrixList.push_back(new MG5du_du());
    MG5matrixList.push_back(new MG5du_dug());
    MG5matrixList.push_back(new MG5du_dugg());
    MG5matrixList.push_back(new MG5du_duggg());
    MG5matrixList.push_back(new MG5du_dxddu());
    MG5matrixList.push_back(new MG5du_dxddug());
    MG5matrixList.push_back(new MG5du_sxdus());
    MG5matrixList.push_back(new MG5du_sxdusg());
    MG5matrixList.push_back(new MG5du_uxduu());
    MG5matrixList.push_back(new MG5du_uxduug());
    MG5matrixList.push_back(new MG5dxd_dxd());
    MG5matrixList.push_back(new MG5dxd_dxdg());
    MG5matrixList.push_back(new MG5dxd_dxdgg());
    MG5matrixList.push_back(new MG5dxd_dxdggg());
    MG5matrixList.push_back(new MG5dxd_gg());
    MG5matrixList.push_back(new MG5dxd_ggg());
    MG5matrixList.push_back(new MG5dxd_gggg());
    MG5matrixList.push_back(new MG5dxd_ggggg());
    MG5matrixList.push_back(new MG5dxd_uxu());
    MG5matrixList.push_back(new MG5dxd_uxug());
    MG5matrixList.push_back(new MG5dxd_uxugg());
    MG5matrixList.push_back(new MG5dxd_uxuggg());
    MG5matrixList.push_back(new MG5dxdx_dxdx());
    MG5matrixList.push_back(new MG5dxdx_dxdxdxd());
    MG5matrixList.push_back(new MG5dxdx_dxdxdxdg());
    MG5matrixList.push_back(new MG5dxdx_dxdxg());
    MG5matrixList.push_back(new MG5dxdx_dxdxgg());
    MG5matrixList.push_back(new MG5dxdx_dxdxggg());
    MG5matrixList.push_back(new MG5dxdx_uxdxdxu());
    MG5matrixList.push_back(new MG5dxdx_uxdxdxug());
    MG5matrixList.push_back(new MG5dxg_dxdxd());
    MG5matrixList.push_back(new MG5dxg_dxdxdg());
    MG5matrixList.push_back(new MG5dxg_dxdxdgg());
    MG5matrixList.push_back(new MG5dxg_dxdxdxdd());
    MG5matrixList.push_back(new MG5dxg_dxg());
    MG5matrixList.push_back(new MG5dxg_dxgg());
    MG5matrixList.push_back(new MG5dxg_dxggg());
    MG5matrixList.push_back(new MG5dxg_dxgggg());
    MG5matrixList.push_back(new MG5dxg_sxuxdxus());
    MG5matrixList.push_back(new MG5dxg_uxdxdxdu());
    MG5matrixList.push_back(new MG5dxg_uxdxu());
    MG5matrixList.push_back(new MG5dxg_uxdxug());
    MG5matrixList.push_back(new MG5dxg_uxdxugg());
    MG5matrixList.push_back(new MG5dxu_dxu());
    MG5matrixList.push_back(new MG5dxu_dxug());
    MG5matrixList.push_back(new MG5dxu_dxugg());
    MG5matrixList.push_back(new MG5dxu_dxuggg());
    MG5matrixList.push_back(new MG5gg_dxd());
    MG5matrixList.push_back(new MG5gg_dxdg());
    MG5matrixList.push_back(new MG5gg_dxdgg());
    MG5matrixList.push_back(new MG5gg_dxdggg());
    MG5matrixList.push_back(new MG5gg_dxdxdd());
    MG5matrixList.push_back(new MG5gg_dxdxddg());
    MG5matrixList.push_back(new MG5gg_gg());
    MG5matrixList.push_back(new MG5gg_ggg());
    MG5matrixList.push_back(new MG5gg_gggg());
    MG5matrixList.push_back(new MG5gg_uxdxdu());
    MG5matrixList.push_back(new MG5gg_uxdxdug());
    MG5matrixList.push_back(new MG5sg_uxdxdus());
    MG5matrixList.push_back(new MG5ug_dxdu());
    MG5matrixList.push_back(new MG5ug_dxdug());
    MG5matrixList.push_back(new MG5ug_dxdugg());
    MG5matrixList.push_back(new MG5ug_dxdxddu());
    MG5matrixList.push_back(new MG5ug_sxdxdus());
    MG5matrixList.push_back(new MG5ug_uxdxduu());
    MG5matrixList.push_back(new MG5uxdx_sxuxdxs());
    MG5matrixList.push_back(new MG5uxdx_sxuxdxsg());
    MG5matrixList.push_back(new MG5uxdx_uxdx());
    MG5matrixList.push_back(new MG5uxdx_uxdxdxd());
    MG5matrixList.push_back(new MG5uxdx_uxdxdxdg());
    MG5matrixList.push_back(new MG5uxdx_uxdxg());
    MG5matrixList.push_back(new MG5uxdx_uxdxgg());
    MG5matrixList.push_back(new MG5uxdx_uxdxggg());
    MG5matrixList.push_back(new MG5uxdx_uxuxdxu());
    MG5matrixList.push_back(new MG5uxdx_uxuxdxug());
    MG5matrixList.push_back(new MG5uxg_sxuxdxds());
    MG5matrixList.push_back(new MG5uxg_uxdxd());
    MG5matrixList.push_back(new MG5uxg_uxdxdg());
    MG5matrixList.push_back(new MG5uxg_uxdxdgg());
    MG5matrixList.push_back(new MG5uxg_uxdxdxdd());
    MG5matrixList.push_back(new MG5uxg_uxuxdxdu());

    // Initialise internal MG5 matrix elements and add them to ampMap
    ampMap.clear();
    for (int i=0; i<(int)MG5matrixList.size(); ++i) {
      MG5matrixList[i]->init();
      // Get list of process labels to associate with this ME
      vector< vector<int> > processList = MG5matrixList[i]->processList();
      for (int j=0; j<(int)processList.size(); ++j)
	ampMap[processList[j]]  = MG5matrixList[i];
    }
    
    // Initialise and add any user-defined MG5 matrix elements
    for (int i=0; i<(int)MG5matrixListUser.size(); ++i) {
      MG5matrixListUser[i]->init();
      // Get list of process labels to associate with this ME
      vector< vector<int> > processList =
	MG5matrixListUser[i]->processList();
      for (int j=0; j<(int)processList.size(); ++j) {
	// Check if this entry already exists and issue warning if replacing
	if (ampMap.find(processList[j]) != ampMap.end()) {
	  cout<<"MG5interface::init(): Warning! Replacing existing MG5 matrix element for "<<processLabel2string(processList[j],true)<<endl;
	} else {
	  cout<<"MG5interface::init(): Adding user-defined MG5 matrix element for "<<processLabel2string(processList[j],true)<<endl;
	}
	ampMap[processList[j]] = MG5matrixListUser[i];
      }
    }
  }

//--------------------------------------------------------------------------

// Get Matrix Element Squared
// (Currently assumes particle state is in Vincia Colour order)
  
  double MG5interface::ME2(vector<Particle> pIn, int iA, int iB) {

    // Make PDG-ordered label for these particles
    vector<int> label = makeProcessLabel(pIn, iA, iB);
    
    // Do we have such an amplitude?
    if (ampMap.find(label) == ampMap.end()) {
      cout<<" Did not find amplitude for "
	  <<processLabel2string(label)<<endl;
      return -1.;
    }

    // Pointer to matrix element
    MG5matrix* matrix = ampMap[label];
    
    // Check if 1 or 2 incoming particles.
    bool isResDec = false;
    if (iB < 0) isResDec = true;

    // Save indices of various types of partons, and create vector of particles
    vector<int> iQ, iQbar, iG, iLep, iGam, iVHsm, iBSM, iHad, i9;
    for (int i=0; i<(int)pIn.size(); ++i) {
      if (pIn[i].id() >= 1 && pIn[i].id() <= 8) iQ.push_back(i);
      else if (pIn[i].id() <= -1 && pIn[i].id() >= -8) iQbar.push_back(i);
      else if (pIn[i].id() == 21) iG.push_back(i);
      else if (abs(pIn[i].id()) >= 12) iLep.push_back(i);
      else if (pIn[i].id() == 22) iGam.push_back(i);
      else if (pIn[i].id() >= 23 && pIn[i].id() <= 25) iVHsm.push_back(i);
      else if (abs(pIn[i].id()) >= 100 && abs(pIn[i].id()) <= 10000)
	iHad.push_back(i);
      else iBSM.push_back(i);
      // Save indices of unpolarised partons
      if (pIn[i].pol() == 9) i9.push_back(i);
    }

    // Explicitly sum over any hel = 9 helicities
    vector< vector<int> > helConf;
    vector<int> helOrg;
    for (int i=0; i<(int)pIn.size(); ++i)
      helOrg.push_back(pIn[i].pol());
    helConf.push_back(helOrg);
    while (i9.size() > 0) {
      int i = i9.back();
      // How many spin states?
      int nS = particleDataPtr->spinType(pIn[i].id());
      // Massless particles max have 2 (physical) spin states
      if (particleDataPtr->m0(pIn[i].id()) == 0.0) nS=min(nS,2);
      // Create nS copies of helConf, one for each spin state 
      int helConfSizeNow = helConf.size();
      for (int iCopy=1; iCopy <= nS; ++iCopy) {
        // Set hel for this particle in this copy.
        // Start from -1, then 1, then 0 (if 3 states)
        int h = -1;
        if (iCopy == 2) h = 1;
        else if (iCopy == 3) h = 0;
        else if (iCopy == 4) h = -2;
        else if (iCopy == 5) h = 2;
        for (int iHelConf=0; iHelConf<helConfSizeNow; ++iHelConf) {
          vector<int> helNow = helConf[iHelConf];
          helNow[i] = h;
          // First copy: use existing
          if (iCopy == 1) helConf[iHelConf] = helNow;
          // Subsequent copies: create new
          else helConf.push_back(helNow);
        }
      }
      // Remove the particle whose helicities have been summed over
      i9.pop_back();
    }

    // VERBOSE DEBUG
    //cout<<" number of helicity configurations = "<<helConf.size()<<endl;
    //for (int i=0; i<(int)helConf.size(); ++i) {
    //  cout<<"  helConf "<<i;
    //  for (int j=0; j<(int)helConf[i].size(); ++j) {
    //    cout<<" "<<helConf[i][j];
    //  }
    //cout<<endl;
    //}
    
    // Set parton configuration (except helicities and colours, see below)
    matrix->reset();
    matrix->setPerm(pIn, iA, iB);
    matrix->setMom(pIn);

    // For LC and Vincia colour, also set ICOL
    if (colorDepth == 0 || colorDepth == 1) {
      // Create colour-list in Vincia numbering (assuming ordered Vincia list)
      vector<int> colChain;
      vector< vector<int> > ccList;      
      for (int i=0; i<(int)pIn.size(); ++i) {
	int sign = (i == iA || i == iB) ? -1 : 1;
	// Octets: just add them to chain (note: max handles one colour ring)
	if (particleDataPtr->colType(pIn[i].id()) == 2)
	  colChain.push_back(i);
	// Beginning of new colour chain
	else if (particleDataPtr->colType(pIn[i].id()) == sign) {
	  if (colChain.size() != 0) ccList.push_back(colChain);
	  colChain.resize(0);
	  colChain.push_back(i);
	}
	// End of colour chain
	else if (particleDataPtr->colType(pIn[i].id()) == -sign) {
	  colChain.push_back(i);
	  ccList.push_back(colChain);
	  colChain.resize(0);
	}
      }
      // Add this contraction to list of contractions
      if (colChain.size() != 0) ccList.push_back(colChain);

      // Get MG5 colour index for this set of contractions
      int icol = matrix->setIcol(ccList);  

      // Sanity check 
      if (icol <= 0) {
	cout<<"MG5interface::ME2(): ERROR! invalid colour structure for "
	    <<matrix->name()<<endl;
	return 0.;
      }
    }
    
    // Compute helicity sum (and save helicity components if needed later)
    double me2 = 0.0;
    me2hel.clear();
    for (int iHC=0; iHC<(int)helConf.size(); ++iHC) {

      // Note: could check here if the helConf is physical
      // (consistent with spins of particles and conserving angular momentum)

      // Verbose output: print helicity configuration in Vincia's enumeration
      //cout<<" helConfNow = "<<iHC<<" size = "<<helConf[iHC].size();
      //for (int h=0; h<(int)helConf[iHC].size(); ++h)
      //  cout<<" "<<helConf[iHC][h];
      //cout<<endl;

      // Get MG5 helicity index for this helConf
      int ihel = matrix->setHel(helConf[iHC]);
      if (ihel <= 0) { 
	cout<<"MG5interface::ME2(): ERROR! invalid helicity structure for "
	    <<matrix->name()<<endl;
      } else {
	double me2now = matrix->getME2(colorDepth);
	// MG may produce inf/nan for unphysical hel combinations
	if ( !std::isnan(me2now) && !std::isinf(me2now) ) {
	  // Save helicity matrix element for possible later use.
	  me2hel[helConf[iHC]] = me2now;
	  // Add this helicity ME to me2
	  me2 += me2now;
	}
      }
    }
    // Return amplitude squared
    return me2;
 }

//--------------------------------------------------------------------------

// Set helicities for a particle state 
// (Note: takes a reference as input and operates on it)

 bool MG5interface::selectHelicities(vector<Particle>& p, int iA, int iB) {

   // Get the matrix element (automatically sums over any h=9 particles)
   double me2sum = ME2(p, iA, iB);

   // Did we find the ME? ( ME2() returns -1 if no ME found ).
   if (me2sum < 0.) return false;

   // Check how many helicity configurations we summed over
   int nHelConf = me2hel.size();
   if (nHelConf <= 0) return false;

   // Random number between zero and me2sum (trivial if only one helConf)
   double ranHelConf = 0.0;
   vector<int> hSelected;
   if (nHelConf >= 2) ranHelConf = rndmPtr->flat() * me2sum;
   for (map< vector<int>, double>::iterator it=me2hel.begin();
	it != me2hel.end(); ++it) {
     // Progressively subtract each me2hel and check when we cross zero
     ranHelConf -= it->second;     
     if (ranHelConf <= 0.0) {
       hSelected = it->first;
       break;
     }
   }
   if (ranHelConf > 0.) return false;
   
   // Set helicity configuration
   for (int i=0; i<(int)p.size(); ++i) p[i].pol(hSelected[i]);
       
   // Everything OK
   return true;
 }
  
//--------------------------------------------------------------------------

// Get Matrix Element Squared
// (Currently assumes particle state is in Vincia Colour order)

 string MG5interface::processLabel2string(vector<int> id, bool useNames) {
   
   string label = "{";
   for (int i=0; i<(int)id.size(); ++i) {
     string idName;
     if (useNames) idName = particleDataPtr->name(id[i]);
     else idName = num2str(id[i]);
     label += idName+" ";
   }
   label += "}";
   return label;
 }
  
//==========================================================================
  
} // end namespace Vincia
