// MergingHooks.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator. 
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2019 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

#include "Vincia/MergingHooks.h"

namespace Vincia {

  //==========================================================================

  // The MergingHooks class.

  //--------------------------------------------------------------------------

  // Initialize.

  void VinciaMergingHooks::init() {

    // Initialize parameters that we are not using.
    doUserMergingSave     = false;
    doMGMergingSave       = false;
    doKTMergingSave       = false;
    doCutBasedMergingSave = false;
    doNL3TreeSave         = false;
    doNL3LoopSave         = false;
    doNL3SubtSave         = false;
    doUNLOPSTreeSave      = false;
    doUNLOPSLoopSave      = false;
    doUNLOPSSubtSave      = false;
    doUNLOPSSubtNLOSave   = false;
    doUMEPSTreeSave       = false;
    doUMEPSSubtSave       = false;
    nReclusterSave        = 0;
    nQuarksMergeSave      = 5;
    includeMassiveSave    = true;
    enforceStrongOrderingSave = false;
    scaleSeparationFactorSave = 1.0;
    orderInRapiditySave   = false;
    nonJoinedNormSave     = 1.0;
    fsrInRecNormSave      = 1.0;
    pickByFullPSave       = true;
    pickByPoPT2Save       = false;
    pickBySumPTSave       = false;
    includeRedundantSave  = false;
    allowColourShufflingSave = false;
    herwigAcollFSRSave    = 1.0;
    herwigAcollISRSave    = 0.9;
    pT0ISRSave            = 2.0;
    pTcutSave             = 0.2;
    pTcutSave             = max(pTcutSave,pT0ISRSave);
    doWeakClusteringSave  = false;
    doSQCDClusteringSave  = false;
    DparameterSave        = 1.0;
    nJetMaxNLOSave        = -1;
    useShowerPluginSave   = true;
    // Parameters to allow setting hard process scales to default (dynamical)
    // Pythia values. TODO
    resetHardQRenSave     = true;
    resetHardQFacSave     = true;

    doPTLundMergingSave   = settingsPtr->flag("Merging:doPTLundMerging");
    processSave           = settingsPtr->word("Merging:Process");

    tmsValueSave          = settingsPtr->parm("Merging:TMS");
    tmsValueNow           = tmsValueSave;
    nJetMaxSave           = settingsPtr->mode("Merging:nJetMax");
    tmsListSave.resize(0);

    includeWGTinXSECSave  = settingsPtr->flag("Merging:includeWeightInXsection");
    doEstimateXSection    = settingsPtr->flag("Merging:doXSectionEstimate");
    applyVeto             = settingsPtr->flag("Merging:applyVeto");

    doRemoveDecayProducts = settingsPtr->flag("Merging:mayRemoveDecayProducts");

    // Clear hard process.
    if (!hardProcess) {
      hardProcess = new HardProcess();
      useOwnHardProcess = true;
    }
    hardProcess->clear();

    // Initialise input event.
    inputEvent.init("(hard process)", particleDataPtr);

    // Initialise the hard process
    if (doMGMergingSave)
      hardProcess->initOnLHEF(lheInputFile, particleDataPtr);
    else
      hardProcess->initOnProcess(processSave, particleDataPtr);

    // Remove whitespace from process string
    while(processSave.find(" ", 0) != string::npos)
      processSave.erase(processSave.begin()+processSave.find(" ",0));

    // Parameters for scale choices.
    unorderedScalePrescipSave   =
      settingsPtr->flag("Merging:useEffectiveScaleInUnorderedEvents") ? 2 :
      settingsPtr->mode("Merging:unorderedScalePrescrip");
    unorderedASscalePrescipSave =
      settingsPtr->mode("Merging:unorderedASscalePrescrip");
    unorderedPDFscalePrescipSave =
      settingsPtr->mode("Merging:unorderedPDFscalePrescrip");
    incompleteScalePrescipSave  =
      settingsPtr->mode("Merging:incompleteScalePrescrip");
    unorderedScaleVariationSave =
      settingsPtr->parm("Merging:unorderedScaleVariation");

    // If effective scale is used, check alphaS settings.
    if (unorderedScalePrescipSave == 2) {
      string checkSetting[11] = { "Vincia:alphaSkMuF", "Vincia:alphaSkMuSplitF",
        "Vincia:alphaSkMuI", "Vincia:alphaSkMuSplitI", "Vincia:alphaSkMuConv",
        "Vincia:pTnormalization", "Vincia:pTnormalizationII",
        "Vincia:pTnormalizationIF", "Vincia:mSnormalizationI",
        "Vincia:mCnormalizationI", "Vincia:mSnormalizationF" };
      for (int i=0; i<11; i++)
        if ( abs(settingsPtr->parm(checkSetting[i])-1.0) > TINY)
          infoPtr->errorMsg("Error in VinciaMergingHooks::init: Effective "
            "scale requested, but " + checkSetting[i] + " not 1.0!");
    }

    doMECsSave             = settingsPtr->flag("Merging:useMECs");
    showerPathsOnlySave    = settingsPtr->flag("Merging:showerPathsOnly");
    nonShowerPathsOnlySave = settingsPtr->flag("Merging:nonShowerPathsOnly");
    checkShowerCutoffSave  = settingsPtr->flag("Merging:checkShowerCutoff");
    showerCutoffSave       = settingsPtr->parm("Merging:showerCutoff");
    //pT0ISRSave             = showerCutoffSave;
    //pTcutSave              = showerCutoffSave;

    // Initialize parameters for communication with the shower.
    useOrderedPathSave    = false;
    useCompletePathSave   = false;
    doOrderedMECsSave     = false;
    nrGoodBranchSave      = 0;

    muFSave               = settingsPtr->parm("Merging:muFac");
    muRSave               = settingsPtr->parm("Merging:muRen");
    // Factorization, renormalization scale used in matrix element calculation.
    muFinMESave           = settingsPtr->parm("Merging:muFacInME");
    muRinMESave           = settingsPtr->parm("Merging:muRenInME");

    // Initialize CKKWL weight.
    weightCKKWLSave       = 1.0;
    weightFIRSTSave       = 0.0;
    nMinMPISave           = 100;
    muMISave              = -1.0;

    hasJetMaxLocal        = false;
    nJetMaxLocal          = nJetMaxSave;
    nJetMaxNLOLocal       = nJetMaxNLOSave;

    bool writeBanner      =  doPTLundMergingSave;

    if (!writeBanner) return;

    // Write banner.
    cout << "\n *------------------ MEPS Merging Initialization  ---------------"
         << "---*";
    cout << "\n |                                                               "
         << "   |\n";
    cout << " | CKKW-L merge                                                  "
         << "   |\n"
         << " |"<< setw(34) << processSave << "  with up to"
         << setw(3) << nJetMaxSave << " additional jets |\n";
    if (tmsValueSave > 10000.0)
      cout << " | Merging scale is defined by Lund pT, with tMS set to infinity    |";
    else 
      cout << " | Merging scale is defined by Lund pT, with value tMS = "
           << setw(6) << fixed << setprecision(1) << tmsValueSave << " GeV |";
    cout << "\n |                                                               "
         << "   |";
    cout << "\n *-------------- END MEPS Merging Initialization  ---------------"
         << "---*\n\n";

  }

  //--------------------------------------------------------------------------

  // Set starting scales.

  bool VinciaMergingHooks::setShowerStartingScales(bool isTrial, 
    bool doMergeFirstEmm, double& pTscaleIn, const Event& event,
    double& pTmaxFSRIn, bool& limitPTmaxFSRIn,
    double& pTmaxISRIn, bool& limitPTmaxISRIn,
    double& pTmaxMPIIn, bool& limitPTmaxMPIIn ) {

    // Local copies of power/wimpy shower booleans and scales.
    bool   limitPTmaxFSR = limitPTmaxFSRIn;
    bool   limitPTmaxISR = limitPTmaxISRIn;
    bool   limitPTmaxMPI = limitPTmaxMPIIn;
    double pTmaxFSR      = pTmaxFSRIn;
    double pTmaxISR      = pTmaxISRIn;
    double pTmaxMPI      = pTmaxMPIIn;
    double pTscale       = pTscaleIn;

    // Check if the process only contains two outgoing partons. If so, then
    // this process could also have been produced by MPI. Thus, the MPI starting
    // scale would need to be set accordingly to correctly attach a
    // "no-MPI-probability" to multi-jet events. ("Hard" MPI are included
    // by not restricting MPI when showering the lowest-multiplicity sample.)
    double pT2to2 = 0;
    int nFinalPartons = 0, nInitialPartons = 0, nFinalOther = 0;
    for ( int i = 0; i < event.size(); ++i ) {
      if ( (event[i].mother1() == 1 || event[i].mother1() == 2 )
        && (event[i].idAbs()   < 6  || event[i].id()      == 21) )
        nInitialPartons++;
      if (event[i].isFinal() && (event[i].idAbs() < 6 || event[i].id() == 21)) {
          nFinalPartons++;
          pT2to2 = event[i].pT();
      } else if ( event[i].isFinal() ) nFinalOther++;
    }
    bool is2to2QCD = (nFinalPartons == 2 && nInitialPartons == 2
      && nFinalOther == 0);
    bool hasMPIoverlap = is2to2QCD;

    double scale   = event.scale();

    double refScale = muF();
    if (limitPTmaxISR || limitPTmaxFSR)
      refScale = sqrt(spaceAntShowerPtr->getBornScale(event));

    // Starting scales for trial shower.
    if (isTrial) {

      // Reset shower and MPI scales.
      pTmaxISR = pTmaxFSR = pTmaxMPI = scale;

      // Reset to minimal scale for wimpy showers.
      if (limitPTmaxISR) pTmaxISR = min(scale,refScale);
      if (limitPTmaxFSR) pTmaxFSR = min(scale,refScale);
      if (limitPTmaxMPI) pTmaxMPI = min(scale,refScale);

      // For pure QCD set the PS starting scales to the pT of the dijet system.
      if (is2to2QCD) {
        pTmaxFSR = pT2to2;
        pTmaxISR = pT2to2;
      }

      // If necessary, set the MPI starting scale to the collider energy.
      if ( hasMPIoverlap ) pTmaxMPI = infoPtr->eCM();

      // Reset phase space limitation flags
      if ( pTscale < infoPtr->eCM() ) {
        limitPTmaxISR = limitPTmaxFSR = limitPTmaxMPI = true;
        // If necessary, set the MPI starting scale to the collider energy.
        if ( hasMPIoverlap ) limitPTmaxMPI = false;
      }

    }

    // Starting scales for regular showers.
    if ( doMergeFirstEmm ) {

      // Reset shower and MPI scales.
      pTmaxISR = pTmaxFSR = pTmaxMPI = scale;

      // Reset to minimal scale for wimpy showers.
      if (limitPTmaxISR) pTmaxISR = min(scale,refScale);
      if (limitPTmaxFSR) pTmaxFSR = min(scale,refScale);
      if (limitPTmaxMPI) pTmaxMPI = min(scale,refScale);

      // For pure QCD set the PS starting scales to the pT of the dijet system.
      if (is2to2QCD) {
        pTmaxFSR = pT2to2;
        pTmaxISR = pT2to2;
      }

      // If necessary, set the MPI starting scale to the collider energy.
      if (hasMPIoverlap) {
        pTmaxMPI = infoPtr->eCM();
        limitPTmaxMPI = false;
      }

    }

    // Reset power/wimpy shower switches iand scales if necessary.
    limitPTmaxFSRIn = limitPTmaxFSR;
    limitPTmaxISRIn = limitPTmaxISR;
    limitPTmaxMPIIn = limitPTmaxMPI;
    pTmaxFSRIn      = pTmaxFSR;
    pTmaxISRIn      = pTmaxISR;
    pTmaxMPIIn      = pTmaxMPI;
    pTscaleIn       = pTscale;

    // Done
    return true;

  }

  //--------------------------------------------------------------------------

  // Calculate the merging scale for the input event.

  double VinciaMergingHooks::tmsCalc(const Event& event) {

    // Only check for first emission.
    if (!isFirstEmission(event)) return 0.0;

    // Find all resonances in the state.
    vector<int> ewResonancePos;
    ewResonancePos.clear();
    for (int i=0; i < event.size(); ++i)
      if ( abs(event[i].status()) == 22 && ( event[i].idAbs() == 6 ||
        (event[i].idAbs() > 21 && event[i].idAbs() < 26) ) )
      ewResonancePos.push_back(i);

    // Declare final parton vectors
    vector <int> FinalPartPos;
    FinalPartPos.clear();
    // Search event record for final state partons.
    // Exclude decay products of resonance.
    for (int i=0; i<(int)event.size(); i++) {
      if (event[i].isFinal() && isInHard(i, event) && event[i].colType() != 0
        && checkAgainstCut(event[i])) {
        bool isDecayProduct = false;
        for (int j=0; j<(int)ewResonancePos.size(); j++)
          if (event[i].isAncestor(ewResonancePos[j])) isDecayProduct = true;
        // Except for e+e- -> jets, do not check radiation in resonance decays.
        if (!isDecayProduct
          || getProcessString().compare("ta+ta->jj") == 0
          || getProcessString().compare("e+e->jj") == 0
          || getProcessString().compare("e+e->(z>jj)") == 0 )
          FinalPartPos.push_back(i);
      }

      // Include photons into the tms definition for "weak+QCD merging".
      if ( getProcessString().find("Ainc") != string::npos
        && event[i].isFinal() && event[i].idAbs() == 22 )
        FinalPartPos.push_back(i);
      // Include Z-bosons into the tms definition for "weak+QCD merging".
      if ( getProcessString().find("Zinc") != string::npos
        && event[i].isFinal() && event[i].idAbs() == 23 )
        FinalPartPos.push_back(i);
      // Include W-bosons into the tms definition for "weak+QCD merging".
      if ( getProcessString().find("Winc") != string::npos
        && event[i].isFinal() && event[i].idAbs() == 24 )
        FinalPartPos.push_back(i);
    }

    // Get index of first incoming
    int in1 = 0;
    for (int i=0; i<(int)event.size(); i++) 
      if (abs(event[i].status()) == 41) {
        in1 = i;
        break;
      }
    // Get index of second incoming
    int in2 = 0;
    for (int i=0; i<(int)event.size(); i++)
      if (abs(event[i].status()) == 42) {
        in2 = i;
        break;
      }

    // If no incoming of the cascade are found, try incoming
    if (in1 == 0 || in2 == 0) {
      // Find current incoming partons
      for (int i=3; i<(int)event.size(); i++) {
        if (!isInHard(i, event)) continue;
        if (event[i].mother1() == 1) in1 = i;
        if (event[i].mother1() == 2) in2 = i;
      }
    }

    int nInitialPartons = 0, nFinalOther = 0;
    for (int i=0; i<(int)event.size(); i++) {
      if ( (event[i].mother1() == 1 || event[i].mother1() == 2)
        && (event[i].idAbs() < 6 || event[i].id() == 21) ) nInitialPartons++;
      if (event[i].isFinal() && event[i].idAbs() >= 6 && event[i].id() != 21)
        nFinalOther++;
    }
    bool is2to2QCD = (FinalPartPos.size() == 2 && nInitialPartons == 2
      && nFinalOther == 0 );

    // For pure QCD set the cut to the pT of the dijet system.
    if (is2to2QCD)
      return min(event[FinalPartPos[0]].pT(), event[FinalPartPos[1]].pT());

    // Find minimal pythia pt in event
    double ptmin = event[0].e();
    for (int i=0; i<(int)FinalPartPos.size(); i++) {
      double pt12 = ptmin;

      // Compute II separation i-jet and first incoming
      if (event[in1].colType() != 0) {
        double temp = rhoVincia( event, in1, FinalPartPos[i], in2, -1 );
        pt12 = min(pt12, temp);
      }

      // Compute II separation i-jet and second incoming
      if (event[in2].colType() != 0) {
        double temp = rhoVincia( event, in2, FinalPartPos[i], in1, -1 );
        pt12 = min(pt12, temp);
      }

      // Compute FF separation between two jets.
      for (int j=0; j<(int)FinalPartPos.size(); j++) {
        for (int k=0; k<(int)FinalPartPos.size(); k++) {
          // Allow any parton as recoiler
          if ( (i != j) && (i != k) && (j != k) ){
            double temp = 0.0;
            temp = rhoVincia( event, FinalPartPos[i], FinalPartPos[j],
                              FinalPartPos[k], 1 );
            pt12 = min(pt12, temp);
            temp = rhoVincia( event, FinalPartPos[j], FinalPartPos[i],
                              FinalPartPos[k], 1 );
            pt12 = min(pt12, temp);
          }
        }
      }

      // Compute IF separation between two jets.
      if (event[in1].colType() != 0 && event[in2].colType() != 0) {
        for (int j=0; j<(int)FinalPartPos.size(); j++) {
          if ( i != j ){
            double temp = pt12;
            // Check with first initial as radiator.
            if (event[in1].colType() != 0)
              temp = rhoVincia( event, in1, FinalPartPos[j], FinalPartPos[i],
                                -1 );
            pt12 = min(pt12, temp);
            // Check with first initial as recoiler
            if (event[in1].colType() != 0) 
              temp = rhoVincia( event, FinalPartPos[i], FinalPartPos[j], in1,
                                1 );
            pt12 = min(pt12, temp);
            // Check with second initial as radiator.
            if (event[in2].colType() != 0)
              temp = rhoVincia( event, in2, FinalPartPos[j], FinalPartPos[i],
                                -1 );
            pt12 = min(pt12, temp);
            // Check with second initial as recoiler.
            if (event[in2].colType() != 0) 
              temp = rhoVincia( event, FinalPartPos[i], FinalPartPos[j], in2,
                                1 );
            pt12 = min(pt12, temp);
          }
        }
      }

      // Reset minimal y separation
      ptmin = min(ptmin,pt12);
    }

    return ptmin;
    
  }

  // Helper function for tms calculation.

  double VinciaMergingHooks::rhoVincia(const Event& event, int rad, int emt,
    int rec, int) {
    map<string,double> stateVars;
    string name;
    bool isFSR = timesAntShowerPtr->isTimelike(event,rad,emt,rec,"");
    if (isFSR) {
      name      = timesAntShowerPtr->getSplittingName(event,rad,emt,rec)[0];
      stateVars = timesAntShowerPtr->getStateVariables(event,rad,emt,rec,name);
    } else {
      name      = spaceAntShowerPtr->getSplittingName(event,rad,emt,rec)[0];
      stateVars = spaceAntShowerPtr->getStateVariables(event,rad,emt,rec,name);
    }
    return ( (stateVars.size() > 0 && stateVars.find("t") != stateVars.end())
             ? sqrt(stateVars["t"]) : 1e15 );
  }

  //--------------------------------------------------------------------------

  // Function to calculate the hard process matrix element.

  double VinciaMergingHooks::hardProcessME(const Event& event) {
    // Simply ask the shower.
    return spaceAntShowerPtr->getBornProb(event);
  }

}
