// ShowerSplittings.cc is a part of the VINCIA plug-in to the PYTHIA 8
// generator. 
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2019 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// TimeAntShower and SpaceAntShower functions used by the matrix-element
// corrections and merging.
// NOTE Only works for Vincia:hyperjet = off (default)
//                     Vincia:massCorrectionSplitF < 2 (default)
//                     Vincia:massCorrectionEmitF = 0 (default)

// Include the Vincia headers 
#include "Vincia/TimeAntShower.h"
#include "Vincia/SpaceAntShower.h"

namespace Vincia {

  const bool printNow    = false;
  const bool printNowAll = false;

  //**********************************************************************
  // TimeAntShower
  //**********************************************************************

  // Return clustered event.
  vector<Event> TimeAntShower::clusteredAll( const Event& eventOld, int iRa, 
    int iEm, int iRe, string name ) {
    if (!isTimelike(eventOld, iRa, iEm, iRe, name)) return vector<Event>();
    string vincia("Vincia:");
    if (name.find(vincia) == string::npos) return vector<Event>();

    Event event = eventOld;

    Particle radiator = event[iRa];
    Particle emission = event[iEm];
    Particle recoiler = event[iRe];

    // Clustering: kinematic part
    vector<int> idsVc;
    vector<Vec4> pVc;
    // Gluon emission or quark from gluon splitting (ij)k
    if (emission.id() == 21 || emission.id() > 0) {
      idsVc.push_back(radiator.id());
      idsVc.push_back(emission.id());
      idsVc.push_back(recoiler.id());
      pVc.push_back(radiator.p());
      pVc.push_back(emission.p());
      pVc.push_back(recoiler.p());
    }
    // Antiquark from gluon splitting i(jk)
    else {
      idsVc.push_back(recoiler.id());
      idsVc.push_back(emission.id());
      idsVc.push_back(radiator.id());
      pVc.push_back(recoiler.p());
      pVc.push_back(emission.p());
      pVc.push_back(radiator.p());
    }
    int ii = 0;
    int ij = 1;
    int ik = 2;
    vector<int> idClu; 
    vector<Vec4> pClu;
    int iant;
    bool swapAB;
    //if (!vinCluPtr->getClu(idsVc, pVc, ii, ij, ik, idClu, pClu, iant, swapAB))
    //  return vector<Event>();
    vinCluPtr->getClu(idsVc, pVc, ii, ij, ik, idClu, pClu, iant, swapAB);

    // Id's and colours
    int idj   = emission.id();
    if (idj == 21) swapAB = true;
    int colj  = emission.col();
    int acolj = emission.acol();
    int idI   = idClu[swapAB ? 0 : 1]; // radiator
    int colI  = radiator.col();
    int acolI = radiator.acol();
    int idK   = idClu[swapAB ? 1 : 0]; // recoiler
    int colK  = recoiler.col();
    int acolK = recoiler.acol();
    // Gluon emission
    if (idj == 21) {
      if (colj == acolK) acolK = colI;
      else if (acolj == colK) colK = acolI;
    }
    // Gluon splitting
    else {
      idI = 21;
      if (colI == 0) colI = colj;
      if (acolI == 0) acolI = acolj;
    }

    // Change id, momentum, colour
    // Replace radiator and recoiler in event record
    event[iRa].id(idI);
    event[iRa].p(pClu[swapAB ? 0 : 1]);
    event[iRa].col(colI);
    event[iRa].acol(acolI);

    event[iRe].id(idK);
    event[iRe].p(pClu[swapAB ? 1 : 0]);
    event[iRe].col(colK);
    event[iRe].acol(acolK);

    // Remove emission from event record
    event.remove(iEm, iEm);
    // Remove emission from mother/daughter lists
    for (int i=0; i<event.size(); i++) {
      // Mother1
      if (event[i].mother1() > iEm) 
        event[i].mother1(event[i].mother1()-1);
      else if (event[i].mother1() == iEm) {
        if (event[i].mother2() == iEm) 
          event[i].mothers(0, 0);
        else if (event[i].mother2() > iEm)
          event[i].mothers(event[i].mother2()-1, 0);
        else
          event[i].mothers(event[i].mother2(), 0);
      }
      // Mother2
      if (event[i].mother2() > iEm) 
        event[i].mother2(event[i].mother2()-1);
      else if (event[i].mother2() == iEm) 
        event[i].mother2(0);
      // Daugther1
      if (event[i].daughter1() > iEm) 
        event[i].daughter1(event[i].daughter1()-1);
      else if (event[i].daughter1() == iEm) {
        if (event[i].daughter2() == iEm)
          event[i].daughters(0, 0);
        else if (event[i].daughter2() > iEm)
          event[i].daughters(iEm, event[i].daughter2()-1);
        else
          event[i].daughters(event[i].daughter2(), 0);
      }
      // Daugther2
      if (event[i].daughter2() > iEm) 
        event[i].daughter2(event[i].daughter2()-1);
      else if (event[i].daughter2() == iEm) {
        if (event[i].daughter1() < iEm-1)
          event[i].daughters(event[i].daughter1(), iEm-1);
        else 
          event[i].daughter2(0);
      }
    }

    // Check energy-momentum conservation
    if (verbose >= 3)  {
      // Start with incoming partons = beam daughters
      Vec4 pTot = -(event[event[1].daughter1()].p() + 
                    event[event[2].daughter1()].p());
      double eLab = abs(pTot.e());
      // Sum final state
      for (int i=0; i<event.size(); ++i)
        if (event[i].isFinal()) pTot += event[i].p();
      // Check energy-momentum/charge conservation.
      double epDev = abs( pTot.e()) + abs(pTot.px()) + abs(pTot.py())
                                    + abs(pTot.pz() );
      if (epDev > vinComPtr->epTolErr * eLab)
        infoPtr->errorMsg("Error in TimeAntShower::clustered: energy-"
          "momentum not conserved");
      else if (epDev > vinComPtr->epTolWarn * eLab)
        infoPtr->errorMsg("Warning in TimeAntShower::clustered: energy-"
          "momentum not quite conserved");
    }

if (printNow) { 
cout << " TimeAntShower::clustered with size = " << eventOld.size()
     << " iEm = " << iEm << " iRa = " << iRa << " iRe = " << iRe << endl;
cout << " Unclustered event for " << name << ":" << endl;
if (printNowAll) {
eventOld.list(); cout << endl;
cout << " Clustered event for " << name << ":" << endl;
event.list(); cout << endl;
}
cout << endl;
}

    vector<Event> clusteredEvents;
    clusteredEvents.push_back(event);

    // Helicities
    if (helicityShower) {
      // Gluon emission: parents inherit helicity, we are done
      // Gluon splitting: recoiler inherits helicity, gluon can be both
      if (emission.id() != 21) {
        clusteredEvents.push_back(event);
        // Emission has been removed, so iRa might have changed
        int iRaUp = ( iRa < iEm ? iRa : iRa-1 );
        clusteredEvents[0][iRaUp].pol( 1);
        clusteredEvents[1][iRaUp].pol(-1);
      }
    }

    return clusteredEvents;
  }

  // Return the evolution variable.
  vector< map<string, double> > TimeAntShower::getStateVariablesAll(
    const Event& event, int iRa, int iEm, int iRe, string name) {
    vector< map<string, double> > stateVariables;
    stateVariables.clear();
    map<string, double> evolVals;
    evolVals.clear();

if (printNow)
cout << " TimeAntShower::getStateVariables with size = " << event.size()
     << " iEm = " << iEm << " iRa = " << iRa << " iRe = " << iRe << endl;

    // Born with iRa=0, iEm=0, iRe=0
    if ( (iRa == 0) && (iEm == 0) && (iRe == 0) ) {
      double scale = getBornScale(event);
      if (scale > 0.0) {
        evolVals.insert( pair<string, double>("scalePDF", scale) );
        stateVariables.push_back(evolVals);
      }
      return stateVariables;
    }

    string vincia("Vincia:");
    if (name.find(vincia) == string::npos) return stateVariables;

    // Erase the Vincia:XX and FF part to minimize string checks
    name.erase(name.begin(), name.begin()+vincia.size()+2);
    name.erase(name.end()-2, name.end());

    // Evolution variable
    double Q2evol  = 0.0;
    double kMu2    = 0.0;
    double normFac = 1.0;

    // 1. Gluon emission: symmetric
    if (name == "emit") {
      Q2evol  = resolutionPtr->Q2E(event[iRa], event[iEm], event[iRe], 
                                   evolutionType);
      kMu2    = aSkMu2;
      normFac = evNormFac;
if (printNow) cout << fixed << setprecision(15) << " emit (FF) = "
<< sqrt(Q2evol) << endl;
    }

    // 2. Gluon splitting: sjk
    else if (name == "split") {
      Q2evol = resolutionPtr->Q2E(event[iRe], event[iEm], event[iRa], 7);
      kMu2   = aSkMu2Split;
if (printNow) cout << fixed << setprecision(15) << " split (FF) = "
<< sqrt(Q2evol) << endl;
    }

    // 3. Not found, use for pTLund
    else if (name == "other") {
      Q2evol = resolutionPtr->Q2E(event[iRe], event[iEm], event[iRa], 7);
      kMu2   = aSkMu2Split;
if (printNow) cout << fixed << setprecision(15) << " other (FF) = "
<< sqrt(Q2evol) << endl;
    }

    // Restart scale
    double Q2restart = Q2evol;

    // AlphaS and PDF scales
    double Q2alphaSraw = kMu2*Q2evol/normFac;
    double Q2alphaS    = max( vinComPtr->mu2min, 
                              vinComPtr->mu2freeze + Q2alphaSraw );
    double Q2pdf       = Q2evol;

if (printNowAll) {    
cout << " Q2evol    for " << name << " = " << Q2evol << endl;
cout << " Q2restart for " << name << " = " << Q2restart << endl;
cout << " Q2alphaS  for " << name << " = " << Q2alphaS << endl;
cout << " Q2pdf     for " << name << " = " << Q2pdf << endl;
}

    // Insert values in map
    evolVals.insert( pair<string, double>("t",          Q2evol)      );
    evolVals.insert( pair<string, double>("tRS",        Q2restart)   );
    evolVals.insert( pair<string, double>("scaleAS",    Q2alphaS)    );
    evolVals.insert( pair<string, double>("scaleASraw", Q2alphaSraw) );
    evolVals.insert( pair<string, double>("scalePDF",   Q2pdf)       );

if (printNow) cout << endl;

    stateVariables.push_back(evolVals);

    // Helicities
    if (helicityShower && (name == "split"))
      stateVariables.push_back(evolVals);

    return stateVariables;
  }

  // Check if attempted clustering is handled by timelike shower
  bool TimeAntShower::isTimelike( const Event& event, int iRa, int iEm, 
    int iRe, string ) {
    // Born with iRa=0, iEm=0, iRe=0
    if ( (iRa == 0) && (iEm == 0) && (iRe == 0) )
      return ( event[3].isLepton() && event[4].isLepton() );
    // Check if all have colours
    if ( event[iRa].colType() == 0 ||
         event[iRe].colType() == 0 ||
         event[iEm].colType() == 0 ) return false;
    return ( event[iRa].isFinal() && event[iEm].isFinal() 
                                  && event[iRe].isFinal() );
  }

  // Return a string identifier of a splitting.
  vector<string> TimeAntShower::getSplittingName( const Event& event, int iRa, 
    int iEm, int iRe ) { 
    if (!isTimelike(event, iRa, iEm, iRe, "")) return vector<string>();

    string name = "";

    int idEm = event[iEm].idAbs();

    // 1. Gluon emission
    if (idEm == 21) {
      int idRa = event[iRa].idAbs();
      int idRe = event[iRe].idAbs();
           if ( (idRa  <  7) && (idRe  <  7) ) name = "Vincia:QQemitFF";
      else if ( (idRa == 21) && (idRe == 21) ) name = "Vincia:GGemitFF";
      else if ( (idRa == 21) && (idRe  <  7) ) name = "Vincia:QGemitFF";
      else if ( (idRa  <  7) && (idRe == 21) ) name = "Vincia:QGemitFF";
    }

    // 2. Gluon splitting
    // emission id = - radiator id
    else if (idEm < 7) {
      if (event[iEm].id() == -event[iRa].id()) {
        int idRe = event[iRe].idAbs();
             if (idRe == 21) name = "Vincia:GGsplitFF";
        else if (idRe  <  7) name = "Vincia:QGsplitFF";
      }
      else name = "Vincia:XXotherFF";
    }

if (printNowAll) {    
event.list();
cout << " TimeAntShower::getSplittingName with size = " << event.size()
     << " iEm = " << iEm << " iRa = " << iRa << " iRe = " << iRe << endl;
if (name != "") cout << " Splitting name = " << name << endl;
else cout << " Did not find splitting" << endl;
}

    vector<string> names;
    names.push_back(name);

    // Helicities
    if (helicityShower && (name == "Vincia:GGsplitFF" || 
      name == "Vincia:QGsplitFF")) names.push_back(name);

    return names;
  }

  // Return the splitting probability.
  vector<double> TimeAntShower::getSplittingProbAll( const Event& event,
    int iRa, int iEm, int iRe, string name ) {
    if (!isTimelike(event, iRa, iEm, iRe, name)) return vector<double>();
    string vincia("Vincia:");
    if (name.find(vincia) == string::npos) return vector<double>();

if (printNow)
cout << " TimeAntShower::getSplittingProb with size = " << event.size()
     << " iEm = " << iEm << " iRa = " << iRa << " iRe = " << iRe << endl;
    
    vector<double> splitProbs;
    
    // Find the right antenna
    for (int iant = iQQemit; iant <= iGGsplit; iant++) {
      if ( !caseInsensComp(name, getAnt(iant)->vinciaName()) ) continue;

      // Swap iRa and iRe for QGemit if Ra is gluon
      // and for gluon splittings (sjk singularity in antenna function)
      if ( ((iant == iQGemit) && (event[iRa].id() == 21))
           || (event[iEm].id() != 21) ) {
        int tmp = iRa;
        iRa = iRe;
        iRe = tmp;
      }

      double sik = (event[iRa].p()+event[iRe].p()).m2Calc();
      double sij = (event[iRa].p()+event[iEm].p()).m2Calc();
      double sjk = (event[iEm].p()+event[iRe].p()).m2Calc();
      double sIK = sik + sij + sjk;

      int polI = event[iRa].pol(), polK = event[iRe].pol();
      int poli = polI, polk = polK, polj = event[iEm].pol();

      double mui = 0.0, muj = 0.0, muk = 0.0;

      // Antenna function and colour factor
      vector<double> ants;
      // Helicities
      if (helicityShower) {
        // Gluon emission: parents inherit helicity, we are done
        // Gluon splitting: recoiler inherits helicity, gluon can be both
        if (event[iEm].id() != 21) {
          ants.push_back( getAnt(iant)->antennaFunction(sij/sIK, sjk/sIK, 
            mui, muj, muk, polI,  1, poli, polj, polk ) / sIK );
          ants.push_back( getAnt(iant)->antennaFunction(sij/sIK, sjk/sIK, 
            mui, muj, muk, polI, -1, poli, polj, polk ) / sIK );
        } else {
          ants.push_back( getAnt(iant)->antennaFunction(sij/sIK, sjk/sIK, 
            mui, muj, muk, polI, polK, poli, polj, polk ) / sIK );
        }
      } else ants.push_back( getAnt(iant)->antennaFunction(sij/sIK, sjk/sIK)
        / sIK );
      double colFac = getAnt(iant)->getChargeFac();

      // Phase Space factor, not inclued for Vincia
      double PSfac = 1.0/sIK;
      PSfac = 1.0;

if (printNowAll) cout << " col*ant = " << colFac*ants[0] << endl << " PSfac   = "
<< PSfac << endl;

      for (int i=0; i<(int)ants.size(); i++) 
        splitProbs.push_back(PSfac*colFac*ants[i]);
      break;
    }

if (printNowAll) event.list();
if (printNow)
cout << " Splitting probability for " << name << " = "
     << scientific << splitProbs[0] << endl << endl;

    return splitProbs;
  }

  // Return matrix-element correction.
  double TimeAntShower::getMEC( const Event& event, int nStep) {
    matrixElementCorrectionsPtr->prepMECs(event, incomingType);
    return matrixElementCorrectionsPtr->performMECs(event, this, 
      ISRshowerPtr, nStep);
  }

  // Return scales, alphaS values, etc in case of only unordered shower paths.
  map<string, double> TimeAntShower::getUnorderedInfos( const Event& event) {
    matrixElementCorrectionsPtr->prepMECs(event, incomingType);
    return matrixElementCorrectionsPtr->getUnorderedInfos(event, this, 
      ISRshowerPtr);
  }

  // Return the Born probability (= ME).
  double TimeAntShower::getBornProb( const Event& event) {
    matrixElementCorrectionsPtr->prepMECs(event, incomingType);
    Event evtCopy = event;
    vector<Particle> eventParts = convertEventToParticles(evtCopy,
      matrixElementCorrectionsPtr->getProcClass());
    double ME = matrixElementCorrectionsPtr->evalME(eventParts);
    if (ME < 0.0) return 1.0;
    return ME;
  }

  // Return the Born scale for merging.
  double TimeAntShower::getBornScale( const Event& event) {
    return ISRshowerPtr->getBornScale(event);
  }

  // Return the shower starting scale for pure shower.
  double TimeAntShower::getShowerStartingScale( const Event& event) {
    return vinComPtr->getShowerStartingScale(event, sbbSav);
  }

  //**********************************************************************
  // SpaceAntShower
  //**********************************************************************

  // Return clustered event.
  vector<Event> SpaceAntShower::clusteredAll( const Event& eventOld, int iRa, 
    int iEm, int iRe, string name ) {
    if (!isSpacelike(eventOld, iRa, iEm, iRe, name)) return vector<Event>();
    string vincia("Vincia:");
    if (name.find(vincia) == string::npos) return vector<Event>();

    Event event = eventOld;

    Particle radiator = event[iRa];
    Particle emission = event[iEm];
    Particle recoiler = event[iRe];

    // Check if II or IF
    bool isII = ( !radiator.isFinal() && !recoiler.isFinal() );

    // II - start
    if (isII) {

      double sab = (radiator.p()+recoiler.p()).m2Calc();
      double saj = (radiator.p()+emission.p()).m2Calc();
      double sjb = (emission.p()+recoiler.p()).m2Calc();
      double sAB = sab - saj - sjb;
      if (sAB < 0.0) return vector<Event>();

      // Clustering: kinematic part
      Vec4 pA = radiator.p();
      Vec4 pB = recoiler.p();
      double rescaleFacA = sqrt(sAB/sab*(sAB+saj)/(sAB+sjb));
      double rescaleFacB = sqrt(sAB/sab*(sAB+sjb)/(sAB+saj));
      pA.rescale4(rescaleFacA);
      pB.rescale4(rescaleFacB);
      // Check energys -> is clustering possible?
      if ( (pA.e() < 0.0) || (pA.e() > radiator.e()) ) return vector<Event>();
      if ( (pB.e() < 0.0) || (pB.e() > recoiler.e()) ) return vector<Event>();

      // Id's and colours
      int idA   = radiator.id();
      int colA  = radiator.col();
      int acolA = radiator.acol();
      int idB   = recoiler.id();
      int colB  = recoiler.col();
      int acolB = recoiler.acol();
      int idj   = emission.id();
      int colj  = emission.col();
      int acolj = emission.acol();
      // Gluon emission
      if (idj == 21) {
        if (colj == colB) colB = acolA;
        else if (acolj == acolB) acolB = colA;
      }
      // Initial gluon splitting
      else if (name == "Vincia:QXsplitII") {
        idA = -idj;
        if (idA > 0) acolA = 0;
        else if (idA < 0) colA = 0;
      }
      // Initial gluon conversion
      else if (name == "Vincia:GXconvII") {
        idA = 21;
        if (colA == 0) colA = acolj;
        if (acolA == 0) acolA = colj;
      } else return vector<Event>();

      // Replace radiator and recoiler in event record
      event[iRa].id(idA);
      event[iRa].p(pA);
      event[iRa].col(colA);
      event[iRa].acol(acolA);

      event[iRe].id(idB);
      event[iRe].p(pB);
      event[iRe].col(colB);
      event[iRe].acol(acolB);

      // Particle vector for FS/IM recoilers
      vector<Particle> FSrecoilers; 
      FSrecoilers.resize(0);
      vector<Particle> IMrecoilers;
      IMrecoilers.resize(0);
      for (int i=0; i<event.size(); i++) {
        // Final state
        if (event[i].isFinal() && (i != iEm)) FSrecoilers.push_back(event[i]);
        // Intermediate state, if present
        else if (event[i].statusAbs() == 22) IMrecoilers.push_back(event[i]);
      }
      // 1) Boost in center-of-mass frame AB, from E,0,0,pz to E,0,0,0
      Vec4 pSum = pA + pB;
      // 2) Boost from E,0,0,0 to E',px',py',pz' 
      //    (FSrecoilers before clustering in lab frame)
      Vec4 pRec(0.0, 0.0, 0.0, 0.0);
      for (int i=0; i<(int)FSrecoilers.size(); i++) pRec += FSrecoilers[i].p();
      // Now we perform the boost back
      for (int i=0; i<(int)FSrecoilers.size(); i++) {
        FSrecoilers[i].bstback(pRec);
        FSrecoilers[i].bst(pSum);
      }
      for (int i=0; i<(int)IMrecoilers.size(); i++) {
        IMrecoilers[i].bstback(pRec);
        IMrecoilers[i].bst(pSum);
      }

      // Change momentum in event record
      int countIM = 0;
      int countFS = 0;
      for (int i=0; i<event.size(); i++) {
        if (event[i].isFinal() && (i != iEm)) {
          event[i].p(FSrecoilers[countFS].p());
          countFS++;
        } else if (event[i].statusAbs() == 22) {
          event[i].p(IMrecoilers[countIM].p());
          countIM++;
        }
      }

    }
    // II - end

    // IF - start
    else {

      Particle isa = (radiator.isFinal() ? recoiler : radiator);
      Particle fsk = (radiator.isFinal() ? radiator : recoiler);
      double sak = (isa.p()+fsk.p()).m2Calc();
      double saj = (isa.p()+emission.p()).m2Calc();
      double sjk = (emission.p()+fsk.p()).m2Calc();
      double sAK = sak + saj - sjk;
      if (sAK < 0.0) return vector<Event>();

      // Clustering: kinematic part
      Vec4 pA = isa.p();
      pA.rescale4(sAK/(sAK+sjk));
      Vec4 pK = pA-isa.p()+fsk.p()+emission.p();
      // Check energys -> is clustering possible?
      if ( (pA.e() < 0.0) || (pK.e() < 0.0) ) return vector<Event>();

      // Id's and colours
      int idA    = isa.id();
      int colA   = isa.col();
      int acolA  = isa.acol();
      int idK    = fsk.id();
      int colK   = fsk.col();
      int acolK  = fsk.acol();
      int idj    = emission.id();
      int colj   = emission.col();
      int acolj  = emission.acol();
      // Gluon emission
      if (idj == 21) {
        if (colj == acolK) acolK = acolA;
        else if (acolj == colK) colK = colA;
      }
      // Initial gluon splitting
      else if (name == "Vincia:QXsplitIF") {
        idA = -idj;
        if (idA > 0) acolA = 0;
        else if (idA < 0) colA = 0;
      }
      // Initial gluon conversion
      else if (name == "Vincia:GXconvIF") {
        idA = 21;
        if (colA == 0) colA = acolj;
        if (acolA == 0) acolA = colj;
      }
      // Final gluon splitting
      else if (name == "Vincia:XGsplitIF") {
        idK = 21;
        if (colK == 0) colK = colj;
        if (acolK == 0) acolK = acolj;
      } else return vector<Event>();
      
      // Replace radiator and recoiler in event record
      if (event[iRa].isFinal()) {
        event[iRe].id(idA);
        event[iRe].p(pA);
        event[iRe].col(colA);
        event[iRe].acol(acolA);

        event[iRa].id(idK);
        event[iRa].p(pK);
        event[iRa].col(colK);
        event[iRa].acol(acolK);
      } else {
        event[iRa].id(idA);
        event[iRa].p(pA);
        event[iRa].col(colA);
        event[iRa].acol(acolA);

        event[iRe].id(idK);
        event[iRe].p(pK);
        event[iRe].col(colK);
        event[iRe].acol(acolK);
      }

    }
    // IF - end

    // Remove emission from event record
    event.remove(iEm, iEm);
    // Remove emission from mother/daughter lists
    for (int i=0; i<event.size(); i++) {
      // Mother1
      if (event[i].mother1() > iEm) 
        event[i].mother1(event[i].mother1()-1);
      else if (event[i].mother1() == iEm) {
        if (event[i].mother2() == iEm) 
          event[i].mothers(0, 0);
        else if (event[i].mother2() > iEm)
          event[i].mothers(event[i].mother2()-1, 0);
        else
          event[i].mothers(event[i].mother2(), 0);
      }
      // Mother2
      if (event[i].mother2() > iEm) 
        event[i].mother2(event[i].mother2()-1);
      else if (event[i].mother2() == iEm) 
        event[i].mother2(0);
      // Daugther1
      if (event[i].daughter1() > iEm) 
        event[i].daughter1(event[i].daughter1()-1);
      else if (event[i].daughter1() == iEm) {
        if (event[i].daughter2() == iEm)
          event[i].daughters(0, 0);
        else if (event[i].daughter2() > iEm)
          event[i].daughters(iEm, event[i].daughter2()-1);
        else
          event[i].daughters(event[i].daughter2(), 0);
      }
      // Daugther2
      if (event[i].daughter2() > iEm) 
        event[i].daughter2(event[i].daughter2()-1);
      else if (event[i].daughter2() == iEm) {
        if (event[i].daughter1() < iEm-1)
          event[i].daughters(event[i].daughter1(), iEm-1);
        else 
          event[i].daughter2(0);
      }
    }

    // Check energy-momentum conservation
    if (verbose >= 3)  {
      // Start with incoming partons = beam daughters
      Vec4 pTot = -(event[event[1].daughter1()].p() + 
                    event[event[2].daughter1()].p());
      double eLab = abs(pTot.e());
      // Sum final state
      for (int i=0; i<event.size(); ++i)
        if (event[i].isFinal()) pTot += event[i].p();
      // Check energy-momentum/charge conservation.
      double epDev = abs( pTot.e()) + abs(pTot.px()) + abs(pTot.py())
                                    + abs(pTot.pz() );
      if (epDev > vinComPtr->epTolErr * eLab)
        infoPtr->errorMsg("Error in SpaceAntShower::clustered: energy-"
          "momentum not conserved");
      else if (epDev > vinComPtr->epTolWarn * eLab)
        infoPtr->errorMsg("Warning in SpaceAntShower::clustered: energy-"
          "momentum not quite conserved");
    }

if (printNow) { 
cout << " SpaceAntShower::clustered with size = " << eventOld.size()
     << " iEm = " << iEm << " iRa = " << iRa << " iRe = " << iRe << endl;
cout << " Unclustered event for " << name << ":" << endl;
if (printNowAll) {
eventOld.list(); cout << endl;
cout << " Clustered event for " << name << ":" << endl;
event.list(); cout << endl;
}
cout << endl;
}

    vector<Event> clusteredEvents;
    clusteredEvents.push_back(event);

    // Helicities
    if (helicityShower) {
      // Emission has been removed, so iRa/iRe might have changed
      int iRaUp = -1, iReUp = -1;
      // Gluon emission: final state parent and initial state quark inherit
      // helicity, initial state gluon can be both if same helicities after
      if (emission.id() == 21) {
        if (isII) {
          // Check both parents
          if ( radiator.id() == 21 && radiator.pol() == emission.pol() )
            iRaUp = ( iRa < iEm ? iRa : iRa-1 );
          if ( recoiler.id() == 21 && recoiler.pol() == emission.pol() )
            iReUp = ( iRe < iEm ? iRe : iRe-1 );
        } else {
          // Check initial state
          int iRaNow = (radiator.isFinal() ? iRe : iRa);
              iRaUp  = ( iRaNow < iEm ? iRaNow : iRaNow-1 );
          if ( clusteredEvents[0][iRaUp].id() != 21 ||
            clusteredEvents[0][iRaUp].pol() != emission.pol() ) iRaUp = -1;
        }
      } else {
        // Gluon backwards evolving to a quark: recoiler inherits helicity,
        // gluon can be both
        if (name == "Vincia:GXconvII" || name == "Vincia:GXconvIF") {
          int iRaNow = (radiator.isFinal() ? iRe : iRa);
              iRaUp  = ( iRaNow < iEm ? iRaNow : iRaNow-1 );
        }
        // Final state gluon splitting: recoiler inherits helicity, gluon can
        // be both
        else if (name == "Vincia:XGsplitIF") {
          int iRaNow = (radiator.isFinal() ? iRa : iRe);
              iRaUp  = ( iRaNow < iEm ? iRaNow : iRaNow-1 );
        }
        // Quark backwards evolving to gluon: recoiler inherits helicity,
        // quark has -pol of emission antiquark
        else if (name == "Vincia:QXsplitII" || name == "Vincia:QXsplitIF") {
          int iRaNow = (radiator.isFinal() ? iRe : iRa);
              iRaNow = ( iRaNow < iEm ? iRaNow : iRaNow-1 );
          clusteredEvents[0][iRaNow].pol(-emission.pol());
        }
      }
      if (iRaUp > 0 && iReUp > 0) {
        for (int i=0; i<2; i++) clusteredEvents.push_back(event);
        clusteredEvents[1][iRaUp].pol(-clusteredEvents[1][iRaUp].pol());
        clusteredEvents[2][iReUp].pol(-clusteredEvents[2][iReUp].pol());
      } else if (iRaUp > 0) {
        clusteredEvents.push_back(event);
        clusteredEvents[0][iRaUp].pol( 1);
        clusteredEvents[1][iRaUp].pol(-1);
      } else if (iReUp > 0) {
        clusteredEvents.push_back(event);
        clusteredEvents[0][iReUp].pol( 1);
        clusteredEvents[1][iReUp].pol(-1);
      }
    }

    return clusteredEvents;
  }

  // Return the evolution variable.
  vector< map<string, double> > SpaceAntShower::getStateVariablesAll(
    const Event& event, int iRa, int iEm, int iRe, string name) {
    vector< map<string, double> > stateVariables;
    stateVariables.clear();
    map<string, double> evolVals;
    evolVals.clear();

if (printNow)
cout << " SpaceAntShower::getStateVariables with size = " << event.size()
     << " iEm = " << iEm << " iRa = " << iRa << " iRe = " << iRe << endl;

    // Born with iRa=0, iEm=0, iRe=0
    if ( (iRa == 0) && (iEm == 0) && (iRe == 0) ) {
      double scale = getBornScale(event);
      if (scale > 0.0) {
        evolVals.insert( pair<string, double>("scalePDF", scale) );
        stateVariables.push_back(evolVals);
      }
      return stateVariables;
    }

    string vincia("Vincia:");
    if (name.find(vincia) == string::npos) return stateVariables;

    // Erase the Vincia: and II/IF part to minimize string checks
    name.erase(name.begin(), name.begin()+vincia.size());
    string nameSav = name;
    name.erase(name.end()-2, name.end());

    bool isII = ( !event[iRa].isFinal() && !event[iRe].isFinal() );
    int idEm = event[iEm].idAbs();

    // Evolution variable
    double Q2evol  = 0.0;
    double kMu2    = 0.0;
    double normFac = 1.0;

    // 1. Gluon emission:
    if (idEm == 21) {
      kMu2 = aSkMu2;
      // 1.1 from II, symmetric
      if (isII) {
        Q2evol  = resolutionPtr->Q2EII(event[iRa], event[iRe], event[iEm], 1);
        normFac = evNormIIpT2;
if (printNow) cout << fixed << setprecision(15) << " emit (II) = "
<< sqrt(Q2evol) << endl;
      } else {
        normFac = evNormIFpT2;
        // 1.2 from IF side1
        if (!event[iRa].isFinal())
          Q2evol = resolutionPtr->Q2EIF(event[iRa], event[iRe], event[iEm], 1);
        // 1.2 from IF side2
        else if (!event[iRe].isFinal())
          Q2evol = resolutionPtr->Q2EIF(event[iRe], event[iRa], event[iEm], 1);
if (printNow) cout << fixed << setprecision(15) << " emit (IF/FI) = "
<< sqrt(Q2evol) << endl;
      }
    }

    // Non-emission
    else if (idEm < 7) {

      // 2. Initial gluon splitting: saj
      if (name == "QXsplit") {
        kMu2    = aSkMu2SplitI;
        normFac = evNormMsplitI;
        // 2.1 from II
        if (isII) {
          Q2evol = resolutionPtr->Q2EII(event[iRa], event[iRe], event[iEm], 
                                        !evAllPT ? 2 : 1);
          if (evAllPT) normFac = evNormIIpT2;
if (printNow) cout << fixed << setprecision(15) << " split (II) = "
<< sqrt(Q2evol) << endl;
        }
        // 2.2 form IF
        else {
          Q2evol = resolutionPtr->Q2EIF(event[iRa], event[iRe], event[iEm], 
                                        !evAllPT ? 2 : 1);
          if (evAllPT) normFac = evNormIFpT2;
if (printNow) cout << fixed << setprecision(15) << " split (IF) = "
<< sqrt(Q2evol) << endl;
        }
      }

      // 3. Gluon conversion: saj
      else if (name == "GXconv") {
        kMu2    = aSkMu2Conv;
        normFac = evNormMconvI;
        // 3.1 from II
        if (isII) {
          Q2evol = resolutionPtr->Q2EII(event[iRa], event[iRe], event[iEm], 
                                        !evAllPT ? 3 : 1);
          if (evAllPT) normFac = evNormIIpT2;
if (printNow) cout << fixed << setprecision(15) << " conv (II) = "
<< sqrt(Q2evol) << endl;
        } 
        // 3.2 form IF
        else {
          Q2evol = resolutionPtr->Q2EIF(event[iRa], event[iRe], event[iEm], 
                                        !evAllPT ? 3 : 1);
          if (evAllPT) normFac = evNormIFpT2;
if (printNow) cout << fixed << setprecision(15) << " conv (IF) = "
<< sqrt(Q2evol) << endl;
        }
      }

      // 4. Final gluon splitting: sjk
      else if (name == "XGsplit") {
        Q2evol  = resolutionPtr->Q2EIF(event[iRe], event[iRa], event[iEm], 
                                       !evAllPT ? 4 : 1);
        kMu2    = aSkMu2SplitF;
        normFac = (evAllPT ? evNormIFpT2 : evNormMsplitF);
if (printNow) cout << fixed << setprecision(15) << " split (FI) = "
<< sqrt(Q2evol) << endl;
      }

      // 5. Not found, use for pTLund
      else {
        if (nameSav == "XXotherFI") {
          Q2evol  = resolutionPtr->Q2EIF(event[iRe], event[iRa], event[iEm], 
                                        !evAllPT ? 4 : 1);
          kMu2    = aSkMu2SplitF;
          normFac = (evAllPT ? evNormIFpT2 : evNormMsplitF);
if (printNow) cout << fixed << setprecision(15) << " other (FI) = "
<< sqrt(Q2evol) << endl;
        } else if (nameSav == "XXotherII") {
          Q2evol  = resolutionPtr->Q2EII(event[iRa], event[iRe], event[iEm], 
                                         !evAllPT ? 2 : 1);
          kMu2    = aSkMu2Conv;
          normFac = (evAllPT ? evNormIIpT2 : evNormMconvI);
if (printNow) cout << fixed << setprecision(15) << " other (II) = "
<< sqrt(Q2evol) << endl;
        } else if (nameSav == "XXotherIF") {
          Q2evol  = resolutionPtr->Q2EIF(event[iRa], event[iRe], event[iEm], 
                                         !evAllPT ? 2 : 1);
          kMu2    = aSkMu2Conv;
          normFac = (evAllPT ? evNormIFpT2 : evNormMconvI);
if (printNow) cout << fixed << setprecision(15) << " other (IF) = "
<< sqrt(Q2evol) << endl;
        }
      }
    }

    // Restart scale
    double Q2restart = Q2evol;

    // AlphaS and PDF scales
    double Q2alphaSraw = kMu2*Q2evol/normFac;
    double Q2alphaS    = max( vinComPtr->mu2min, 
                              vinComPtr->mu2freeze + Q2alphaSraw );
    double Q2pdf       = Q2evol;

if (printNowAll) {
cout << " Q2evol    for " << name << " = " << Q2evol << endl;
cout << " Q2restart for " << name << " = " << Q2restart << endl;
cout << " Q2alphaS  for " << name << " = " << Q2alphaS << endl;
cout << " Q2pdf     for " << name << " = " << Q2pdf << endl;
}

    // Insert values in map
    evolVals.insert( pair<string, double>("t",          Q2evol)      );
    evolVals.insert( pair<string, double>("tRS",        Q2restart)   );
    evolVals.insert( pair<string, double>("scaleAS",    Q2alphaS)    );
    evolVals.insert( pair<string, double>("scaleASraw", Q2alphaSraw) );
    evolVals.insert( pair<string, double>("scalePDF",   Q2pdf)       );

if (printNow) cout << endl;

    stateVariables.push_back(evolVals);

    // Helicities
    if (helicityShower) {
      if (name == "XGsplit" || name == "GXconv")
        stateVariables.push_back(evolVals);
      else if (nameSav == "GGemitII") {
        if ( event[iEm].pol() == event[iRa].pol() && 
          event[iEm].pol() == event[iRe].pol() ) {
          for (int i=0; i<2; i++) stateVariables.push_back(evolVals);
        } else if ( event[iEm].pol() == event[iRa].pol() ||
          event[iEm].pol() == event[iRe].pol() )
            stateVariables.push_back(evolVals);
      } else if (nameSav == "GQemitII") {
        int iRaNow = ( event[iRa].id() == 21 ? iRa : iRe );
        if (event[iEm].pol() == event[iRaNow].pol())
          stateVariables.push_back(evolVals);
      } else if (nameSav == "GGemitIF" || nameSav == "GQemitIF") {
        int iRaNow = (event[iRa].isFinal() ? iRe : iRa);
        if (event[iEm].pol() == event[iRaNow].pol())
          stateVariables.push_back(evolVals);
      }
    }

    return stateVariables;
  }

  // Check if attempted clustering is handled by spacelike shower
  bool SpaceAntShower::isSpacelike( const Event& event, int iRa, int iEm, 
    int iRe, string ) {
    // Born with iRa=0, iEm=0, iRe=0
    if ( (iRa == 0) && (iEm == 0) && (iRe == 0) )
      return ( !event[3].isLepton() || !event[4].isLepton() );
    // Check if all have colours
    if ( event[iRa].colType() == 0 ||
         event[iRe].colType() == 0 ||
         event[iEm].colType() == 0 ) return false;
    return !( event[iRa].isFinal() && event[iEm].isFinal() 
                                   && event[iRe].isFinal() );
  }

  // Return a string identifier of a splitting.
  vector<string> SpaceAntShower::getSplittingName( const Event& event, int iRa, 
    int iEm, int iRe ) { 
    if (!isSpacelike(event, iRa, iEm, iRe, "")) return vector<string>();

    string name = "";

    bool iRaIS = !event[iRa].isFinal();
    bool iReIS = !event[iRe].isFinal();
    bool isII  = (iRaIS && iReIS);
    int  idRa  = event[iRa].idAbs();
    int  idEm  = event[iEm].idAbs();
    int  idRe  = event[iRe].idAbs();

    // 1. Gluon emission
    if (idEm == 21) {
      // 1.1 from II
      if (isII) {
             if ( (idRa  <  7) && (idRe  <  7) ) name = "Vincia:QQemitII";
        else if ( (idRa == 21) && (idRe == 21) ) name = "Vincia:GGemitII";
        else if ( (idRa == 21) && (idRe  <  7) ) name = "Vincia:GQemitII";
        else if ( (idRa  <  7) && (idRe == 21) ) name = "Vincia:GQemitII";
      }
      // 1.2 from IF
      else if (iRaIS || iReIS) {
        int idIS = (iReIS ? idRe : idRa);
        int idFS = (iReIS ? idRa : idRe);
             if ( (idIS  <  7) && (idFS  <  7) ) name = "Vincia:QQemitIF";
        else if ( (idIS == 21) && (idFS == 21) ) name = "Vincia:GGemitIF";
        else if ( (idIS == 21) && (idFS  <  7) ) name = "Vincia:GQemitIF";
        else if ( (idIS  <  7) && (idFS == 21) ) name = "Vincia:QGemitIF";
      }
    }

    // Non-emission
    else if (idEm < 7) {

      // 2. Initial gluon splitting
      // radiator = initial gluon, radiator in initial state
      if ( (idRa == 21) && iRaIS ) {
        // 2.1 from II
        if (isII) name = "Vincia:QXsplitII";
        // 2.2 form IF
        else      name = "Vincia:QXsplitIF";
      }

      // 3. Gluon conversion
      // radiator id = emission id, radiator in initial state
      else if ( (event[iEm].id() == event[iRa].id()) && iRaIS ) {
        // 2.1 from II
        if (isII) name = "Vincia:GXconvII";
        // 2.2 form IF
        else      name = "Vincia:GXconvIF";
      }

      // 4. Final gluon splitting
      // emission id = - radiator id, radiator in final state
      else if ( (event[iEm].id() == -event[iRa].id()) && !iRaIS ) {
        name = "Vincia:XGsplitIF";
      }

      else {
        if (isII)       name = "Vincia:XXotherII";
        else if (iRaIS) name = "Vincia:XXotherIF";
        else            name = "Vincia:XXotherFI";
      }

    }

if (printNowAll) {
event.list();
cout << " SpaceAntShower::getSplittingName with size = " << event.size()
     << " iEm = " << iEm << " iRa = " << iRa << " iRe = " << iRe << endl;
if (name != "") cout << " Splitting name = " << name << endl;
else cout << " Did not find splitting" << endl;
cout << endl;
}

    vector<string> names;
    names.push_back(name);

    // Helicities
    if (helicityShower) {
      if (name == "Vincia:XGsplitIF" || name == "Vincia:GXconvII" ||
        name == "Vincia:GXconvIF") names.push_back(name);
      else if (name == "Vincia:GGemitII") {
        if ( event[iEm].pol() == event[iRa].pol() && 
          event[iEm].pol() == event[iRe].pol() ) {
          for (int i=0; i<2; i++) names.push_back(name);
        } else if ( event[iEm].pol() == event[iRa].pol() ||
          event[iEm].pol() == event[iRe].pol() ) names.push_back(name);
      } else if (name == "Vincia:GQemitII") {
        int iRaNow = ( idRa == 21 ? iRa : iRe );
        if (event[iEm].pol() == event[iRaNow].pol()) names.push_back(name);
      } else if (name == "Vincia:GGemitIF" || name == "Vincia:GQemitIF") {
        int iRaNow = (event[iRa].isFinal() ? iRe : iRa);
        if (event[iEm].pol() == event[iRaNow].pol()) names.push_back(name);
      }
    }

    return names;
  }

  // Return the splitting probability.
  vector<double> SpaceAntShower::getSplittingProbAll( const Event& event,
    int iRa, int iEm, int iRe, string name ) {
    if (!isSpacelike(event, iRa, iEm, iRe, name)) return vector<double>();
    string vincia("Vincia:");
    if (name.find(vincia) == string::npos) return vector<double>();

if (printNow)
cout << " SpaceAntShower::getSplittingProb with size = " << event.size()
     << " iEm = " << iEm << " iRa = " << iRa << " iRe = " << iRe << endl;

    vector<double> splitProbs;

    bool isII = ( !event[iRa].isFinal() && !event[iRe].isFinal() );

    // Find the right antenna for II
    if (isII) { for (int iant = iQQemitII; iant <= iGXconvII; iant++) {
      if ( !caseInsensComp(name, getAnt(iant)->vinciaName()) ) continue;

      // Swap iRa and iRe for GQemitII if Re is gluon
      //                  for iQXsplitII if Em is not cc to Ra
      //                  for iGXconvII if id(Em) is not id(Ra)
      int colEm = (event[iEm].id() > 0 ? event[iEm].col() : event[iEm].acol());
      int colRa = (event[iEm].id() > 0 ? event[iRa].col() : event[iRa].acol());
      if ( (iant == iGQemitII  && event[iRe].id() == 21) ||
           (iant == iQXsplitII && colEm != colRa) ||
           (iant == iGXconvII  && event[iEm].id() != event[iRa].id()) ) {
        int tmp = iRa;
        iRa = iRe;
        iRe = tmp;
      }

      double sab = (event[iRa].p()+event[iRe].p()).m2Calc();
      double saj = (event[iRa].p()+event[iEm].p()).m2Calc();
      double sjb = (event[iEm].p()+event[iRe].p()).m2Calc();
      double sAB = (sab - saj - sjb);
      if (sAB < 0.0) return vector<double>();

      int polA = event[iRa].pol(), polB = event[iRe].pol();
      int pola = polA, polb = polB, polj = event[iEm].pol();

      // Antenna function and colour factor
      vector<double> ants;
      // Helicities
      if (helicityShower) {
        // Gluon emission: quarks inherit helicity, gluons can be both if same
        // helicities after
        if (event[iEm].id() == 21 && iant != iQQemitII) {
          if (iant == iGGemitII && pola == polj && polb == polj) {
            ants.push_back( getAnt(iant)->antennaFunction(saj, sjb, sAB,  polA,
               polB, pola, polj, polb) );
            ants.push_back( getAnt(iant)->antennaFunction(saj, sjb, sAB, -polA,
               polB, pola, polj, polb) );
            ants.push_back( getAnt(iant)->antennaFunction(saj, sjb, sAB,  polA,
              -polB, pola, polj, polb) );
          } else if (event[iRa].id() == 21 && pola == polj) {
            ants.push_back( getAnt(iant)->antennaFunction(saj, sjb, sAB,  1,
              polB, pola, polj, polb) );
            ants.push_back( getAnt(iant)->antennaFunction(saj, sjb, sAB, -1,
              polB, pola, polj, polb) );
          } else if (event[iRe].id() == 21 && polb == polj) {
            ants.push_back( getAnt(iant)->antennaFunction(saj, sjb, sAB, polA,
               1, pola, polj, polb) );
            ants.push_back( getAnt(iant)->antennaFunction(saj, sjb, sAB, polA,
              -1, pola, polj, polb) );
          } else {
            ants.push_back( getAnt(iant)->antennaFunction(saj, sjb, sAB, polA,
              polB, pola, polj, polb) );
          }
        }
        // Gluon backwards evolving to a quark: recoiler inherits helicity, 
        // gluon can be both
        else if (iant == iGXconvII) {
          ants.push_back( getAnt(iant)->antennaFunction(saj, sjb, sAB,  1,
            polB, pola, polj, polb) );
          ants.push_back( getAnt(iant)->antennaFunction(saj, sjb, sAB, -1,
            polB, pola, polj, polb) );
        }
        // Quark backwards evolving to gluon: recoiler inherits helicity,
        // quark has -pol of emission antiquark
        else if (iant == iQXsplitII) {
          ants.push_back( getAnt(iant)->antennaFunction(saj, sjb, sAB, -polj,
            polB, pola, polj, polb) );
        }
        // All other cases: parents inherit helicities
        else {
          ants.push_back( getAnt(iant)->antennaFunction(saj, sjb, sAB, polA,
            polB, pola, polj, polb) );
        }
      } else ants.push_back( getAnt(iant)->antennaFunction(saj, sjb, sAB) );
      double colFac = getAnt(iant)->getChargeFac();

      // Phase Space factor, not inclued for Vincia
      double PSfac = sAB/sab/sab;
      PSfac = 1.0;

if (printNowAll) cout << " col*ant = " << colFac*ants[0] << endl << " PSfac   = "
<< PSfac << endl;

      for (int i=0; i<(int)ants.size(); i++) 
        splitProbs.push_back(PSfac*colFac*ants[i]);
      break;
    } }

    // Find the right antenna for IF
    else { for (int iant = iQQemitIF; iant <= iXGsplitIF; iant++) {
      // Find the right antenna
      if (name != getAnt(iant)->vinciaName()) continue;

      // Swap iRa and iRe if iRe is initial
      if (!event[iRe].isFinal()) {
        int tmp = iRa;
        iRa = iRe;
        iRe = tmp;
      }

      double sak = (event[iRa].p()+event[iRe].p()).m2Calc();
      double saj = (event[iRa].p()+event[iEm].p()).m2Calc();
      double sjk = (event[iEm].p()+event[iRe].p()).m2Calc();
      double sAK = (sak + saj - sjk);
      if (sAK < 0.0) return vector<double>();

      int polA = event[iRa].pol(), polK = event[iRe].pol();
      int pola = polA, polk = polK, polj = event[iEm].pol();

      // Antenna function and colour factor
      vector<double> ants;
      // Helicities
      if (helicityShower) {
        // Gluon emission: final state parent and initial state quark inherit
        // helicity, initial state gluon can be both if same helicities after
        if ( (iant == iGQemitIF || iant == iGGemitIF) && (pola == polj) ) {
          ants.push_back( getAnt(iant)->antennaFunction(saj, sjk, sAK,  1,
            polK, pola, polj, polk) );
          ants.push_back( getAnt(iant)->antennaFunction(saj, sjk, sAK, -1,
            polK, pola, polj, polk) );
        }
        // Gluon backwards evolving to a quark: recoiler inherits helicity, 
        // gluon can be both
        else if (iant == iGXconvIF) {
          ants.push_back( getAnt(iant)->antennaFunction(saj, sjk, sAK,  1,
            polK, pola, polj, polk) );
          ants.push_back( getAnt(iant)->antennaFunction(saj, sjk, sAK, -1,
            polK, pola, polj, polk) );
        }
        // Final state gluon splitting: recoiler inherits helicity, gluon can
        // be both
        else if (iant == iXGsplitIF) {
          ants.push_back( getAnt(iant)->antennaFunction(saj, sjk, sAK, polA,
             1, pola, polj, polk) );
          ants.push_back( getAnt(iant)->antennaFunction(saj, sjk, sAK, polA,
            -1, pola, polj, polk) );
        }
        // Quark backwards evolving to gluon: recoiler inherits helicity,
        // quark has -pol of emission antiquark
        else if (iant == iQXsplitIF) {
          ants.push_back( getAnt(iant)->antennaFunction(saj, sjk, sAK, -polj,
            polK, pola, polj, polk) );
        }
        // All other cases: parents inherit helicities
        else {
          ants.push_back( getAnt(iant)->antennaFunction(saj, sjk, sAK, polA,
            polK, pola, polj, polk) );
        }
      } else ants.push_back( getAnt(iant)->antennaFunction(saj, sjk, sAK) );
      double colFac = getAnt(iant)->getChargeFac();

      // Phase Space factor, not inclued for Vincia
      double PSfac = sAK/(sAK+sjk);
      PSfac = 1.0;

if (printNowAll) cout << " col*ant = " << colFac*ants[0] << endl << " PSfac   = "
<< PSfac << endl;

      for (int i=0; i<(int)ants.size(); i++) 
        splitProbs.push_back(PSfac*colFac*ants[i]);
      break;
    } }

if (printNowAll) event.list();
if (printNow)
cout << " Splitting probability for " << name << " = "
     << scientific << splitProbs[0] << endl << endl;

    return splitProbs;
  }

  // Return matrix-element correction.
  double SpaceAntShower::getMEC( const Event& event, int nStep) {
    matrixElementCorrectionsPtr->prepMECs(event, incomingType);
    return matrixElementCorrectionsPtr->performMECs(event, FSRshowerPtr, this,
      nStep);
  }

  // Return scales, alphaS values, etc in case of only unordered shower paths.
  map<string, double> SpaceAntShower::getUnorderedInfos( const Event& event) {
    matrixElementCorrectionsPtr->prepMECs(event, incomingType);
    return matrixElementCorrectionsPtr->getUnorderedInfos(event, FSRshowerPtr, 
      this);
  }

  // Return Born probability (= ME)
  double SpaceAntShower::getBornProb( const Event& event) {
    matrixElementCorrectionsPtr->prepMECs(event, incomingType);
    Event evtCopy = event;
    vector<Particle> eventParts = convertEventToParticles(evtCopy,
      matrixElementCorrectionsPtr->getProcClass());
    double ME = matrixElementCorrectionsPtr->evalME(eventParts);
    if (ME < 0.0) return 1.0;
    return ME;
  }

  // Return the Born scale.
  double SpaceAntShower::getBornScale( const Event& event) {
    double scale = -1.0;
    if ( event[1].isLepton() && event[2].isLepton() ) {
      // sHat for LEP-like process
      scale = m2(event[3], event[4]);
if (printNow) cout << " LL: sHat = " << sqrt(scale) << endl;
    } else if ( event[1].isHadron() && event[2].isHadron() ) {
      vector<int> iFS; iFS.clear(); iFS.resize(0);
      for (int i=0; i<(int)event.size(); i++)
        if (event[i].isFinal()) iFS.push_back(i);
      bool is2to1 = (iFS.size() == 1);
      if (!is2to1 && (iFS.size() == 2))
        is2to1 = (event[iFS[0]].isLepton() && event[iFS[1]].isLepton());
      if (is2to1) {
        // sHat for LHC-like 2 -> 1 processes
        scale = m2(event[3], event[4]);
if (printNow) cout << " HH: sHat = " << sqrt(scale) << endl;
      } else if (iFS.size() == 2) {
        // Geometric mean of outgoing mT2's for LHC-like 2 -> 2 processes
        scale = sqrt(event[iFS[0]].mT2()*event[iFS[1]].mT2());
if (printNow) cout << " HH: <mT2's> = " << sqrt(scale) << endl;
      } else if (iFS.size() == 4) {
        if (event[iFS[0]].isLepton() && event[iFS[1]].isLepton() &&
          event[iFS[2]].isLepton() && event[iFS[3]].isLepton()) {
          int mot1 = 0, mot2 = 0;
          for (int i=0; i<(int)iFS.size(); i++) 
            if (event[iFS[i]].mother1() > 0) {
                   if (mot1 == 0) mot1 = event[iFS[i]].mother1();
              else if (mot2 == 0) mot1 = event[iFS[i]].mother1();
            }
          // Geometric mean of outgoing mT2's for LHC-like 2 -> 2 processes
          if ( (mot1 > 0) && (mot2 > 0) && (mot1 != mot2) )
            scale = sqrt(event[mot1].mT2()*event[mot2].mT2());
if (printNow) cout << " HH: <mT2's> = " << sqrt(scale) << endl;
        } else {
          // Not found, report
          infoPtr->errorMsg("Warning in SpaceAntShower::getBornScale: unknown "
            "process, no scale found.");
        }
      } else {
        // Not found, report
        infoPtr->errorMsg("Warning in SpaceAntShower::getBornScale: unknown "
          "process, no scale found.");
      }
    } else {
      // Not found, report
      infoPtr->errorMsg("Warning in SpaceAntShower::getBornScale: unknown "
        "process, no scale found.");
    }
    return scale;
  }

  // Return the shower starting scale for pure shower.
  double SpaceAntShower::getShowerStartingScale( const Event& event) {
    return vinComPtr->getShowerStartingScale(event, sbbSav);
  }

} // end namespace VINCIA
