// TrialGeneratorsFSR.cc is a part of the VINCIA plug-in to the PYTHIA 8 
// generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2019 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for the 
// TrialGeneratorFSR base class and derived classes

// Include the Vincia headers 
#include "Vincia/TrialGeneratorsFSR.h"

namespace Vincia {
  
  // *****
  // Base class TrialGeneratorFSR
  // Note: base class is coded for a soft-eikonal trial function

  // Init Pointers
  void TrialGeneratorFSR::initPtr(Rndm* rndmPtrIn, Settings* settingsPtrIn) {
    rndmPtr     = rndmPtrIn;
    settingsPtr = settingsPtrIn;
    isInitPtr   = true;
  }

  // -----------------------------------------------------------------
  // Virtual functions 
  // Defining the particular kind of trial generator
  // -----------------------------------------------------------------

  // Initialize
  void TrialGeneratorFSR::init(int evTypeIn, double evNormFacIn) {
    // Check if pointers are initialized
    if ( !isInitPtr ) {
      printOut(name()+"::init()","Error! Pointers not initialized.");
      return;
    }
    // Set verbose level
    verbose          = settingsPtr->mode("Vincia:Verbose");
    // Set evolution variable type and (optionally) normalization
    evTypeSav        = evTypeIn;
    if ( evTypeSav <= 0 || evTypeSav >= 4) {
      printOut(name()+"::init()","Error! Unrecognized evolution variable "
               +num2str(evTypeSav,2)+", reverting to pT-evolution");
      evTypeSav      = 1;
      evNormFacIn = 1.0;
    }
    evNormFacSav  = evNormFacIn;
    // All clear
    isInit           = true;
  }
  
  // The trial antenna function, optionally in dimensionful form
  // Note: does not include color and headroom factors 
  double TrialGeneratorFSR::aTrial(double yij, double yjk, double sIK) {
    if (yij < 0. || yjk < 0.) return 0.;
    return 2./yij/yjk/sIK;
  }
  
  // Generate a new Q value, with first-order running alphaS
  double TrialGeneratorFSR::genQ2run(double q2old, double sIK, 
    double zMin, double zMax, double colFac,
    double b0, double kR, double Lambda, double headroomFac, double biasFac) {
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sIK < 0. || q2old < 0.) return 0.0;

    // Limit evolution scale by invariant mass of branching dipole
    double q2max  = sIK;
    if (evTypeSav == 1) q2max *= evNormFacSav/4.;
    else if (evTypeSav == 2) q2max *= evNormFacSav/2.;
    q2old = min(q2old,q2max);

    // Bias factors < 1 are treated by modyfing accept probability, with
    // trial probability left unchanged
    biasFac = max(biasFac,1.0);
    
    // Generate new trial scale
    double q2new  = 0.0;
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.*M_PI*b0/Iz/colFac/(headroomFac*biasFac);
    double ran    = rndmPtr->flat();
    double facLam = pow2(Lambda/kR) * evNormFacSav;
    double xNew   = 0.0;
    // pT evolution
    if ( evTypeSav == 1) {
      xNew = pow(ran,comFac) * log(q2old / facLam);
    } 
    // mD evolution: kernel multiplied by 2 to include both phase-space branches
    else if ( evTypeSav == 2 ) {
      xNew = pow(ran,comFac/4.) * log(q2old / facLam);
    } 
    // E* evolution not available with running inside
    else if ( evTypeSav == 3 ) {
      printOut(name()+"::genQ()","Error! Cannot use running trial alphaS"
               " with E*-ordering");
      return 0.0;
    }
    q2new = exp(xNew) * facLam;
    return q2new;
  }

  // Generate a new Q value, with constant trial alphaS
  double TrialGeneratorFSR::genQ2(double q2old, double sIK,
    double zMin, double zMax, double colFac, double alphaSvalue,
    double headroomFac, double biasFac) {
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sIK < 0. || q2old < 0.) return 0.0;

    // Limit evolution scale by invariant mass of branching dipole
    double q2max  = sIK;
    if (evTypeSav == 1) q2max *= 4./evNormFacSav;
    else if (evTypeSav == 2) q2max *= 2./evNormFacSav;
    q2old = min(q2old,q2max);

    // Bias factors < 1 are treated by modyfing accept probability, with
    // trial probability left unchanged
    biasFac = max(biasFac,1.0);

    // Generate new trial scale
    double q2new  = 0.0;
    double Iz     = getIz(zMin,zMax);
    double comFac = 4*M_PI/alphaSvalue/Iz/colFac/(headroomFac*biasFac);
    double ran    = rndmPtr->flat();
    // pT evolution
    if ( evTypeSav == 1 ) {
      q2new = q2old * pow(ran,comFac);
    } 
    // mD evolution: kernel multiplied by 2 to include both phase-space branches
    else if ( evTypeSav == 2) {
      q2new = q2old * pow(ran,comFac/4.);
    }
    // E* evolution
    else if ( evTypeSav == 3 ) {
      q2new = pow2( sqrt(q2old) - sqrt(sIK)*comFac*log(ran) );
    }
    return q2new;
  }

  // Generate a new zeta value in [zMin,zMax]
  double TrialGeneratorFSR::genZ(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return -1.;
    double ran  = rndmPtr->flat();
    // pT evolution
    if ( evTypeSav == 1 ) {
      double invZ = 1 + (1-zMin)/zMin*pow(zMin*(1-zMax)/zMax/(1-zMin),ran);
      return 1./invZ;
    } 
    // mD evolution
    else if ( evTypeSav == 2 ) {
      // Randomize branches for mD type evolution variable (zeta pos/neg)
      double sign = 1.0;
      ran *= 2;
      if (ran > 1.0) { 
        ran -= 1.0;
        sign = -1.0;
      }
      return sign * zMin * pow(zMax/zMin,ran);
    } 
    // E* evolution
    else if ( evTypeSav == 3 ) {
      return zMax*zMin/(zMax - ran*(zMax-zMin));
    }
    else return -1.;
  }
  
  // The zeta integral
  double TrialGeneratorFSR::getIz(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return 0.0;
    if ( evTypeSav == 1 ) return log(zMax*(1-zMin)/zMin/(1-zMax)); 
    else if ( evTypeSav == 2 ) return log(zMax/zMin);
    else if ( evTypeSav == 3 ) return 1./zMin - 1./zMax;
    else return 0.0;
  }
  
  // The zeta boundaries, for a given value of the evolution variable
  // Note: for Type 2, zeta boundaries are for the branch yij < yjk
  // Note: for Type 3, zeta boundaries are nominally [0,1] (returned here)
  //       and should instead be bounded by the value used for the 
  //       hadronization cutoff and/or evolution window cutoffs
  double TrialGeneratorFSR::getZmin(double yE) {
    if ( evTypeSav == 1 ) return 0.5*(1-sqrt(1.-4.*yE/evNormFacSav));
    else if ( evTypeSav == 2 ) return yE/evNormFacSav;
    else if ( evTypeSav == 3 ) return 0.;
    return 0.;
  }
  double TrialGeneratorFSR::getZmax(double yE) {
    if ( evTypeSav == 1 ) return 0.5*(1+sqrt(1.-4.*yE/evNormFacSav));
    else if ( evTypeSav == 2 ) return 1.-yE/evNormFacSav;
    else if ( evTypeSav == 3 ) return 1.;
    return 0.;
  }
  
  // Inverse transforms
  // To obtain yij and yjk from yE and zeta
  double TrialGeneratorFSR::getYij(double yE, double zeta) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getYjk(yE,-zeta);
    // Sanity check
    if (yE < 0. || (1-zeta) <= 0. || zeta <= 0.) {
      printOut(name()+"::getYij()","Error! Out of range");
      return 0.;
    }
    // Return yij, with zeta definitions as given in the compendium
    double yij = 0.;
    if ( evTypeSav == 1 ) {
      yij = sqrt(zeta/(1-zeta) * yE/evNormFacSav);
    } else if ( evTypeSav == 2 ) {
      yij = yE/evNormFacSav;
    } else if ( evTypeSav == 3 ) {
      yij = zeta * sqrt(yE);
    }
    return yij;
  }
  double TrialGeneratorFSR::getYjk(double yE, double zeta) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getYij(yE,-zeta);
    // Sanity check
    if (yE < 0. || (1-zeta) <= 0. || zeta <= 0.) {
      printOut(name()+"::getYjk()","Error! Out of range");
      return 0.;
    }
    // Return yjk, with zeta definitions as given in the compendium
    double yjk = 0.;
    if ( evTypeSav == 1 ) {
      yjk = sqrt((1-zeta)/zeta * yE/evNormFacSav);
    } else if ( evTypeSav == 2 ) {
      yjk = zeta;
    } else if ( evTypeSav == 3 ) {
      yjk = (1-zeta) * sqrt(yE);
    }
    return yjk;
  }
  
  // -----------------------------------------------------------------
  // Non-Virtual TrialGeneratorFSR base-class functions
  // -----------------------------------------------------------------
  
  bool TrialGeneratorFSR::checkInit() {
    if (!isInit || !isInitPtr) {
      printOut(name()+"::checkInit()","Error! Not initialized");
      return false;        
    } else {
        return true;        
    }
  }

  // -----------------------------------------------------------------
  // TrialCollI derived TrialGeneratorFSR Class
  // (for additional sector-collinear term for parton I)
  // -----------------------------------------------------------------
  
  // The trial antenna function, optionally in dimensionful form
  // Note: does not include color and headroom factors 
  double TrialCollI::aTrial(double yij, double yjk, double sIK) {
    if (yij < 0. || yjk < 0.) return 0.;
    return 2./yij/(1.-yjk)/sIK;
  }
  
  // Generate a new Q value, with first-order running alphaS
  double TrialCollI::genQ2run(double q2old, double sIK,
    double zMin, double zMax, double colFac, double b0, double kR,
    double Lambda, double headroomFac, double biasFac) {
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sIK < 0. || q2old < 0.) return 0.0;

    // Limit evolution scale by invariant mass of branching dipole
    double q2max  = sIK;
    if (evTypeSav == 1) q2max *= evNormFacSav/4.;
    else if (evTypeSav == 2) q2max *= evNormFacSav/2.;
    q2old = min(q2old,q2max);

    // Bias factors < 1 are treated by modyfing accept probability, with
    // trial probability left unchanged
    biasFac = max(biasFac,1.0);

    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4*M_PI*b0/Iz/colFac/(headroomFac*biasFac);
    double ran    = rndmPtr->flat();
    double facLam = evNormFacSav * pow2(Lambda/kR);
    double xNew   = 0.0;
    if ( evTypeSav == 1 ) 
      xNew = pow(ran,comFac/2.) * log(q2old / facLam);
    else if ( evTypeSav == 2 ) 
      xNew = pow(ran,comFac/2.) * log(q2old / facLam);
    return exp(xNew) * facLam;

  }

  // Generate a new Q value, with constant trial alphaS
  double TrialCollI::genQ2(double q2old, double sIK, double zMin, double zMax, 
    double colFac, double alphaSvalue, double headroomFac, double biasFac) {
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sIK < 0. || q2old < 0.) return 0.0;

    // Limit evolution scale by invariant mass of branching dipole
    double q2max  = sIK;
    if (evTypeSav == 1) q2max *= evNormFacSav/4.;
    else if (evTypeSav == 2) q2max *= evNormFacSav/2.;
    q2old = min(q2old,q2max);

    // Bias factors < 1 are treated by modyfing accept probability, with
    // trial probability left unchanged
    biasFac = max(biasFac,1.0);

    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4*M_PI/alphaSvalue/Iz/colFac/(headroomFac*biasFac);
    double ran    = rndmPtr->flat();
    if (evTypeSav == 1) 
      return q2old * pow(ran, comFac/2.);
    else if (evTypeSav == 2) 
      return q2old * pow(ran, comFac/2.);
    return 0.;

  }

  // Generate a new zeta value in [zMin,zMax]
  double TrialCollI::genZ(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return -1.;
    double ran   = rndmPtr->flat();
    double xMin = 1. - zMin;
    double xMax = 1. - zMax;
    if (evTypeSav == 1 || evTypeSav == 2) 
      return 1. - xMin*pow(xMax/xMin,ran);
    else
      return 0.;
  }
  
  // The zeta integral
  double TrialCollI::getIz(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return 0.0;
    double xMin = 1. - zMin;
    double xMax = 1. - zMax;
    if (evTypeSav == 1 || evTypeSav == 2) 
      return log(xMin/xMax);
    else 
      return 0.;
  }
  
  // The zeta boundaries, for a given value of the evolution variable
  double TrialCollI::getZmin(double yE) {
    if (evTypeSav == 1) {
      return 0.5 * (1. - sqrt(1. - 4*yE/evNormFacSav));
    } else if (evTypeSav == 2) {
      return yE/evNormFacSav;
    } else {
      return 0.;
    }
  }
  double TrialCollI::getZmax(double yE) {
    if (evTypeSav == 1) {
      return 0.5 * (1. + sqrt(1. - 4*yE/evNormFacSav));
    } else if (evTypeSav == 2) {
      return 1.-yE/evNormFacSav;
    } else {
      return 0.;
    }
  }
  
  // Inverse transforms
  // To obtain yij and yjk from yE and zeta
  double TrialCollI::getYij(double yE, double zeta) {
    // Sanity check
    if (yE < 0. || (1-zeta) <= 0. || zeta <= 0.) {
      printOut(name()+"::getYij()","Error! Out of range");
      return 0.;
    }
    if (evTypeSav == 1) {
      return 1./zeta * yE/evNormFacSav;
    } else if (evTypeSav == 2) {
      return yE / evNormFacSav;
    }
    else return 0.;
  }
  double TrialCollI::getYjk(double yE, double zeta) {
    // Sanity check
    if (yE < 0. || (1-zeta) <= 0. || zeta <= 0.) {
      printOut(name()+"::getYjk()","Error! Out of range");
      return 0.;
    }
    return zeta;
  }
  
  // -----------------------------------------------------------------
  // TrialSplitK derived TrialGeneratorFSR Class
  // Note: always uses mjk^2 as evolution variable and yij as zeta
  // -----------------------------------------------------------------
  
  // Trial function
  double TrialSplitK::aTrial(double yij, double yjk, double sIK) {
    if (yij < 0. || yjk < 0.) return 0.;
    return 1./yjk/sIK;
  }
  
  // Generate a new Q value, with first-order running alphaS
  double TrialSplitK::genQ2run(double q2old, double sIK,
    double zMin, double zMax, double colFac,
    double b0, double kR, double Lambda, double headroomFac, double biasFac) {
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sIK < 0. || q2old < 0.) return 0.0;
    
    // Limit evolution scale by invariant mass of branching dipole
    double q2max  = sIK;
    q2old = min(q2old,q2max);

    // Bias factors < 1 are treated by modyfing accept probability, with
    // trial probability left unchanged
    biasFac = max(biasFac,1.0);

    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4*M_PI*b0/Iz/colFac/(headroomFac*biasFac);
    double ran    = rndmPtr->flat();
    double facLam = pow2(Lambda/kR);
    double xNew   = pow(ran,comFac) * log(q2old / facLam);
    return exp(xNew) * facLam;
  }

  // Generate a new Q value, with constant trial alphaS
  double TrialSplitK::genQ2(double q2old, double sIK, double zMin, double zMax, 
    double colFac, double alphaSvalue, double headroomFac, double biasFac) {
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sIK < 0. || q2old < 0.) return 0.0;

    // Limit evolution scale by invariant mass of branching dipole
    double q2max  = sIK;
    q2old = min(q2old,q2max);

    // Bias factors < 1 are treated by modyfing accept probability, with
    // trial probability left unchanged
    biasFac = max(biasFac,1.0);

    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4*M_PI/alphaSvalue/Iz/colFac/(headroomFac*biasFac);
    double ran    = rndmPtr->flat();
    return q2old * pow(ran,comFac);
  }

  // Generate a new zeta value in [zMin,zMax]
  double TrialSplitK::genZ(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return -1.;
    double ran = rndmPtr->flat();
    return zMin + ran*(zMax - zMin);
  }
  
  // The zeta integral
  double TrialSplitK::getIz(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return 0.0;
    return zMax - zMin;
  }
  
  // The zeta boundaries, for a given value of the evolution variable
  double TrialSplitK::getZmin(double) {
    return 0.;
  }
  double TrialSplitK::getZmax(double yE) {
    return 1. - yE;
  }
  
  // Inverse transforms
  // To obtain yij and yjk from yE and zeta
  double TrialSplitK::getYij(double yE, double zeta) {
    // Sanity check
    if (yE < 0. || (1-zeta) <= 0. || zeta <= 0.) {
      printOut(name()+"::getYij()","Error! Out of range");
      return 0.;
    }
    return zeta;
  }
  double TrialSplitK::getYjk(double yE, double zeta) {
    // Sanity check
    if (yE < 0. || (1-zeta) <= 0. || zeta <= 0.) {
      printOut(name()+"::getYjk()","Error! Out of range");
      return 0.;
    }
    return yE;
  }
  
}
