// VinciaMHV.h is a part of the VINCIA plug-in to the
// PYTHIA 8 generator.
// Copyright (C) 2019 P. Skands.
// VINCIA is licenced under the GNU GPL v2 or later.
// Please respect the MCnet Guidelines for Event Generator Authors and Users,
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for VICNIA's MHV amplitudes' classes
// Added by A Lifson. Includes QCD amplitudes for up to 6 gluons; 2 quarks and
// up to 5 gluons and 4 quarks with up to 1 gluon. Also contains Drell-Yan with
// up to 5 gluons.

#ifndef VINCIAMHV_H
#define VINCIAMHV_H

#include "Pythia8/ParticleData.h"
#include "Pythia8/Settings.h"
#include "Pythia8/StandardModel.h"
#include "Pythia8/PythiaComplex.h"
#include "Vincia/VinciaCommon.h"

namespace Vincia {

// Declare complex error variable pie
const Pythia8::complex pie(-M_PI, exp(1));

// For ease of writing, put FourPi as variable
const double FourPi = 4.*M_PI;

// Declare maximum number of partons
const int nPartonsMax = 9;

// Declare maximum number of partons to be polarised
const int nPartonsToPolMax = 5;

// Declare maximum number of amplitudes in vector of amplitude permutations
const int nAmpsMax = 720; //  = 6!, i.e. 7 gluons, or 2 quarks 6 gluons

//Declare values of leading colour factors
const double LCAG4 = 152./3;
const double LCAG5 = 3640./27;
const double LCAG6 = 29128./81;
const double LCAG7 = 77672./81;
const double LC2Q0 = 3.;
const double LC2Q1 = 8.;
const double LC2Q4 = 64./3.;
const double LC2Q5 = 512./9.;
const double LC2Q6 = 4096./27.0;
const double LC2Q7 = 32768./81.;
const double LC4Q  = 8.;
const double LC4Q1 = 64./3;
const double LC4Q2 = -1;

// Declare values of full colour factors
const double FC4Q  = 8.;
const double FCAG4 = 72.;
const double FCAG5 = 216.;
const double FCAG6 = 648.;

// Declare a complex one value
const Pythia8::complex one(1., 0.);

// Declare a complex zero
const Pythia8::complex zero(0.,0.);

// Declare colour matrices. Order is 12...n gluons permuted lexicographically
const double qQBar2g [2][2] = {
                               {64./3., -8./3.},
                               {-8./3., 64./3.},
};

const double qQBar3g [6][6] =
  {
   {512./9., -64./9., -64./9., 8./9., 8./9., 80./9.}, //123
   {-64./9., 512./9., 8./9., 80./9., -64./9., 8./9.}, //132
   {-64./9., 8./9., 512./9., -64./9., 80./9., 8./9.}, //213
   {8./9., 80./9., -64./9., 512./9., 8./9., -64./9.}, //231
   {8./9., -64./9., 80./9., 8./9., 512./9., -64./9.}, //312
   {80./9., 8./9., 8./9., -64./9., -64./9., 512./9.}, //321
  };

const double qQBar4g [24][24] =
  {
   // Row     0 : MG ordering  = [ 3, 1, 2, 4, 5, 6 ]
   { 4096/27.,   -512/27.,   -512/27.,     64/27.,     64/27.,    640/27.,
     -512/27.,     64/27.,     64/27.,     -8/27.,     -8/27.,    -80/27.,
     64/27.,     -8/27.,    640/27.,    -80/27.,    568/27.,    496/27.,
     -8/27.,    -80/27.,    -80/27.,    496/27.,    496/27.,   -224/27.},
   // Row     1 : MG ordering  = [ 3, 1, 2, 4, 6, 5 ]
   {  -512/27.,   4096/27.,     64/27.,    640/27.,   -512/27.,     64/27.,
      64/27.,   -512/27.,     -8/27.,    -80/27.,     64/27.,     -8/27.,
      -8/27.,    -80/27.,    -80/27.,    496/27.,    496/27.,   -224/27.,
      64/27.,     -8/27.,    640/27.,    -80/27.,    568/27.,    496/27.},
   // Row     2 : MG ordering  = [ 3, 1, 2, 5, 4, 6 ]
   {  -512/27.,     64/27.,   4096/27.,   -512/27.,    640/27.,     64/27.,
      64/27.,     -8/27.,    640/27.,    -80/27.,    568/27.,    496/27.,
      -512/27.,     64/27.,     64/27.,     -8/27.,     -8/27.,    -80/27.,
      -80/27.,     -8/27.,    496/27.,   -224/27.,    -80/27.,    496/27.},
   // Row     3 : MG ordering  = [ 3, 1, 2, 5, 6, 4 ]
   {    64/27.,    640/27.,   -512/27.,   4096/27.,     64/27.,   -512/27.,
        -8/27.,    -80/27.,    -80/27.,    496/27.,    496/27.,   -224/27.,
        64/27.,   -512/27.,     -8/27.,    -80/27.,     64/27.,     -8/27.,
        -8/27.,     64/27.,    568/27.,    496/27.,    640/27.,    -80/27.},
   // Row     4 : MG ordering  = [ 3, 1, 2, 6, 4, 5 ]
   {    64/27.,   -512/27.,    640/27.,     64/27.,   4096/27.,   -512/27.,
        -8/27.,     64/27.,    568/27.,    496/27.,    640/27.,    -80/27.,
        -80/27.,     -8/27.,    496/27.,   -224/27.,    -80/27.,    496/27.,
        -512/27.,     64/27.,     64/27.,     -8/27.,     -8/27.,    -80/27.},
   // Row     5 : MG ordering  = [ 3, 1, 2, 6, 5, 4 ]
   {   640/27.,     64/27.,     64/27.,   -512/27.,   -512/27.,   4096/27.,
       -80/27.,     -8/27.,    496/27.,   -224/27.,    -80/27.,    496/27.,
       -8/27.,     64/27.,    568/27.,    496/27.,    640/27.,    -80/27.,
       64/27.,   -512/27.,     -8/27.,    -80/27.,     64/27.,     -8/27.},
   // Row     6 : MG ordering  = [ 3, 1, 4, 2, 5, 6 ]
   {   -512/27.,     64/27.,     64/27.,     -8/27.,     -8/27.,    -80/27.,
       4096/27.,   -512/27.,   -512/27.,     64/27.,     64/27.,    640/27.,
       640/27.,    -80/27.,     64/27.,     -8/27.,    496/27.,    568/27.,
       -80/27.,    496/27.,     -8/27.,    -80/27.,   -224/27.,    496/27.},
   // Row     7 : MG ordering  = [ 3, 1, 4, 2, 6, 5 ]
   {     64/27.,   -512/27.,     -8/27.,    -80/27.,     64/27.,     -8/27.,
         -512/27.,   4096/27.,     64/27.,    640/27.,   -512/27.,     64/27.,
         -80/27.,    496/27.,     -8/27.,    -80/27.,   -224/27.,    496/27.,
         640/27.,    -80/27.,     64/27.,     -8/27.,    496/27.,    568/27.},
   // Row     8 : MG ordering  = [ 3, 1, 4, 5, 2, 6 ]
   {     64/27.,     -8/27.,    640/27.,    -80/27.,    568/27.,    496/27.,
         -512/27.,     64/27.,   4096/27.,   -512/27.,    640/27.,     64/27.,
         64/27.,     -8/27.,   -512/27.,     64/27.,    -80/27.,     -8/27.,
         496/27.,   -224/27.,    -80/27.,     -8/27.,    496/27.,    -80/27.},
   // Row     9 : MG ordering  = [ 3, 1, 4, 5, 6, 2 ]
   {     -8/27.,    -80/27.,    -80/27.,    496/27.,    496/27.,   -224/27.,
         64/27.,    640/27.,   -512/27.,   4096/27.,     64/27.,   -512/27.,
         -8/27.,    -80/27.,     64/27.,   -512/27.,     -8/27.,     64/27.,
         568/27.,    496/27.,     -8/27.,     64/27.,    -80/27.,    640/27.},
   // Row    10 : MG ordering  = [ 3, 1, 4, 6, 2, 5 ]
   {     -8/27.,     64/27.,    568/27.,    496/27.,    640/27.,    -80/27.,
         64/27.,   -512/27.,    640/27.,     64/27.,   4096/27.,   -512/27.,
         496/27.,   -224/27.,    -80/27.,     -8/27.,    496/27.,    -80/27.,
         64/27.,     -8/27.,   -512/27.,     64/27.,    -80/27.,     -8/27.},
   // Row    11 : MG ordering  = [ 3, 1, 4, 6, 5, 2 ]
   {    -80/27.,     -8/27.,    496/27.,   -224/27.,    -80/27.,    496/27.,
        640/27.,     64/27.,     64/27.,   -512/27.,   -512/27.,   4096/27.,
        568/27.,    496/27.,     -8/27.,     64/27.,    -80/27.,    640/27.,
        -8/27.,    -80/27.,     64/27.,   -512/27.,     -8/27.,     64/27.},
   // Row    12 : MG ordering  = [ 3, 1, 5, 2, 4, 6 ]
   {     64/27.,     -8/27.,   -512/27.,     64/27.,    -80/27.,     -8/27.,
         640/27.,    -80/27.,     64/27.,     -8/27.,    496/27.,    568/27.,
         4096/27.,   -512/27.,   -512/27.,     64/27.,     64/27.,    640/27.,
         496/27.,    -80/27.,   -224/27.,    496/27.,     -8/27.,    -80/27.},
   // Row    13 : MG ordering  = [ 3, 1, 5, 2, 6, 4 ]
   {     -8/27.,    -80/27.,     64/27.,   -512/27.,     -8/27.,     64/27.,
         -80/27.,    496/27.,     -8/27.,    -80/27.,   -224/27.,    496/27.,
         -512/27.,   4096/27.,     64/27.,    640/27.,   -512/27.,     64/27.,
         -80/27.,    640/27.,    496/27.,    568/27.,     64/27.,     -8/27.},
   // Row    14 : MG ordering  = [ 3, 1, 5, 4, 2, 6 ]
   {    640/27.,    -80/27.,     64/27.,     -8/27.,    496/27.,    568/27.,
        64/27.,     -8/27.,   -512/27.,     64/27.,    -80/27.,     -8/27.,
        -512/27.,     64/27.,   4096/27.,   -512/27.,    640/27.,     64/27.,
        -224/27.,    496/27.,    496/27.,    -80/27.,    -80/27.,     -8/27.},
   // Row    15 : MG ordering  = [ 3, 1, 5, 4, 6, 2 ]
   {    -80/27.,    496/27.,     -8/27.,    -80/27.,   -224/27.,    496/27.,
        -8/27.,    -80/27.,     64/27.,   -512/27.,     -8/27.,     64/27.,
        64/27.,    640/27.,   -512/27.,   4096/27.,     64/27.,   -512/27.,
        496/27.,    568/27.,    -80/27.,    640/27.,     -8/27.,     64/27.},
   // Row    16 : MG ordering  = [ 3, 1, 5, 6, 2, 4 ]
   {    568/27.,    496/27.,     -8/27.,     64/27.,    -80/27.,    640/27.,
        496/27.,   -224/27.,    -80/27.,     -8/27.,    496/27.,    -80/27.,
        64/27.,   -512/27.,    640/27.,     64/27.,   4096/27.,   -512/27.,
        -8/27.,     64/27.,    -80/27.,     -8/27.,   -512/27.,     64/27.},
   // Row    17 : MG ordering  = [ 3, 1, 5, 6, 4, 2 ]
   {    496/27.,   -224/27.,    -80/27.,     -8/27.,    496/27.,    -80/27.,
        568/27.,    496/27.,     -8/27.,     64/27.,    -80/27.,    640/27.,
        640/27.,     64/27.,     64/27.,   -512/27.,   -512/27.,   4096/27.,
        -80/27.,     -8/27.,     -8/27.,     64/27.,     64/27.,   -512/27.},
   // Row    18 : MG ordering  = [ 3, 1, 6, 2, 4, 5 ]
   {     -8/27.,     64/27.,    -80/27.,     -8/27.,   -512/27.,     64/27.,
         -80/27.,    640/27.,    496/27.,    568/27.,     64/27.,     -8/27.,
         496/27.,    -80/27.,   -224/27.,    496/27.,     -8/27.,    -80/27.,
         4096/27.,   -512/27.,   -512/27.,     64/27.,     64/27.,    640/27.},
   // Row    19 : MG ordering  = [ 3, 1, 6, 2, 5, 4 ]
   {    -80/27.,     -8/27.,     -8/27.,     64/27.,     64/27.,   -512/27.,
        496/27.,    -80/27.,   -224/27.,    496/27.,     -8/27.,    -80/27.,
        -80/27.,    640/27.,    496/27.,    568/27.,     64/27.,     -8/27.,
        -512/27.,   4096/27.,     64/27.,    640/27.,   -512/27.,     64/27.},
   // Row    20 : MG ordering  = [ 3, 1, 6, 4, 2, 5 ]
   {    -80/27.,    640/27.,    496/27.,    568/27.,     64/27.,     -8/27.,
        -8/27.,     64/27.,    -80/27.,     -8/27.,   -512/27.,     64/27.,
        -224/27.,    496/27.,    496/27.,    -80/27.,    -80/27.,     -8/27.,
        -512/27.,     64/27.,   4096/27.,   -512/27.,    640/27.,     64/27.},
   // Row    21 : MG ordering  = [ 3, 1, 6, 4, 5, 2 ]
   {    496/27.,    -80/27.,   -224/27.,    496/27.,     -8/27.,    -80/27.,
        -80/27.,     -8/27.,     -8/27.,     64/27.,     64/27.,   -512/27.,
        496/27.,    568/27.,    -80/27.,    640/27.,     -8/27.,     64/27.,
        64/27.,    640/27.,   -512/27.,   4096/27.,     64/27.,   -512/27.},
   // Row    22 : MG ordering  = [ 3, 1, 6, 5, 2, 4 ]
   {    496/27.,    568/27.,    -80/27.,    640/27.,     -8/27.,     64/27.,
        -224/27.,    496/27.,    496/27.,    -80/27.,    -80/27.,     -8/27.,
        -8/27.,     64/27.,    -80/27.,     -8/27.,   -512/27.,     64/27.,
        64/27.,   -512/27.,    640/27.,     64/27.,   4096/27.,   -512/27.},
   // Row    23 : MG ordering  = [ 3, 1, 6, 5, 4, 2 ]
   {   -224/27.,    496/27.,    496/27.,    -80/27.,    -80/27.,     -8/27.,
       496/27.,    568/27.,    -80/27.,    640/27.,     -8/27.,     64/27.,
       -80/27.,     -8/27.,     -8/27.,     64/27.,     64/27.,   -512/27.,
       640/27.,     64/27.,     64/27.,   -512/27.,   -512/27.,   4096/27.}
  };

const double qQBar5g [120][120] =
  {
   // Row     0 : MG ordering  = [ 3, 1, 2, 4, 5, 6, 7 ]
   {  32768/81.,  -4096/81.,  -4096/81.,    512/81.,    512/81.,   5120/81.,
      -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
      512/81.,    -64/81.,   5120/81.,   -640/81.,   4544/81.,   3968/81.,
      -64/81.,   -640/81.,   -640/81.,   3968/81.,   3968/81.,  -1792/81.,
      -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
      512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
      -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
      8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
      512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
      5120/81.,   -640/81.,   -640/81.,     80/81.,     80/81.,    800/81.,
      4544/81.,   -568/81.,   3968/81.,   -496/81.,   4040/81.,   4112/81.,
      -568/81.,    152/81.,   -496/81.,   -424/81.,  -1072/81.,   -352/81.,
      -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
      -640/81.,     80/81.,   3968/81.,   -496/81.,    152/81.,   -424/81.,
      3968/81.,   -496/81.,  -1792/81.,    224/81.,  -1072/81.,   -352/81.,
      4040/81.,  -1072/81.,  -1072/81.,   3536/81.,   3536/81.,   -928/81.,
      8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
      80/81.,    800/81.,   -496/81.,   4112/81.,   -424/81.,   -352/81.,
      -496/81.,   -424/81.,    224/81.,   -352/81.,   3536/81.,   -928/81.,
      4112/81.,   -352/81.,   -352/81.,   -928/81.,   -928/81.,   1088/81.},
   // Row     1 : MG ordering  = [ 3, 1, 2, 4, 5, 7, 6 ]
   {  -4096/81.,  32768/81.,    512/81.,   5120/81.,  -4096/81.,    512/81.,
      512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
      -64/81.,   -640/81.,   -640/81.,   3968/81.,   3968/81.,  -1792/81.,
      512/81.,    -64/81.,   5120/81.,   -640/81.,   4544/81.,   3968/81.,
      512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
      -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
      8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
      -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
      -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
      -640/81.,   5120/81.,     80/81.,    800/81.,   -640/81.,     80/81.,
      -568/81.,    152/81.,   -496/81.,   -424/81.,  -1072/81.,   -352/81.,
      4544/81.,   -568/81.,   3968/81.,   -496/81.,   4040/81.,   4112/81.,
      8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
      80/81.,    800/81.,   -496/81.,   4112/81.,   -424/81.,   -352/81.,
      -496/81.,   -424/81.,    224/81.,   -352/81.,   3536/81.,   -928/81.,
      4112/81.,   -352/81.,   -352/81.,   -928/81.,   -928/81.,   1088/81.,
      -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
      -640/81.,     80/81.,   3968/81.,   -496/81.,    152/81.,   -424/81.,
      3968/81.,   -496/81.,  -1792/81.,    224/81.,  -1072/81.,   -352/81.,
      4040/81.,  -1072/81.,  -1072/81.,   3536/81.,   3536/81.,   -928/81.},
   // Row     2 : MG ordering  = [ 3, 1, 2, 4, 6, 5, 7 ]
   {  -4096/81.,    512/81.,  32768/81.,  -4096/81.,   5120/81.,    512/81.,
      512/81.,    -64/81.,   5120/81.,   -640/81.,   4544/81.,   3968/81.,
      -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
      -640/81.,    -64/81.,   3968/81.,  -1792/81.,   -640/81.,   3968/81.,
      512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
      -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
      512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
      80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
      -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
      -640/81.,     80/81.,   3968/81.,   -496/81.,    152/81.,   -424/81.,
      3968/81.,   -496/81.,  -1792/81.,    224/81.,  -1072/81.,   -352/81.,
      4040/81.,  -1072/81.,  -1072/81.,   3536/81.,   3536/81.,   -928/81.,
      512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
      5120/81.,   -640/81.,   -640/81.,     80/81.,     80/81.,    800/81.,
      4544/81.,   -568/81.,   3968/81.,   -496/81.,   4040/81.,   4112/81.,
      -568/81.,    152/81.,   -496/81.,   -424/81.,  -1072/81.,   -352/81.,
      80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
      800/81.,     80/81.,   -424/81.,   -352/81.,   -496/81.,   4112/81.,
      4112/81.,   -352/81.,   -352/81.,   -928/81.,   -928/81.,   1088/81.,
      -496/81.,   -424/81.,    224/81.,   -352/81.,   3536/81.,   -928/81.},
   // Row     3 : MG ordering  = [ 3, 1, 2, 4, 6, 7, 5 ]
   {    512/81.,   5120/81.,  -4096/81.,  32768/81.,    512/81.,  -4096/81.,
        -64/81.,   -640/81.,   -640/81.,   3968/81.,   3968/81.,  -1792/81.,
        512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
        -64/81.,    512/81.,   4544/81.,   3968/81.,   5120/81.,   -640/81.,
        -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
        8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
        -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
        8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
        8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
        80/81.,    800/81.,   -496/81.,   4112/81.,   -424/81.,   -352/81.,
        -496/81.,   -424/81.,    224/81.,   -352/81.,   3536/81.,   -928/81.,
        4112/81.,   -352/81.,   -352/81.,   -928/81.,   -928/81.,   1088/81.,
        -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
        -640/81.,   5120/81.,     80/81.,    800/81.,   -640/81.,     80/81.,
        -568/81.,    152/81.,   -496/81.,   -424/81.,  -1072/81.,   -352/81.,
        4544/81.,   -568/81.,   3968/81.,   -496/81.,   4040/81.,   4112/81.,
        8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
        80/81.,   -640/81.,    152/81.,   -424/81.,   3968/81.,   -496/81.,
        4040/81.,  -1072/81.,  -1072/81.,   3536/81.,   3536/81.,   -928/81.,
        3968/81.,   -496/81.,  -1792/81.,    224/81.,  -1072/81.,   -352/81.},
   // Row     4 : MG ordering  = [ 3, 1, 2, 4, 7, 5, 6 ]
   {    512/81.,  -4096/81.,   5120/81.,    512/81.,  32768/81.,  -4096/81.,
        -64/81.,    512/81.,   4544/81.,   3968/81.,   5120/81.,   -640/81.,
        -640/81.,    -64/81.,   3968/81.,  -1792/81.,   -640/81.,   3968/81.,
        -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
        -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
        8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
        80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
        512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
        8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
        80/81.,   -640/81.,    152/81.,   -424/81.,   3968/81.,   -496/81.,
        4040/81.,  -1072/81.,  -1072/81.,   3536/81.,   3536/81.,   -928/81.,
        3968/81.,   -496/81.,  -1792/81.,    224/81.,  -1072/81.,   -352/81.,
        80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
        800/81.,     80/81.,   -424/81.,   -352/81.,   -496/81.,   4112/81.,
        4112/81.,   -352/81.,   -352/81.,   -928/81.,   -928/81.,   1088/81.,
        -496/81.,   -424/81.,    224/81.,   -352/81.,   3536/81.,   -928/81.,
        512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
        5120/81.,   -640/81.,   -640/81.,     80/81.,     80/81.,    800/81.,
        4544/81.,   -568/81.,   3968/81.,   -496/81.,   4040/81.,   4112/81.,
        -568/81.,    152/81.,   -496/81.,   -424/81.,  -1072/81.,   -352/81.},
   // Row     5 : MG ordering  = [ 3, 1, 2, 4, 7, 6, 5 ]
   {   5120/81.,    512/81.,    512/81.,  -4096/81.,  -4096/81.,  32768/81.,
       -640/81.,    -64/81.,   3968/81.,  -1792/81.,   -640/81.,   3968/81.,
       -64/81.,    512/81.,   4544/81.,   3968/81.,   5120/81.,   -640/81.,
       512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
       8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
       -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
       80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
       800/81.,     80/81.,   -424/81.,   -352/81.,   -496/81.,   4112/81.,
       4112/81.,   -352/81.,   -352/81.,   -928/81.,   -928/81.,   1088/81.,
       -496/81.,   -424/81.,    224/81.,   -352/81.,   3536/81.,   -928/81.,
       8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
       80/81.,   -640/81.,    152/81.,   -424/81.,   3968/81.,   -496/81.,
       4040/81.,  -1072/81.,  -1072/81.,   3536/81.,   3536/81.,   -928/81.,
       3968/81.,   -496/81.,  -1792/81.,    224/81.,  -1072/81.,   -352/81.,
       -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
       -640/81.,   5120/81.,     80/81.,    800/81.,   -640/81.,     80/81.,
       -568/81.,    152/81.,   -496/81.,   -424/81.,  -1072/81.,   -352/81.,
       4544/81.,   -568/81.,   3968/81.,   -496/81.,   4040/81.,   4112/81.},
   // Row     6 : MG ordering  = [ 3, 1, 2, 5, 4, 6, 7 ]
   {  -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
      32768/81.,  -4096/81.,  -4096/81.,    512/81.,    512/81.,   5120/81.,
      5120/81.,   -640/81.,    512/81.,    -64/81.,   3968/81.,   4544/81.,
      -640/81.,   3968/81.,    -64/81.,   -640/81.,  -1792/81.,   3968/81.,
      512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
      5120/81.,   -640/81.,   -640/81.,     80/81.,     80/81.,    800/81.,
      4544/81.,   -568/81.,   3968/81.,   -496/81.,   4040/81.,   4112/81.,
      -568/81.,    152/81.,   -496/81.,   -424/81.,  -1072/81.,   -352/81.,
      -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
      512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
      -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
      8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
      -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
      3968/81.,   -496/81.,  -1792/81.,    224/81.,  -1072/81.,   -352/81.,
      -640/81.,     80/81.,   3968/81.,   -496/81.,    152/81.,   -424/81.,
      -1072/81.,   4040/81.,   3536/81.,   -928/81.,  -1072/81.,   3536/81.,
      80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
      -496/81.,   -424/81.,    224/81.,   -352/81.,   3536/81.,   -928/81.,
      80/81.,    800/81.,   -496/81.,   4112/81.,   -424/81.,   -352/81.,
      -352/81.,   4112/81.,   -928/81.,   1088/81.,   -352/81.,   -928/81.},
   // Row     7 : MG ordering  = [ 3, 1, 2, 5, 4, 7, 6 ]
   {    512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
        -4096/81.,  32768/81.,    512/81.,   5120/81.,  -4096/81.,    512/81.,
        -640/81.,   3968/81.,    -64/81.,   -640/81.,  -1792/81.,   3968/81.,
        5120/81.,   -640/81.,    512/81.,    -64/81.,   3968/81.,   4544/81.,
        -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
        -640/81.,   5120/81.,     80/81.,    800/81.,   -640/81.,     80/81.,
        -568/81.,    152/81.,   -496/81.,   -424/81.,  -1072/81.,   -352/81.,
        4544/81.,   -568/81.,   3968/81.,   -496/81.,   4040/81.,   4112/81.,
        512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
        -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
        8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
        -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
        80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
        -496/81.,   -424/81.,    224/81.,   -352/81.,   3536/81.,   -928/81.,
        80/81.,    800/81.,   -496/81.,   4112/81.,   -424/81.,   -352/81.,
        -352/81.,   4112/81.,   -928/81.,   1088/81.,   -352/81.,   -928/81.,
        -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
        3968/81.,   -496/81.,  -1792/81.,    224/81.,  -1072/81.,   -352/81.,
        -640/81.,     80/81.,   3968/81.,   -496/81.,    152/81.,   -424/81.,
        -1072/81.,   4040/81.,   3536/81.,   -928/81.,  -1072/81.,   3536/81.},
   // Row     8 : MG ordering  = [ 3, 1, 2, 5, 6, 4, 7 ]
   {    512/81.,    -64/81.,   5120/81.,   -640/81.,   4544/81.,   3968/81.,
        -4096/81.,    512/81.,  32768/81.,  -4096/81.,   5120/81.,    512/81.,
        512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
        3968/81.,  -1792/81.,   -640/81.,    -64/81.,   3968/81.,   -640/81.,
        -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
        -640/81.,     80/81.,   3968/81.,   -496/81.,    152/81.,   -424/81.,
        3968/81.,   -496/81.,  -1792/81.,    224/81.,  -1072/81.,   -352/81.,
        4040/81.,  -1072/81.,  -1072/81.,   3536/81.,   3536/81.,   -928/81.,
        512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
        -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
        512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
        80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
        -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
        4544/81.,   -568/81.,   3968/81.,   -496/81.,   4040/81.,   4112/81.,
        5120/81.,   -640/81.,   -640/81.,     80/81.,     80/81.,    800/81.,
        152/81.,   -568/81.,  -1072/81.,   -352/81.,   -496/81.,   -424/81.,
        -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
        4112/81.,   -352/81.,   -352/81.,   -928/81.,   -928/81.,   1088/81.,
        800/81.,     80/81.,   -424/81.,   -352/81.,   -496/81.,   4112/81.,
        -424/81.,   -496/81.,   3536/81.,   -928/81.,    224/81.,   -352/81.},
   // Row     9 : MG ordering  = [ 3, 1, 2, 5, 6, 7, 4 ]
   {    -64/81.,   -640/81.,   -640/81.,   3968/81.,   3968/81.,  -1792/81.,
        512/81.,   5120/81.,  -4096/81.,  32768/81.,    512/81.,  -4096/81.,
        -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
        4544/81.,   3968/81.,    -64/81.,    512/81.,   -640/81.,   5120/81.,
        8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
        80/81.,    800/81.,   -496/81.,   4112/81.,   -424/81.,   -352/81.,
        -496/81.,   -424/81.,    224/81.,   -352/81.,   3536/81.,   -928/81.,
        4112/81.,   -352/81.,   -352/81.,   -928/81.,   -928/81.,   1088/81.,
        -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
        8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
        -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
        8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
        8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
        -568/81.,    152/81.,   -496/81.,   -424/81.,  -1072/81.,   -352/81.,
        -640/81.,   5120/81.,     80/81.,    800/81.,   -640/81.,     80/81.,
        -568/81.,   4544/81.,   4040/81.,   4112/81.,   3968/81.,   -496/81.,
        -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
        4040/81.,  -1072/81.,  -1072/81.,   3536/81.,   3536/81.,   -928/81.,
        80/81.,   -640/81.,    152/81.,   -424/81.,   3968/81.,   -496/81.,
        -496/81.,   3968/81.,  -1072/81.,   -352/81.,  -1792/81.,    224/81.},
   // Row    10 : MG ordering  = [ 3, 1, 2, 5, 7, 4, 6 ]
   {    -64/81.,    512/81.,   4544/81.,   3968/81.,   5120/81.,   -640/81.,
        512/81.,  -4096/81.,   5120/81.,    512/81.,  32768/81.,  -4096/81.,
        3968/81.,  -1792/81.,   -640/81.,    -64/81.,   3968/81.,   -640/81.,
        512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
        8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
        80/81.,   -640/81.,    152/81.,   -424/81.,   3968/81.,   -496/81.,
        4040/81.,  -1072/81.,  -1072/81.,   3536/81.,   3536/81.,   -928/81.,
        3968/81.,   -496/81.,  -1792/81.,    224/81.,  -1072/81.,   -352/81.,
        -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
        8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
        80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
        512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
        -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
        4112/81.,   -352/81.,   -352/81.,   -928/81.,   -928/81.,   1088/81.,
        800/81.,     80/81.,   -424/81.,   -352/81.,   -496/81.,   4112/81.,
        -424/81.,   -496/81.,   3536/81.,   -928/81.,    224/81.,   -352/81.,
        -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
        4544/81.,   -568/81.,   3968/81.,   -496/81.,   4040/81.,   4112/81.,
        5120/81.,   -640/81.,   -640/81.,     80/81.,     80/81.,    800/81.,
        152/81.,   -568/81.,  -1072/81.,   -352/81.,   -496/81.,   -424/81.},
   // Row    11 : MG ordering  = [ 3, 1, 2, 5, 7, 6, 4 ]
   {   -640/81.,    -64/81.,   3968/81.,  -1792/81.,   -640/81.,   3968/81.,
       5120/81.,    512/81.,    512/81.,  -4096/81.,  -4096/81.,  32768/81.,
       4544/81.,   3968/81.,    -64/81.,    512/81.,   -640/81.,   5120/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
       80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
       800/81.,     80/81.,   -424/81.,   -352/81.,   -496/81.,   4112/81.,
       4112/81.,   -352/81.,   -352/81.,   -928/81.,   -928/81.,   1088/81.,
       -496/81.,   -424/81.,    224/81.,   -352/81.,   3536/81.,   -928/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
       8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
       -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       4040/81.,  -1072/81.,  -1072/81.,   3536/81.,   3536/81.,   -928/81.,
       80/81.,   -640/81.,    152/81.,   -424/81.,   3968/81.,   -496/81.,
       -496/81.,   3968/81.,  -1072/81.,   -352/81.,  -1792/81.,    224/81.,
       8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
       -568/81.,    152/81.,   -496/81.,   -424/81.,  -1072/81.,   -352/81.,
       -640/81.,   5120/81.,     80/81.,    800/81.,   -640/81.,     80/81.,
       -568/81.,   4544/81.,   4040/81.,   4112/81.,   3968/81.,   -496/81.},
   // Row    12 : MG ordering  = [ 3, 1, 2, 6, 4, 5, 7 ]
   {    512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
        5120/81.,   -640/81.,    512/81.,    -64/81.,   3968/81.,   4544/81.,
        32768/81.,  -4096/81.,  -4096/81.,    512/81.,    512/81.,   5120/81.,
        3968/81.,   -640/81.,  -1792/81.,   3968/81.,    -64/81.,   -640/81.,
        -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
        4544/81.,   -568/81.,   3968/81.,   -496/81.,   4040/81.,   4112/81.,
        5120/81.,   -640/81.,   -640/81.,     80/81.,     80/81.,    800/81.,
        152/81.,   -568/81.,  -1072/81.,   -352/81.,   -496/81.,   -424/81.,
        -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
        3968/81.,   -496/81.,  -1792/81.,    224/81.,  -1072/81.,   -352/81.,
        -640/81.,     80/81.,   3968/81.,   -496/81.,    152/81.,   -424/81.,
        -1072/81.,   4040/81.,   3536/81.,   -928/81.,  -1072/81.,   3536/81.,
        -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
        512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
        -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
        8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
        -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
        -424/81.,   -496/81.,   3536/81.,   -928/81.,    224/81.,   -352/81.,
        -352/81.,   4112/81.,   -928/81.,   1088/81.,   -352/81.,   -928/81.,
        80/81.,    800/81.,   -496/81.,   4112/81.,   -424/81.,   -352/81.},
   // Row    13 : MG ordering  = [ 3, 1, 2, 6, 4, 7, 5 ]
   {    -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
        -640/81.,   3968/81.,    -64/81.,   -640/81.,  -1792/81.,   3968/81.,
        -4096/81.,  32768/81.,    512/81.,   5120/81.,  -4096/81.,    512/81.,
        -640/81.,   5120/81.,   3968/81.,   4544/81.,    512/81.,    -64/81.,
        8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
        -568/81.,    152/81.,   -496/81.,   -424/81.,  -1072/81.,   -352/81.,
        -640/81.,   5120/81.,     80/81.,    800/81.,   -640/81.,     80/81.,
        -568/81.,   4544/81.,   4040/81.,   4112/81.,   3968/81.,   -496/81.,
        80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
        -496/81.,   -424/81.,    224/81.,   -352/81.,   3536/81.,   -928/81.,
        80/81.,    800/81.,   -496/81.,   4112/81.,   -424/81.,   -352/81.,
        -352/81.,   4112/81.,   -928/81.,   1088/81.,   -352/81.,   -928/81.,
        512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
        -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
        8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
        -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
        80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
        -496/81.,   3968/81.,  -1072/81.,   -352/81.,  -1792/81.,    224/81.,
        -1072/81.,   4040/81.,   3536/81.,   -928/81.,  -1072/81.,   3536/81.,
        -640/81.,     80/81.,   3968/81.,   -496/81.,    152/81.,   -424/81.},
   // Row    14 : MG ordering  = [ 3, 1, 2, 6, 5, 4, 7 ]
   {   5120/81.,   -640/81.,    512/81.,    -64/81.,   3968/81.,   4544/81.,
       512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
       -4096/81.,    512/81.,  32768/81.,  -4096/81.,   5120/81.,    512/81.,
       -1792/81.,   3968/81.,   3968/81.,   -640/81.,   -640/81.,    -64/81.,
       -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
       3968/81.,   -496/81.,  -1792/81.,    224/81.,  -1072/81.,   -352/81.,
       -640/81.,     80/81.,   3968/81.,   -496/81.,    152/81.,   -424/81.,
       -1072/81.,   4040/81.,   3536/81.,   -928/81.,  -1072/81.,   3536/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       4544/81.,   -568/81.,   3968/81.,   -496/81.,   4040/81.,   4112/81.,
       5120/81.,   -640/81.,   -640/81.,     80/81.,     80/81.,    800/81.,
       152/81.,   -568/81.,  -1072/81.,   -352/81.,   -496/81.,   -424/81.,
       512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
       -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
       512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
       80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       -352/81.,   4112/81.,   -928/81.,   1088/81.,   -352/81.,   -928/81.,
       -424/81.,   -496/81.,   3536/81.,   -928/81.,    224/81.,   -352/81.,
       800/81.,     80/81.,   -424/81.,   -352/81.,   -496/81.,   4112/81.},
   // Row    15 : MG ordering  = [ 3, 1, 2, 6, 5, 7, 4 ]
   {   -640/81.,   3968/81.,    -64/81.,   -640/81.,  -1792/81.,   3968/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
       512/81.,   5120/81.,  -4096/81.,  32768/81.,    512/81.,  -4096/81.,
       3968/81.,   4544/81.,   -640/81.,   5120/81.,    -64/81.,    512/81.,
       80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
       -496/81.,   -424/81.,    224/81.,   -352/81.,   3536/81.,   -928/81.,
       80/81.,    800/81.,   -496/81.,   4112/81.,   -424/81.,   -352/81.,
       -352/81.,   4112/81.,   -928/81.,   1088/81.,   -352/81.,   -928/81.,
       8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
       -568/81.,    152/81.,   -496/81.,   -424/81.,  -1072/81.,   -352/81.,
       -640/81.,   5120/81.,     80/81.,    800/81.,   -640/81.,     80/81.,
       -568/81.,   4544/81.,   4040/81.,   4112/81.,   3968/81.,   -496/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
       8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
       -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
       8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       -1072/81.,   4040/81.,   3536/81.,   -928/81.,  -1072/81.,   3536/81.,
       -496/81.,   3968/81.,  -1072/81.,   -352/81.,  -1792/81.,    224/81.,
       80/81.,   -640/81.,    152/81.,   -424/81.,   3968/81.,   -496/81.},
   // Row    16 : MG ordering  = [ 3, 1, 2, 6, 7, 4, 5 ]
   {   4544/81.,   3968/81.,    -64/81.,    512/81.,   -640/81.,   5120/81.,
       3968/81.,  -1792/81.,   -640/81.,    -64/81.,   3968/81.,   -640/81.,
       512/81.,  -4096/81.,   5120/81.,    512/81.,  32768/81.,  -4096/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       4040/81.,  -1072/81.,  -1072/81.,   3536/81.,   3536/81.,   -928/81.,
       80/81.,   -640/81.,    152/81.,   -424/81.,   3968/81.,   -496/81.,
       -496/81.,   3968/81.,  -1072/81.,   -352/81.,  -1792/81.,    224/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       4112/81.,   -352/81.,   -352/81.,   -928/81.,   -928/81.,   1088/81.,
       800/81.,     80/81.,   -424/81.,   -352/81.,   -496/81.,   4112/81.,
       -424/81.,   -496/81.,   3536/81.,   -928/81.,    224/81.,   -352/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
       80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
       512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       -568/81.,   4544/81.,   4040/81.,   4112/81.,   3968/81.,   -496/81.,
       152/81.,   -568/81.,  -1072/81.,   -352/81.,   -496/81.,   -424/81.,
       5120/81.,   -640/81.,   -640/81.,     80/81.,     80/81.,    800/81.},
   // Row    17 : MG ordering  = [ 3, 1, 2, 6, 7, 5, 4 ]
   {   3968/81.,  -1792/81.,   -640/81.,    -64/81.,   3968/81.,   -640/81.,
       4544/81.,   3968/81.,    -64/81.,    512/81.,   -640/81.,   5120/81.,
       5120/81.,    512/81.,    512/81.,  -4096/81.,  -4096/81.,  32768/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       4112/81.,   -352/81.,   -352/81.,   -928/81.,   -928/81.,   1088/81.,
       800/81.,     80/81.,   -424/81.,   -352/81.,   -496/81.,   4112/81.,
       -424/81.,   -496/81.,   3536/81.,   -928/81.,    224/81.,   -352/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       4040/81.,  -1072/81.,  -1072/81.,   3536/81.,   3536/81.,   -928/81.,
       80/81.,   -640/81.,    152/81.,   -424/81.,   3968/81.,   -496/81.,
       -496/81.,   3968/81.,  -1072/81.,   -352/81.,  -1792/81.,    224/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
       8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
       -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       152/81.,   -568/81.,  -1072/81.,   -352/81.,   -496/81.,   -424/81.,
       -568/81.,   4544/81.,   4040/81.,   4112/81.,   3968/81.,   -496/81.,
       -640/81.,   5120/81.,     80/81.,    800/81.,   -640/81.,     80/81.},
   // Row    18 : MG ordering  = [ 3, 1, 2, 7, 4, 5, 6 ]
   {    -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
        -640/81.,   5120/81.,   3968/81.,   4544/81.,    512/81.,    -64/81.,
        3968/81.,   -640/81.,  -1792/81.,   3968/81.,    -64/81.,   -640/81.,
        32768/81.,  -4096/81.,  -4096/81.,    512/81.,    512/81.,   5120/81.,
        8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
        -568/81.,   4544/81.,   4040/81.,   4112/81.,   3968/81.,   -496/81.,
        152/81.,   -568/81.,  -1072/81.,   -352/81.,   -496/81.,   -424/81.,
        5120/81.,   -640/81.,   -640/81.,     80/81.,     80/81.,    800/81.,
        80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
        -496/81.,   3968/81.,  -1072/81.,   -352/81.,  -1792/81.,    224/81.,
        -1072/81.,   4040/81.,   3536/81.,   -928/81.,  -1072/81.,   3536/81.,
        -640/81.,     80/81.,   3968/81.,   -496/81.,    152/81.,   -424/81.,
        -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
        -424/81.,   -496/81.,   3536/81.,   -928/81.,    224/81.,   -352/81.,
        -352/81.,   4112/81.,   -928/81.,   1088/81.,   -352/81.,   -928/81.,
        80/81.,    800/81.,   -496/81.,   4112/81.,   -424/81.,   -352/81.,
        -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
        512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
        -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
        8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.},
   // Row    19 : MG ordering  = [ 3, 1, 2, 7, 4, 6, 5 ]
   {   -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       3968/81.,   -640/81.,  -1792/81.,   3968/81.,    -64/81.,   -640/81.,
       -640/81.,   5120/81.,   3968/81.,   4544/81.,    512/81.,    -64/81.,
       -4096/81.,  32768/81.,    512/81.,   5120/81.,  -4096/81.,    512/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       152/81.,   -568/81.,  -1072/81.,   -352/81.,   -496/81.,   -424/81.,
       -568/81.,   4544/81.,   4040/81.,   4112/81.,   3968/81.,   -496/81.,
       -640/81.,   5120/81.,     80/81.,    800/81.,   -640/81.,     80/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       -424/81.,   -496/81.,   3536/81.,   -928/81.,    224/81.,   -352/81.,
       -352/81.,   4112/81.,   -928/81.,   1088/81.,   -352/81.,   -928/81.,
       80/81.,    800/81.,   -496/81.,   4112/81.,   -424/81.,   -352/81.,
       80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
       -496/81.,   3968/81.,  -1072/81.,   -352/81.,  -1792/81.,    224/81.,
       -1072/81.,   4040/81.,   3536/81.,   -928/81.,  -1072/81.,   3536/81.,
       -640/81.,     80/81.,   3968/81.,   -496/81.,    152/81.,   -424/81.,
       512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
       -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
       8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
       -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.},
   // Row    20 : MG ordering  = [ 3, 1, 2, 7, 5, 4, 6 ]
   {   -640/81.,   5120/81.,   3968/81.,   4544/81.,    512/81.,    -64/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       -1792/81.,   3968/81.,   3968/81.,   -640/81.,   -640/81.,    -64/81.,
       -4096/81.,    512/81.,  32768/81.,  -4096/81.,   5120/81.,    512/81.,
       80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
       -496/81.,   3968/81.,  -1072/81.,   -352/81.,  -1792/81.,    224/81.,
       -1072/81.,   4040/81.,   3536/81.,   -928/81.,  -1072/81.,   3536/81.,
       -640/81.,     80/81.,   3968/81.,   -496/81.,    152/81.,   -424/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       -568/81.,   4544/81.,   4040/81.,   4112/81.,   3968/81.,   -496/81.,
       152/81.,   -568/81.,  -1072/81.,   -352/81.,   -496/81.,   -424/81.,
       5120/81.,   -640/81.,   -640/81.,     80/81.,     80/81.,    800/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       -352/81.,   4112/81.,   -928/81.,   1088/81.,   -352/81.,   -928/81.,
       -424/81.,   -496/81.,   3536/81.,   -928/81.,    224/81.,   -352/81.,
       800/81.,     80/81.,   -424/81.,   -352/81.,   -496/81.,   4112/81.,
       512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
       -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
       512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
       80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.},
   // Row    21 : MG ordering  = [ 3, 1, 2, 7, 5, 6, 4 ]
   {   3968/81.,   -640/81.,  -1792/81.,   3968/81.,    -64/81.,   -640/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       3968/81.,   4544/81.,   -640/81.,   5120/81.,    -64/81.,    512/81.,
       512/81.,   5120/81.,  -4096/81.,  32768/81.,    512/81.,  -4096/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       -424/81.,   -496/81.,   3536/81.,   -928/81.,    224/81.,   -352/81.,
       -352/81.,   4112/81.,   -928/81.,   1088/81.,   -352/81.,   -928/81.,
       80/81.,    800/81.,   -496/81.,   4112/81.,   -424/81.,   -352/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       152/81.,   -568/81.,  -1072/81.,   -352/81.,   -496/81.,   -424/81.,
       -568/81.,   4544/81.,   4040/81.,   4112/81.,   3968/81.,   -496/81.,
       -640/81.,   5120/81.,     80/81.,    800/81.,   -640/81.,     80/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       -1072/81.,   4040/81.,   3536/81.,   -928/81.,  -1072/81.,   3536/81.,
       -496/81.,   3968/81.,  -1072/81.,   -352/81.,  -1792/81.,    224/81.,
       80/81.,   -640/81.,    152/81.,   -424/81.,   3968/81.,   -496/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
       8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
       -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
       8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.},
   // Row    22 : MG ordering  = [ 3, 1, 2, 7, 6, 4, 5 ]
   {   3968/81.,   4544/81.,   -640/81.,   5120/81.,    -64/81.,    512/81.,
       -1792/81.,   3968/81.,   3968/81.,   -640/81.,   -640/81.,    -64/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       512/81.,  -4096/81.,   5120/81.,    512/81.,  32768/81.,  -4096/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       -1072/81.,   4040/81.,   3536/81.,   -928/81.,  -1072/81.,   3536/81.,
       -496/81.,   3968/81.,  -1072/81.,   -352/81.,  -1792/81.,    224/81.,
       80/81.,   -640/81.,    152/81.,   -424/81.,   3968/81.,   -496/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       -352/81.,   4112/81.,   -928/81.,   1088/81.,   -352/81.,   -928/81.,
       -424/81.,   -496/81.,   3536/81.,   -928/81.,    224/81.,   -352/81.,
       800/81.,     80/81.,   -424/81.,   -352/81.,   -496/81.,   4112/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       -568/81.,   4544/81.,   4040/81.,   4112/81.,   3968/81.,   -496/81.,
       152/81.,   -568/81.,  -1072/81.,   -352/81.,   -496/81.,   -424/81.,
       5120/81.,   -640/81.,   -640/81.,     80/81.,     80/81.,    800/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
       80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
       512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.},
   // Row    23 : MG ordering  = [ 3, 1, 2, 7, 6, 5, 4 ]
   {  -1792/81.,   3968/81.,   3968/81.,   -640/81.,   -640/81.,    -64/81.,
      3968/81.,   4544/81.,   -640/81.,   5120/81.,    -64/81.,    512/81.,
      -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
      5120/81.,    512/81.,    512/81.,  -4096/81.,  -4096/81.,  32768/81.,
      224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
      -352/81.,   4112/81.,   -928/81.,   1088/81.,   -352/81.,   -928/81.,
      -424/81.,   -496/81.,   3536/81.,   -928/81.,    224/81.,   -352/81.,
      800/81.,     80/81.,   -424/81.,   -352/81.,   -496/81.,   4112/81.,
      -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
      -1072/81.,   4040/81.,   3536/81.,   -928/81.,  -1072/81.,   3536/81.,
      -496/81.,   3968/81.,  -1072/81.,   -352/81.,  -1792/81.,    224/81.,
      80/81.,   -640/81.,    152/81.,   -424/81.,   3968/81.,   -496/81.,
      80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
      152/81.,   -568/81.,  -1072/81.,   -352/81.,   -496/81.,   -424/81.,
      -568/81.,   4544/81.,   4040/81.,   4112/81.,   3968/81.,   -496/81.,
      -640/81.,   5120/81.,     80/81.,    800/81.,   -640/81.,     80/81.,
      -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
      80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
      8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
      -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.},
   // Row    24 : MG ordering  = [ 3, 1, 4, 2, 5, 6, 7 ]
   {  -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
      512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
      -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
      8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
      32768/81.,  -4096/81.,  -4096/81.,    512/81.,    512/81.,   5120/81.,
      -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
      512/81.,    -64/81.,   5120/81.,   -640/81.,   4544/81.,   3968/81.,
      -64/81.,   -640/81.,   -640/81.,   3968/81.,   3968/81.,  -1792/81.,
      5120/81.,   -640/81.,   -640/81.,     80/81.,     80/81.,    800/81.,
      512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
      3968/81.,   -496/81.,   4544/81.,   -568/81.,   4112/81.,   4040/81.,
      -496/81.,   -424/81.,   -568/81.,    152/81.,   -352/81.,  -1072/81.,
      -640/81.,     80/81.,   3968/81.,   -496/81.,    152/81.,   -424/81.,
      -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
      -1792/81.,    224/81.,   3968/81.,   -496/81.,   -352/81.,  -1072/81.,
      -1072/81.,   3536/81.,   4040/81.,  -1072/81.,   -928/81.,   3536/81.,
      80/81.,    800/81.,   -496/81.,   4112/81.,   -424/81.,   -352/81.,
      8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
      224/81.,   -352/81.,   -496/81.,   -424/81.,   -928/81.,   3536/81.,
      -352/81.,   -928/81.,   4112/81.,   -352/81.,   1088/81.,   -928/81.},
   // Row    25 : MG ordering  = [ 3, 1, 4, 2, 5, 7, 6 ]
   {    512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
        -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
        8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
        -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
        -4096/81.,  32768/81.,    512/81.,   5120/81.,  -4096/81.,    512/81.,
        512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
        -64/81.,   -640/81.,   -640/81.,   3968/81.,   3968/81.,  -1792/81.,
        512/81.,    -64/81.,   5120/81.,   -640/81.,   4544/81.,   3968/81.,
        -640/81.,   5120/81.,     80/81.,    800/81.,   -640/81.,     80/81.,
        -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
        -496/81.,   -424/81.,   -568/81.,    152/81.,   -352/81.,  -1072/81.,
        3968/81.,   -496/81.,   4544/81.,   -568/81.,   4112/81.,   4040/81.,
        80/81.,    800/81.,   -496/81.,   4112/81.,   -424/81.,   -352/81.,
        8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
        224/81.,   -352/81.,   -496/81.,   -424/81.,   -928/81.,   3536/81.,
        -352/81.,   -928/81.,   4112/81.,   -352/81.,   1088/81.,   -928/81.,
        -640/81.,     80/81.,   3968/81.,   -496/81.,    152/81.,   -424/81.,
        -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
        -1792/81.,    224/81.,   3968/81.,   -496/81.,   -352/81.,  -1072/81.,
        -1072/81.,   3536/81.,   4040/81.,  -1072/81.,   -928/81.,   3536/81.},
   // Row    26 : MG ordering  = [ 3, 1, 4, 2, 6, 5, 7 ]
   {    512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
        -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
        512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
        80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
        -4096/81.,    512/81.,  32768/81.,  -4096/81.,   5120/81.,    512/81.,
        512/81.,    -64/81.,   5120/81.,   -640/81.,   4544/81.,   3968/81.,
        -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
        -640/81.,    -64/81.,   3968/81.,  -1792/81.,   -640/81.,   3968/81.,
        -640/81.,     80/81.,   3968/81.,   -496/81.,    152/81.,   -424/81.,
        -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
        -1792/81.,    224/81.,   3968/81.,   -496/81.,   -352/81.,  -1072/81.,
        -1072/81.,   3536/81.,   4040/81.,  -1072/81.,   -928/81.,   3536/81.,
        5120/81.,   -640/81.,   -640/81.,     80/81.,     80/81.,    800/81.,
        512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
        3968/81.,   -496/81.,   4544/81.,   -568/81.,   4112/81.,   4040/81.,
        -496/81.,   -424/81.,   -568/81.,    152/81.,   -352/81.,  -1072/81.,
        800/81.,     80/81.,   -424/81.,   -352/81.,   -496/81.,   4112/81.,
        80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
        -352/81.,   -928/81.,   4112/81.,   -352/81.,   1088/81.,   -928/81.,
        224/81.,   -352/81.,   -496/81.,   -424/81.,   -928/81.,   3536/81.},
   // Row    27 : MG ordering  = [ 3, 1, 4, 2, 6, 7, 5 ]
   {    -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
        8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
        -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
        8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
        512/81.,   5120/81.,  -4096/81.,  32768/81.,    512/81.,  -4096/81.,
        -64/81.,   -640/81.,   -640/81.,   3968/81.,   3968/81.,  -1792/81.,
        512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
        -64/81.,    512/81.,   4544/81.,   3968/81.,   5120/81.,   -640/81.,
        80/81.,    800/81.,   -496/81.,   4112/81.,   -424/81.,   -352/81.,
        8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
        224/81.,   -352/81.,   -496/81.,   -424/81.,   -928/81.,   3536/81.,
        -352/81.,   -928/81.,   4112/81.,   -352/81.,   1088/81.,   -928/81.,
        -640/81.,   5120/81.,     80/81.,    800/81.,   -640/81.,     80/81.,
        -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
        -496/81.,   -424/81.,   -568/81.,    152/81.,   -352/81.,  -1072/81.,
        3968/81.,   -496/81.,   4544/81.,   -568/81.,   4112/81.,   4040/81.,
        80/81.,   -640/81.,    152/81.,   -424/81.,   3968/81.,   -496/81.,
        8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
        -1072/81.,   3536/81.,   4040/81.,  -1072/81.,   -928/81.,   3536/81.,
        -1792/81.,    224/81.,   3968/81.,   -496/81.,   -352/81.,  -1072/81.},
   // Row    28 : MG ordering  = [ 3, 1, 4, 2, 7, 5, 6 ]
   {    -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
        8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
        80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
        512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
        512/81.,  -4096/81.,   5120/81.,    512/81.,  32768/81.,  -4096/81.,
        -64/81.,    512/81.,   4544/81.,   3968/81.,   5120/81.,   -640/81.,
        -640/81.,    -64/81.,   3968/81.,  -1792/81.,   -640/81.,   3968/81.,
        -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
        80/81.,   -640/81.,    152/81.,   -424/81.,   3968/81.,   -496/81.,
        8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
        -1072/81.,   3536/81.,   4040/81.,  -1072/81.,   -928/81.,   3536/81.,
        -1792/81.,    224/81.,   3968/81.,   -496/81.,   -352/81.,  -1072/81.,
        800/81.,     80/81.,   -424/81.,   -352/81.,   -496/81.,   4112/81.,
        80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
        -352/81.,   -928/81.,   4112/81.,   -352/81.,   1088/81.,   -928/81.,
        224/81.,   -352/81.,   -496/81.,   -424/81.,   -928/81.,   3536/81.,
        5120/81.,   -640/81.,   -640/81.,     80/81.,     80/81.,    800/81.,
        512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
        3968/81.,   -496/81.,   4544/81.,   -568/81.,   4112/81.,   4040/81.,
        -496/81.,   -424/81.,   -568/81.,    152/81.,   -352/81.,  -1072/81.},
   // Row    29 : MG ordering  = [ 3, 1, 4, 2, 7, 6, 5 ]
   {   -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
       8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
       -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
       5120/81.,    512/81.,    512/81.,  -4096/81.,  -4096/81.,  32768/81.,
       -640/81.,    -64/81.,   3968/81.,  -1792/81.,   -640/81.,   3968/81.,
       -64/81.,    512/81.,   4544/81.,   3968/81.,   5120/81.,   -640/81.,
       512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
       800/81.,     80/81.,   -424/81.,   -352/81.,   -496/81.,   4112/81.,
       80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
       -352/81.,   -928/81.,   4112/81.,   -352/81.,   1088/81.,   -928/81.,
       224/81.,   -352/81.,   -496/81.,   -424/81.,   -928/81.,   3536/81.,
       80/81.,   -640/81.,    152/81.,   -424/81.,   3968/81.,   -496/81.,
       8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
       -1072/81.,   3536/81.,   4040/81.,  -1072/81.,   -928/81.,   3536/81.,
       -1792/81.,    224/81.,   3968/81.,   -496/81.,   -352/81.,  -1072/81.,
       -640/81.,   5120/81.,     80/81.,    800/81.,   -640/81.,     80/81.,
       -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
       -496/81.,   -424/81.,   -568/81.,    152/81.,   -352/81.,  -1072/81.,
       3968/81.,   -496/81.,   4544/81.,   -568/81.,   4112/81.,   4040/81.},
   // Row    30 : MG ordering  = [ 3, 1, 4, 5, 2, 6, 7 ]
   {    512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
        5120/81.,   -640/81.,   -640/81.,     80/81.,     80/81.,    800/81.,
        4544/81.,   -568/81.,   3968/81.,   -496/81.,   4040/81.,   4112/81.,
        -568/81.,    152/81.,   -496/81.,   -424/81.,  -1072/81.,   -352/81.,
        -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
        32768/81.,  -4096/81.,  -4096/81.,    512/81.,    512/81.,   5120/81.,
        5120/81.,   -640/81.,    512/81.,    -64/81.,   3968/81.,   4544/81.,
        -640/81.,   3968/81.,    -64/81.,   -640/81.,  -1792/81.,   3968/81.,
        512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
        -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
        -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
        80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
        3968/81.,   -496/81.,  -1792/81.,    224/81.,  -1072/81.,   -352/81.,
        -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
        3968/81.,   -496/81.,   -640/81.,     80/81.,   -424/81.,    152/81.,
        3536/81.,   -928/81.,  -1072/81.,   4040/81.,   3536/81.,  -1072/81.,
        -496/81.,   -424/81.,    224/81.,   -352/81.,   3536/81.,   -928/81.,
        80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
        -496/81.,   4112/81.,     80/81.,    800/81.,   -352/81.,   -424/81.,
        -928/81.,   1088/81.,   -352/81.,   4112/81.,   -928/81.,   -352/81.},
   // Row    31 : MG ordering  = [ 3, 1, 4, 5, 2, 7, 6 ]
   {    -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
        -640/81.,   5120/81.,     80/81.,    800/81.,   -640/81.,     80/81.,
        -568/81.,    152/81.,   -496/81.,   -424/81.,  -1072/81.,   -352/81.,
        4544/81.,   -568/81.,   3968/81.,   -496/81.,   4040/81.,   4112/81.,
        512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
        -4096/81.,  32768/81.,    512/81.,   5120/81.,  -4096/81.,    512/81.,
        -640/81.,   3968/81.,    -64/81.,   -640/81.,  -1792/81.,   3968/81.,
        5120/81.,   -640/81.,    512/81.,    -64/81.,   3968/81.,   4544/81.,
        -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
        512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
        80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
        -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
        -496/81.,   -424/81.,    224/81.,   -352/81.,   3536/81.,   -928/81.,
        80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
        -496/81.,   4112/81.,     80/81.,    800/81.,   -352/81.,   -424/81.,
        -928/81.,   1088/81.,   -352/81.,   4112/81.,   -928/81.,   -352/81.,
        3968/81.,   -496/81.,  -1792/81.,    224/81.,  -1072/81.,   -352/81.,
        -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
        3968/81.,   -496/81.,   -640/81.,     80/81.,   -424/81.,    152/81.,
        3536/81.,   -928/81.,  -1072/81.,   4040/81.,   3536/81.,  -1072/81.},
   // Row    32 : MG ordering  = [ 3, 1, 4, 5, 6, 2, 7 ]
   {    -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
        -640/81.,     80/81.,   3968/81.,   -496/81.,    152/81.,   -424/81.,
        3968/81.,   -496/81.,  -1792/81.,    224/81.,  -1072/81.,   -352/81.,
        4040/81.,  -1072/81.,  -1072/81.,   3536/81.,   3536/81.,   -928/81.,
        512/81.,    -64/81.,   5120/81.,   -640/81.,   4544/81.,   3968/81.,
        -4096/81.,    512/81.,  32768/81.,  -4096/81.,   5120/81.,    512/81.,
        512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
        3968/81.,  -1792/81.,   -640/81.,    -64/81.,   3968/81.,   -640/81.,
        -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
        512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
        -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
        -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
        4544/81.,   -568/81.,   3968/81.,   -496/81.,   4040/81.,   4112/81.,
        -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
        -640/81.,     80/81.,   5120/81.,   -640/81.,    800/81.,     80/81.,
        -1072/81.,   -352/81.,    152/81.,   -568/81.,   -424/81.,   -496/81.,
        4112/81.,   -352/81.,   -352/81.,   -928/81.,   -928/81.,   1088/81.,
        -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
        -424/81.,   -352/81.,    800/81.,     80/81.,   4112/81.,   -496/81.,
        3536/81.,   -928/81.,   -424/81.,   -496/81.,   -352/81.,    224/81.},
   // Row    33 : MG ordering  = [ 3, 1, 4, 5, 6, 7, 2 ]
   {    8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
        80/81.,    800/81.,   -496/81.,   4112/81.,   -424/81.,   -352/81.,
        -496/81.,   -424/81.,    224/81.,   -352/81.,   3536/81.,   -928/81.,
        4112/81.,   -352/81.,   -352/81.,   -928/81.,   -928/81.,   1088/81.,
        -64/81.,   -640/81.,   -640/81.,   3968/81.,   3968/81.,  -1792/81.,
        512/81.,   5120/81.,  -4096/81.,  32768/81.,    512/81.,  -4096/81.,
        -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
        4544/81.,   3968/81.,    -64/81.,    512/81.,   -640/81.,   5120/81.,
        8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
        -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
        8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
        -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
        -568/81.,    152/81.,   -496/81.,   -424/81.,  -1072/81.,   -352/81.,
        8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
        80/81.,    800/81.,   -640/81.,   5120/81.,     80/81.,   -640/81.,
        4040/81.,   4112/81.,   -568/81.,   4544/81.,   -496/81.,   3968/81.,
        4040/81.,  -1072/81.,  -1072/81.,   3536/81.,   3536/81.,   -928/81.,
        -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
        152/81.,   -424/81.,     80/81.,   -640/81.,   -496/81.,   3968/81.,
        -1072/81.,   -352/81.,   -496/81.,   3968/81.,    224/81.,  -1792/81.},
   // Row    34 : MG ordering  = [ 3, 1, 4, 5, 7, 2, 6 ]
   {     8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
         80/81.,   -640/81.,    152/81.,   -424/81.,   3968/81.,   -496/81.,
         4040/81.,  -1072/81.,  -1072/81.,   3536/81.,   3536/81.,   -928/81.,
         3968/81.,   -496/81.,  -1792/81.,    224/81.,  -1072/81.,   -352/81.,
         -64/81.,    512/81.,   4544/81.,   3968/81.,   5120/81.,   -640/81.,
         512/81.,  -4096/81.,   5120/81.,    512/81.,  32768/81.,  -4096/81.,
         3968/81.,  -1792/81.,   -640/81.,    -64/81.,   3968/81.,   -640/81.,
         512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
         8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
         -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
         -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
         -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
         4112/81.,   -352/81.,   -352/81.,   -928/81.,   -928/81.,   1088/81.,
         -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
         -424/81.,   -352/81.,    800/81.,     80/81.,   4112/81.,   -496/81.,
         3536/81.,   -928/81.,   -424/81.,   -496/81.,   -352/81.,    224/81.,
         4544/81.,   -568/81.,   3968/81.,   -496/81.,   4040/81.,   4112/81.,
         -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
         -640/81.,     80/81.,   5120/81.,   -640/81.,    800/81.,     80/81.,
         -1072/81.,   -352/81.,    152/81.,   -568/81.,   -424/81.,   -496/81.},
   // Row    35 : MG ordering  = [ 3, 1, 4, 5, 7, 6, 2 ]
   {    80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
        800/81.,     80/81.,   -424/81.,   -352/81.,   -496/81.,   4112/81.,
        4112/81.,   -352/81.,   -352/81.,   -928/81.,   -928/81.,   1088/81.,
        -496/81.,   -424/81.,    224/81.,   -352/81.,   3536/81.,   -928/81.,
        -640/81.,    -64/81.,   3968/81.,  -1792/81.,   -640/81.,   3968/81.,
        5120/81.,    512/81.,    512/81.,  -4096/81.,  -4096/81.,  32768/81.,
        4544/81.,   3968/81.,    -64/81.,    512/81.,   -640/81.,   5120/81.,
        -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
        80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
        -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
        -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
        8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
        4040/81.,  -1072/81.,  -1072/81.,   3536/81.,   3536/81.,   -928/81.,
        -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
        152/81.,   -424/81.,     80/81.,   -640/81.,   -496/81.,   3968/81.,
        -1072/81.,   -352/81.,   -496/81.,   3968/81.,    224/81.,  -1792/81.,
        -568/81.,    152/81.,   -496/81.,   -424/81.,  -1072/81.,   -352/81.,
        8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
        80/81.,    800/81.,   -640/81.,   5120/81.,     80/81.,   -640/81.,
        4040/81.,   4112/81.,   -568/81.,   4544/81.,   -496/81.,   3968/81.},
   // Row    36 : MG ordering  = [ 3, 1, 4, 6, 2, 5, 7 ]
   {    -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
        4544/81.,   -568/81.,   3968/81.,   -496/81.,   4040/81.,   4112/81.,
        5120/81.,   -640/81.,   -640/81.,     80/81.,     80/81.,    800/81.,
        152/81.,   -568/81.,  -1072/81.,   -352/81.,   -496/81.,   -424/81.,
        512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
        5120/81.,   -640/81.,    512/81.,    -64/81.,   3968/81.,   4544/81.,
        32768/81.,  -4096/81.,  -4096/81.,    512/81.,    512/81.,   5120/81.,
        3968/81.,   -640/81.,  -1792/81.,   3968/81.,    -64/81.,   -640/81.,
        3968/81.,   -496/81.,  -1792/81.,    224/81.,  -1072/81.,   -352/81.,
        -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
        3968/81.,   -496/81.,   -640/81.,     80/81.,   -424/81.,    152/81.,
        3536/81.,   -928/81.,  -1072/81.,   4040/81.,   3536/81.,  -1072/81.,
        512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
        -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
        -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
        80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
        -424/81.,   -496/81.,   3536/81.,   -928/81.,    224/81.,   -352/81.,
        -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
        -928/81.,   1088/81.,   -352/81.,   4112/81.,   -928/81.,   -352/81.,
        -496/81.,   4112/81.,     80/81.,    800/81.,   -352/81.,   -424/81.},
   // Row    37 : MG ordering  = [ 3, 1, 4, 6, 2, 7, 5 ]
   {    8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
        -568/81.,    152/81.,   -496/81.,   -424/81.,  -1072/81.,   -352/81.,
        -640/81.,   5120/81.,     80/81.,    800/81.,   -640/81.,     80/81.,
        -568/81.,   4544/81.,   4040/81.,   4112/81.,   3968/81.,   -496/81.,
        -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
        -640/81.,   3968/81.,    -64/81.,   -640/81.,  -1792/81.,   3968/81.,
        -4096/81.,  32768/81.,    512/81.,   5120/81.,  -4096/81.,    512/81.,
        -640/81.,   5120/81.,   3968/81.,   4544/81.,    512/81.,    -64/81.,
        -496/81.,   -424/81.,    224/81.,   -352/81.,   3536/81.,   -928/81.,
        80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
        -496/81.,   4112/81.,     80/81.,    800/81.,   -352/81.,   -424/81.,
        -928/81.,   1088/81.,   -352/81.,   4112/81.,   -928/81.,   -352/81.,
        -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
        512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
        80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
        -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
        -496/81.,   3968/81.,  -1072/81.,   -352/81.,  -1792/81.,    224/81.,
        80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
        3536/81.,   -928/81.,  -1072/81.,   4040/81.,   3536/81.,  -1072/81.,
        3968/81.,   -496/81.,   -640/81.,     80/81.,   -424/81.,    152/81.},
   // Row    38 : MG ordering  = [ 3, 1, 4, 6, 5, 2, 7 ]
   {   -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
       3968/81.,   -496/81.,  -1792/81.,    224/81.,  -1072/81.,   -352/81.,
       -640/81.,     80/81.,   3968/81.,   -496/81.,    152/81.,   -424/81.,
       -1072/81.,   4040/81.,   3536/81.,   -928/81.,  -1072/81.,   3536/81.,
       5120/81.,   -640/81.,    512/81.,    -64/81.,   3968/81.,   4544/81.,
       512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
       -4096/81.,    512/81.,  32768/81.,  -4096/81.,   5120/81.,    512/81.,
       -1792/81.,   3968/81.,   3968/81.,   -640/81.,   -640/81.,    -64/81.,
       4544/81.,   -568/81.,   3968/81.,   -496/81.,   4040/81.,   4112/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       -640/81.,     80/81.,   5120/81.,   -640/81.,    800/81.,     80/81.,
       -1072/81.,   -352/81.,    152/81.,   -568/81.,   -424/81.,   -496/81.,
       -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
       512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       -352/81.,   4112/81.,   -928/81.,   1088/81.,   -352/81.,   -928/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       3536/81.,   -928/81.,   -424/81.,   -496/81.,   -352/81.,    224/81.,
       -424/81.,   -352/81.,    800/81.,     80/81.,   4112/81.,   -496/81.},
   // Row    39 : MG ordering  = [ 3, 1, 4, 6, 5, 7, 2 ]
   {     80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
         -496/81.,   -424/81.,    224/81.,   -352/81.,   3536/81.,   -928/81.,
         80/81.,    800/81.,   -496/81.,   4112/81.,   -424/81.,   -352/81.,
         -352/81.,   4112/81.,   -928/81.,   1088/81.,   -352/81.,   -928/81.,
         -640/81.,   3968/81.,    -64/81.,   -640/81.,  -1792/81.,   3968/81.,
         -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
         512/81.,   5120/81.,  -4096/81.,  32768/81.,    512/81.,  -4096/81.,
         3968/81.,   4544/81.,   -640/81.,   5120/81.,    -64/81.,    512/81.,
         -568/81.,    152/81.,   -496/81.,   -424/81.,  -1072/81.,   -352/81.,
         8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
         80/81.,    800/81.,   -640/81.,   5120/81.,     80/81.,   -640/81.,
         4040/81.,   4112/81.,   -568/81.,   4544/81.,   -496/81.,   3968/81.,
         8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
         -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
         8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
         -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
         -1072/81.,   4040/81.,   3536/81.,   -928/81.,  -1072/81.,   3536/81.,
         -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
         -1072/81.,   -352/81.,   -496/81.,   3968/81.,    224/81.,  -1792/81.,
         152/81.,   -424/81.,     80/81.,   -640/81.,   -496/81.,   3968/81.},
   // Row    40 : MG ordering  = [ 3, 1, 4, 6, 7, 2, 5 ]
   {   -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       4040/81.,  -1072/81.,  -1072/81.,   3536/81.,   3536/81.,   -928/81.,
       80/81.,   -640/81.,    152/81.,   -424/81.,   3968/81.,   -496/81.,
       -496/81.,   3968/81.,  -1072/81.,   -352/81.,  -1792/81.,    224/81.,
       4544/81.,   3968/81.,    -64/81.,    512/81.,   -640/81.,   5120/81.,
       3968/81.,  -1792/81.,   -640/81.,    -64/81.,   3968/81.,   -640/81.,
       512/81.,  -4096/81.,   5120/81.,    512/81.,  32768/81.,  -4096/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       4112/81.,   -352/81.,   -352/81.,   -928/81.,   -928/81.,   1088/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       -424/81.,   -352/81.,    800/81.,     80/81.,   4112/81.,   -496/81.,
       3536/81.,   -928/81.,   -424/81.,   -496/81.,   -352/81.,    224/81.,
       8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       -568/81.,   4544/81.,   4040/81.,   4112/81.,   3968/81.,   -496/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       -1072/81.,   -352/81.,    152/81.,   -568/81.,   -424/81.,   -496/81.,
       -640/81.,     80/81.,   5120/81.,   -640/81.,    800/81.,     80/81.},
   // Row    41 : MG ordering  = [ 3, 1, 4, 6, 7, 5, 2 ]
   {   -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       4112/81.,   -352/81.,   -352/81.,   -928/81.,   -928/81.,   1088/81.,
       800/81.,     80/81.,   -424/81.,   -352/81.,   -496/81.,   4112/81.,
       -424/81.,   -496/81.,   3536/81.,   -928/81.,    224/81.,   -352/81.,
       3968/81.,  -1792/81.,   -640/81.,    -64/81.,   3968/81.,   -640/81.,
       4544/81.,   3968/81.,    -64/81.,    512/81.,   -640/81.,   5120/81.,
       5120/81.,    512/81.,    512/81.,  -4096/81.,  -4096/81.,  32768/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       4040/81.,  -1072/81.,  -1072/81.,   3536/81.,   3536/81.,   -928/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       152/81.,   -424/81.,     80/81.,   -640/81.,   -496/81.,   3968/81.,
       -1072/81.,   -352/81.,   -496/81.,   3968/81.,    224/81.,  -1792/81.,
       80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
       152/81.,   -568/81.,  -1072/81.,   -352/81.,   -496/81.,   -424/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       4040/81.,   4112/81.,   -568/81.,   4544/81.,   -496/81.,   3968/81.,
       80/81.,    800/81.,   -640/81.,   5120/81.,     80/81.,   -640/81.},
   // Row    42 : MG ordering  = [ 3, 1, 4, 7, 2, 5, 6 ]
   {    8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
        -568/81.,   4544/81.,   4040/81.,   4112/81.,   3968/81.,   -496/81.,
        152/81.,   -568/81.,  -1072/81.,   -352/81.,   -496/81.,   -424/81.,
        5120/81.,   -640/81.,   -640/81.,     80/81.,     80/81.,    800/81.,
        -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
        -640/81.,   5120/81.,   3968/81.,   4544/81.,    512/81.,    -64/81.,
        3968/81.,   -640/81.,  -1792/81.,   3968/81.,    -64/81.,   -640/81.,
        32768/81.,  -4096/81.,  -4096/81.,    512/81.,    512/81.,   5120/81.,
        -496/81.,   3968/81.,  -1072/81.,   -352/81.,  -1792/81.,    224/81.,
        80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
        3536/81.,   -928/81.,  -1072/81.,   4040/81.,   3536/81.,  -1072/81.,
        3968/81.,   -496/81.,   -640/81.,     80/81.,   -424/81.,    152/81.,
        -424/81.,   -496/81.,   3536/81.,   -928/81.,    224/81.,   -352/81.,
        -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
        -928/81.,   1088/81.,   -352/81.,   4112/81.,   -928/81.,   -352/81.,
        -496/81.,   4112/81.,     80/81.,    800/81.,   -352/81.,   -424/81.,
        512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
        -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
        -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
        80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.},
   // Row    43 : MG ordering  = [ 3, 1, 4, 7, 2, 6, 5 ]
   {     80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
         152/81.,   -568/81.,  -1072/81.,   -352/81.,   -496/81.,   -424/81.,
         -568/81.,   4544/81.,   4040/81.,   4112/81.,   3968/81.,   -496/81.,
         -640/81.,   5120/81.,     80/81.,    800/81.,   -640/81.,     80/81.,
         -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
         3968/81.,   -640/81.,  -1792/81.,   3968/81.,    -64/81.,   -640/81.,
         -640/81.,   5120/81.,   3968/81.,   4544/81.,    512/81.,    -64/81.,
         -4096/81.,  32768/81.,    512/81.,   5120/81.,  -4096/81.,    512/81.,
         -424/81.,   -496/81.,   3536/81.,   -928/81.,    224/81.,   -352/81.,
         -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
         -928/81.,   1088/81.,   -352/81.,   4112/81.,   -928/81.,   -352/81.,
         -496/81.,   4112/81.,     80/81.,    800/81.,   -352/81.,   -424/81.,
         -496/81.,   3968/81.,  -1072/81.,   -352/81.,  -1792/81.,    224/81.,
         80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
         3536/81.,   -928/81.,  -1072/81.,   4040/81.,   3536/81.,  -1072/81.,
         3968/81.,   -496/81.,   -640/81.,     80/81.,   -424/81.,    152/81.,
         -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
         512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
         80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
         -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.},
   // Row    44 : MG ordering  = [ 3, 1, 4, 7, 5, 2, 6 ]
   {     80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
         -496/81.,   3968/81.,  -1072/81.,   -352/81.,  -1792/81.,    224/81.,
         -1072/81.,   4040/81.,   3536/81.,   -928/81.,  -1072/81.,   3536/81.,
         -640/81.,     80/81.,   3968/81.,   -496/81.,    152/81.,   -424/81.,
         -640/81.,   5120/81.,   3968/81.,   4544/81.,    512/81.,    -64/81.,
         -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
         -1792/81.,   3968/81.,   3968/81.,   -640/81.,   -640/81.,    -64/81.,
         -4096/81.,    512/81.,  32768/81.,  -4096/81.,   5120/81.,    512/81.,
         -568/81.,   4544/81.,   4040/81.,   4112/81.,   3968/81.,   -496/81.,
         8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
         -1072/81.,   -352/81.,    152/81.,   -568/81.,   -424/81.,   -496/81.,
         -640/81.,     80/81.,   5120/81.,   -640/81.,    800/81.,     80/81.,
         -352/81.,   4112/81.,   -928/81.,   1088/81.,   -352/81.,   -928/81.,
         224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
         3536/81.,   -928/81.,   -424/81.,   -496/81.,   -352/81.,    224/81.,
         -424/81.,   -352/81.,    800/81.,     80/81.,   4112/81.,   -496/81.,
         -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
         512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
         -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
         -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.},
   // Row    45 : MG ordering  = [ 3, 1, 4, 7, 5, 6, 2 ]
   {   -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       -424/81.,   -496/81.,   3536/81.,   -928/81.,    224/81.,   -352/81.,
       -352/81.,   4112/81.,   -928/81.,   1088/81.,   -352/81.,   -928/81.,
       80/81.,    800/81.,   -496/81.,   4112/81.,   -424/81.,   -352/81.,
       3968/81.,   -640/81.,  -1792/81.,   3968/81.,    -64/81.,   -640/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       3968/81.,   4544/81.,   -640/81.,   5120/81.,    -64/81.,    512/81.,
       512/81.,   5120/81.,  -4096/81.,  32768/81.,    512/81.,  -4096/81.,
       152/81.,   -568/81.,  -1072/81.,   -352/81.,   -496/81.,   -424/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       4040/81.,   4112/81.,   -568/81.,   4544/81.,   -496/81.,   3968/81.,
       80/81.,    800/81.,   -640/81.,   5120/81.,     80/81.,   -640/81.,
       -1072/81.,   4040/81.,   3536/81.,   -928/81.,  -1072/81.,   3536/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       -1072/81.,   -352/81.,   -496/81.,   3968/81.,    224/81.,  -1792/81.,
       152/81.,   -424/81.,     80/81.,   -640/81.,   -496/81.,   3968/81.,
       8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
       8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.},
   // Row    46 : MG ordering  = [ 3, 1, 4, 7, 6, 2, 5 ]
   {   -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       -1072/81.,   4040/81.,   3536/81.,   -928/81.,  -1072/81.,   3536/81.,
       -496/81.,   3968/81.,  -1072/81.,   -352/81.,  -1792/81.,    224/81.,
       80/81.,   -640/81.,    152/81.,   -424/81.,   3968/81.,   -496/81.,
       3968/81.,   4544/81.,   -640/81.,   5120/81.,    -64/81.,    512/81.,
       -1792/81.,   3968/81.,   3968/81.,   -640/81.,   -640/81.,    -64/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       512/81.,  -4096/81.,   5120/81.,    512/81.,  32768/81.,  -4096/81.,
       -352/81.,   4112/81.,   -928/81.,   1088/81.,   -352/81.,   -928/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       3536/81.,   -928/81.,   -424/81.,   -496/81.,   -352/81.,    224/81.,
       -424/81.,   -352/81.,    800/81.,     80/81.,   4112/81.,   -496/81.,
       -568/81.,   4544/81.,   4040/81.,   4112/81.,   3968/81.,   -496/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       -1072/81.,   -352/81.,    152/81.,   -568/81.,   -424/81.,   -496/81.,
       -640/81.,     80/81.,   5120/81.,   -640/81.,    800/81.,     80/81.,
       8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.},
   // Row    47 : MG ordering  = [ 3, 1, 4, 7, 6, 5, 2 ]
   {    224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
        -352/81.,   4112/81.,   -928/81.,   1088/81.,   -352/81.,   -928/81.,
        -424/81.,   -496/81.,   3536/81.,   -928/81.,    224/81.,   -352/81.,
        800/81.,     80/81.,   -424/81.,   -352/81.,   -496/81.,   4112/81.,
        -1792/81.,   3968/81.,   3968/81.,   -640/81.,   -640/81.,    -64/81.,
        3968/81.,   4544/81.,   -640/81.,   5120/81.,    -64/81.,    512/81.,
        -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
        5120/81.,    512/81.,    512/81.,  -4096/81.,  -4096/81.,  32768/81.,
        -1072/81.,   4040/81.,   3536/81.,   -928/81.,  -1072/81.,   3536/81.,
        -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
        -1072/81.,   -352/81.,   -496/81.,   3968/81.,    224/81.,  -1792/81.,
        152/81.,   -424/81.,     80/81.,   -640/81.,   -496/81.,   3968/81.,
        152/81.,   -568/81.,  -1072/81.,   -352/81.,   -496/81.,   -424/81.,
        80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
        4040/81.,   4112/81.,   -568/81.,   4544/81.,   -496/81.,   3968/81.,
        80/81.,    800/81.,   -640/81.,   5120/81.,     80/81.,   -640/81.,
        80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
        -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
        -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
        8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.},
   // Row    48 : MG ordering  = [ 3, 1, 5, 2, 4, 6, 7 ]
   {    512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
        -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
        -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
        80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
        5120/81.,   -640/81.,   -640/81.,     80/81.,     80/81.,    800/81.,
        512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
        3968/81.,   -496/81.,   4544/81.,   -568/81.,   4112/81.,   4040/81.,
        -496/81.,   -424/81.,   -568/81.,    152/81.,   -352/81.,  -1072/81.,
        32768/81.,  -4096/81.,  -4096/81.,    512/81.,    512/81.,   5120/81.,
        -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
        512/81.,    -64/81.,   5120/81.,   -640/81.,   4544/81.,   3968/81.,
        -64/81.,   -640/81.,   -640/81.,   3968/81.,   3968/81.,  -1792/81.,
        3968/81.,   -496/81.,   -640/81.,     80/81.,   -424/81.,    152/81.,
        -1792/81.,    224/81.,   3968/81.,   -496/81.,   -352/81.,  -1072/81.,
        -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
        3536/81.,  -1072/81.,   -928/81.,   3536/81.,   4040/81.,  -1072/81.,
        -496/81.,   4112/81.,     80/81.,    800/81.,   -352/81.,   -424/81.,
        224/81.,   -352/81.,   -496/81.,   -424/81.,   -928/81.,   3536/81.,
        8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
        -928/81.,   -352/81.,   1088/81.,   -928/81.,   4112/81.,   -352/81.},
   // Row    49 : MG ordering  = [ 3, 1, 5, 2, 4, 7, 6 ]
   {    -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
        512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
        80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
        -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
        -640/81.,   5120/81.,     80/81.,    800/81.,   -640/81.,     80/81.,
        -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
        -496/81.,   -424/81.,   -568/81.,    152/81.,   -352/81.,  -1072/81.,
        3968/81.,   -496/81.,   4544/81.,   -568/81.,   4112/81.,   4040/81.,
        -4096/81.,  32768/81.,    512/81.,   5120/81.,  -4096/81.,    512/81.,
        512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
        -64/81.,   -640/81.,   -640/81.,   3968/81.,   3968/81.,  -1792/81.,
        512/81.,    -64/81.,   5120/81.,   -640/81.,   4544/81.,   3968/81.,
        -496/81.,   4112/81.,     80/81.,    800/81.,   -352/81.,   -424/81.,
        224/81.,   -352/81.,   -496/81.,   -424/81.,   -928/81.,   3536/81.,
        8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
        -928/81.,   -352/81.,   1088/81.,   -928/81.,   4112/81.,   -352/81.,
        3968/81.,   -496/81.,   -640/81.,     80/81.,   -424/81.,    152/81.,
        -1792/81.,    224/81.,   3968/81.,   -496/81.,   -352/81.,  -1072/81.,
        -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
        3536/81.,  -1072/81.,   -928/81.,   3536/81.,   4040/81.,  -1072/81.},
   // Row    50 : MG ordering  = [ 3, 1, 5, 2, 6, 4, 7 ]
   {    -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
        512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
        -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
        -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
        -640/81.,     80/81.,   3968/81.,   -496/81.,    152/81.,   -424/81.,
        -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
        -1792/81.,    224/81.,   3968/81.,   -496/81.,   -352/81.,  -1072/81.,
        -1072/81.,   3536/81.,   4040/81.,  -1072/81.,   -928/81.,   3536/81.,
        -4096/81.,    512/81.,  32768/81.,  -4096/81.,   5120/81.,    512/81.,
        512/81.,    -64/81.,   5120/81.,   -640/81.,   4544/81.,   3968/81.,
        -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
        -640/81.,    -64/81.,   3968/81.,  -1792/81.,   -640/81.,   3968/81.,
        -640/81.,     80/81.,   5120/81.,   -640/81.,    800/81.,     80/81.,
        3968/81.,   -496/81.,   4544/81.,   -568/81.,   4112/81.,   4040/81.,
        512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
        -424/81.,   -496/81.,   -352/81.,  -1072/81.,   -568/81.,    152/81.,
        -424/81.,   -352/81.,    800/81.,     80/81.,   4112/81.,   -496/81.,
        -352/81.,   -928/81.,   4112/81.,   -352/81.,   1088/81.,   -928/81.,
        80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
        -352/81.,    224/81.,   -928/81.,   3536/81.,   -496/81.,   -424/81.},
   // Row    51 : MG ordering  = [ 3, 1, 5, 2, 6, 7, 4 ]
   {    8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
        -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
        8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
        -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
        80/81.,    800/81.,   -496/81.,   4112/81.,   -424/81.,   -352/81.,
        8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
        224/81.,   -352/81.,   -496/81.,   -424/81.,   -928/81.,   3536/81.,
        -352/81.,   -928/81.,   4112/81.,   -352/81.,   1088/81.,   -928/81.,
        512/81.,   5120/81.,  -4096/81.,  32768/81.,    512/81.,  -4096/81.,
        -64/81.,   -640/81.,   -640/81.,   3968/81.,   3968/81.,  -1792/81.,
        512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
        -64/81.,    512/81.,   4544/81.,   3968/81.,   5120/81.,   -640/81.,
        80/81.,    800/81.,   -640/81.,   5120/81.,     80/81.,   -640/81.,
        -496/81.,   -424/81.,   -568/81.,    152/81.,   -352/81.,  -1072/81.,
        -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
        -496/81.,   3968/81.,   4112/81.,   4040/81.,   4544/81.,   -568/81.,
        152/81.,   -424/81.,     80/81.,   -640/81.,   -496/81.,   3968/81.,
        -1072/81.,   3536/81.,   4040/81.,  -1072/81.,   -928/81.,   3536/81.,
        8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
        224/81.,  -1792/81.,   -352/81.,  -1072/81.,   3968/81.,   -496/81.},
   // Row    52 : MG ordering  = [ 3, 1, 5, 2, 7, 4, 6 ]
   {    8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
        -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
        -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
        -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
        80/81.,   -640/81.,    152/81.,   -424/81.,   3968/81.,   -496/81.,
        8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
        -1072/81.,   3536/81.,   4040/81.,  -1072/81.,   -928/81.,   3536/81.,
        -1792/81.,    224/81.,   3968/81.,   -496/81.,   -352/81.,  -1072/81.,
        512/81.,  -4096/81.,   5120/81.,    512/81.,  32768/81.,  -4096/81.,
        -64/81.,    512/81.,   4544/81.,   3968/81.,   5120/81.,   -640/81.,
        -640/81.,    -64/81.,   3968/81.,  -1792/81.,   -640/81.,   3968/81.,
        -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
        -424/81.,   -352/81.,    800/81.,     80/81.,   4112/81.,   -496/81.,
        -352/81.,   -928/81.,   4112/81.,   -352/81.,   1088/81.,   -928/81.,
        80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
        -352/81.,    224/81.,   -928/81.,   3536/81.,   -496/81.,   -424/81.,
        -640/81.,     80/81.,   5120/81.,   -640/81.,    800/81.,     80/81.,
        3968/81.,   -496/81.,   4544/81.,   -568/81.,   4112/81.,   4040/81.,
        512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
        -424/81.,   -496/81.,   -352/81.,  -1072/81.,   -568/81.,    152/81.},
   // Row    53 : MG ordering  = [ 3, 1, 5, 2, 7, 6, 4 ]
   {     80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
         -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
         -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
         8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
         800/81.,     80/81.,   -424/81.,   -352/81.,   -496/81.,   4112/81.,
         80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
         -352/81.,   -928/81.,   4112/81.,   -352/81.,   1088/81.,   -928/81.,
         224/81.,   -352/81.,   -496/81.,   -424/81.,   -928/81.,   3536/81.,
         5120/81.,    512/81.,    512/81.,  -4096/81.,  -4096/81.,  32768/81.,
         -640/81.,    -64/81.,   3968/81.,  -1792/81.,   -640/81.,   3968/81.,
         -64/81.,    512/81.,   4544/81.,   3968/81.,   5120/81.,   -640/81.,
         512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
         152/81.,   -424/81.,     80/81.,   -640/81.,   -496/81.,   3968/81.,
         -1072/81.,   3536/81.,   4040/81.,  -1072/81.,   -928/81.,   3536/81.,
         8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
         224/81.,  -1792/81.,   -352/81.,  -1072/81.,   3968/81.,   -496/81.,
         80/81.,    800/81.,   -640/81.,   5120/81.,     80/81.,   -640/81.,
         -496/81.,   -424/81.,   -568/81.,    152/81.,   -352/81.,  -1072/81.,
         -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
         -496/81.,   3968/81.,   4112/81.,   4040/81.,   4544/81.,   -568/81.},
   // Row    54 : MG ordering  = [ 3, 1, 5, 4, 2, 6, 7 ]
   {   5120/81.,   -640/81.,   -640/81.,     80/81.,     80/81.,    800/81.,
       512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
       3968/81.,   -496/81.,   4544/81.,   -568/81.,   4112/81.,   4040/81.,
       -496/81.,   -424/81.,   -568/81.,    152/81.,   -352/81.,  -1072/81.,
       512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
       -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
       -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
       80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
       -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
       32768/81.,  -4096/81.,  -4096/81.,    512/81.,    512/81.,   5120/81.,
       5120/81.,   -640/81.,    512/81.,    -64/81.,   3968/81.,   4544/81.,
       -640/81.,   3968/81.,    -64/81.,   -640/81.,  -1792/81.,   3968/81.,
       -1792/81.,    224/81.,   3968/81.,   -496/81.,   -352/81.,  -1072/81.,
       3968/81.,   -496/81.,   -640/81.,     80/81.,   -424/81.,    152/81.,
       -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
       -928/81.,   3536/81.,   3536/81.,  -1072/81.,  -1072/81.,   4040/81.,
       224/81.,   -352/81.,   -496/81.,   -424/81.,   -928/81.,   3536/81.,
       -496/81.,   4112/81.,     80/81.,    800/81.,   -352/81.,   -424/81.,
       80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
       1088/81.,   -928/81.,   -928/81.,   -352/81.,   -352/81.,   4112/81.},
   // Row    55 : MG ordering  = [ 3, 1, 5, 4, 2, 7, 6 ]
   {   -640/81.,   5120/81.,     80/81.,    800/81.,   -640/81.,     80/81.,
       -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
       -496/81.,   -424/81.,   -568/81.,    152/81.,   -352/81.,  -1072/81.,
       3968/81.,   -496/81.,   4544/81.,   -568/81.,   4112/81.,   4040/81.,
       -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
       512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
       80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
       -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
       512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
       -4096/81.,  32768/81.,    512/81.,   5120/81.,  -4096/81.,    512/81.,
       -640/81.,   3968/81.,    -64/81.,   -640/81.,  -1792/81.,   3968/81.,
       5120/81.,   -640/81.,    512/81.,    -64/81.,   3968/81.,   4544/81.,
       224/81.,   -352/81.,   -496/81.,   -424/81.,   -928/81.,   3536/81.,
       -496/81.,   4112/81.,     80/81.,    800/81.,   -352/81.,   -424/81.,
       80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
       1088/81.,   -928/81.,   -928/81.,   -352/81.,   -352/81.,   4112/81.,
       -1792/81.,    224/81.,   3968/81.,   -496/81.,   -352/81.,  -1072/81.,
       3968/81.,   -496/81.,   -640/81.,     80/81.,   -424/81.,    152/81.,
       -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
       -928/81.,   3536/81.,   3536/81.,  -1072/81.,  -1072/81.,   4040/81.},
   // Row    56 : MG ordering  = [ 3, 1, 5, 4, 6, 2, 7 ]
   {   -640/81.,     80/81.,   3968/81.,   -496/81.,    152/81.,   -424/81.,
       -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
       -1792/81.,    224/81.,   3968/81.,   -496/81.,   -352/81.,  -1072/81.,
       -1072/81.,   3536/81.,   4040/81.,  -1072/81.,   -928/81.,   3536/81.,
       -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
       512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       512/81.,    -64/81.,   5120/81.,   -640/81.,   4544/81.,   3968/81.,
       -4096/81.,    512/81.,  32768/81.,  -4096/81.,   5120/81.,    512/81.,
       512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
       3968/81.,  -1792/81.,   -640/81.,    -64/81.,   3968/81.,   -640/81.,
       3968/81.,   -496/81.,   4544/81.,   -568/81.,   4112/81.,   4040/81.,
       -640/81.,     80/81.,   5120/81.,   -640/81.,    800/81.,     80/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       -352/81.,  -1072/81.,   -424/81.,   -496/81.,    152/81.,   -568/81.,
       -352/81.,   -928/81.,   4112/81.,   -352/81.,   1088/81.,   -928/81.,
       -424/81.,   -352/81.,    800/81.,     80/81.,   4112/81.,   -496/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       -928/81.,   3536/81.,   -352/81.,    224/81.,   -424/81.,   -496/81.},
   // Row    57 : MG ordering  = [ 3, 1, 5, 4, 6, 7, 2 ]
   {     80/81.,    800/81.,   -496/81.,   4112/81.,   -424/81.,   -352/81.,
         8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
         224/81.,   -352/81.,   -496/81.,   -424/81.,   -928/81.,   3536/81.,
         -352/81.,   -928/81.,   4112/81.,   -352/81.,   1088/81.,   -928/81.,
         8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
         -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
         8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
         -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
         -64/81.,   -640/81.,   -640/81.,   3968/81.,   3968/81.,  -1792/81.,
         512/81.,   5120/81.,  -4096/81.,  32768/81.,    512/81.,  -4096/81.,
         -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
         4544/81.,   3968/81.,    -64/81.,    512/81.,   -640/81.,   5120/81.,
         -496/81.,   -424/81.,   -568/81.,    152/81.,   -352/81.,  -1072/81.,
         80/81.,    800/81.,   -640/81.,   5120/81.,     80/81.,   -640/81.,
         8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
         4112/81.,   4040/81.,   -496/81.,   3968/81.,   -568/81.,   4544/81.,
         -1072/81.,   3536/81.,   4040/81.,  -1072/81.,   -928/81.,   3536/81.,
         152/81.,   -424/81.,     80/81.,   -640/81.,   -496/81.,   3968/81.,
         -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
         -352/81.,  -1072/81.,    224/81.,  -1792/81.,   -496/81.,   3968/81.},
   // Row    58 : MG ordering  = [ 3, 1, 5, 4, 7, 2, 6 ]
   {     80/81.,   -640/81.,    152/81.,   -424/81.,   3968/81.,   -496/81.,
         8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
         -1072/81.,   3536/81.,   4040/81.,  -1072/81.,   -928/81.,   3536/81.,
         -1792/81.,    224/81.,   3968/81.,   -496/81.,   -352/81.,  -1072/81.,
         8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
         -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
         -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
         -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
         -64/81.,    512/81.,   4544/81.,   3968/81.,   5120/81.,   -640/81.,
         512/81.,  -4096/81.,   5120/81.,    512/81.,  32768/81.,  -4096/81.,
         3968/81.,  -1792/81.,   -640/81.,    -64/81.,   3968/81.,   -640/81.,
         512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
         -352/81.,   -928/81.,   4112/81.,   -352/81.,   1088/81.,   -928/81.,
         -424/81.,   -352/81.,    800/81.,     80/81.,   4112/81.,   -496/81.,
         -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
         -928/81.,   3536/81.,   -352/81.,    224/81.,   -424/81.,   -496/81.,
         3968/81.,   -496/81.,   4544/81.,   -568/81.,   4112/81.,   4040/81.,
         -640/81.,     80/81.,   5120/81.,   -640/81.,    800/81.,     80/81.,
         -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
         -352/81.,  -1072/81.,   -424/81.,   -496/81.,    152/81.,   -568/81.},
   // Row    59 : MG ordering  = [ 3, 1, 5, 4, 7, 6, 2 ]
   {    800/81.,     80/81.,   -424/81.,   -352/81.,   -496/81.,   4112/81.,
        80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
        -352/81.,   -928/81.,   4112/81.,   -352/81.,   1088/81.,   -928/81.,
        224/81.,   -352/81.,   -496/81.,   -424/81.,   -928/81.,   3536/81.,
        80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
        -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
        -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
        8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
        -640/81.,    -64/81.,   3968/81.,  -1792/81.,   -640/81.,   3968/81.,
        5120/81.,    512/81.,    512/81.,  -4096/81.,  -4096/81.,  32768/81.,
        4544/81.,   3968/81.,    -64/81.,    512/81.,   -640/81.,   5120/81.,
        -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
        -1072/81.,   3536/81.,   4040/81.,  -1072/81.,   -928/81.,   3536/81.,
        152/81.,   -424/81.,     80/81.,   -640/81.,   -496/81.,   3968/81.,
        -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
        -352/81.,  -1072/81.,    224/81.,  -1792/81.,   -496/81.,   3968/81.,
        -496/81.,   -424/81.,   -568/81.,    152/81.,   -352/81.,  -1072/81.,
        80/81.,    800/81.,   -640/81.,   5120/81.,     80/81.,   -640/81.,
        8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
        4112/81.,   4040/81.,   -496/81.,   3968/81.,   -568/81.,   4544/81.},
   // Row    60 : MG ordering  = [ 3, 1, 5, 6, 2, 4, 7 ]
   {   4544/81.,   -568/81.,   3968/81.,   -496/81.,   4040/81.,   4112/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       -640/81.,     80/81.,   5120/81.,   -640/81.,    800/81.,     80/81.,
       -1072/81.,   -352/81.,    152/81.,   -568/81.,   -424/81.,   -496/81.,
       3968/81.,   -496/81.,  -1792/81.,    224/81.,  -1072/81.,   -352/81.,
       -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
       3968/81.,   -496/81.,   -640/81.,     80/81.,   -424/81.,    152/81.,
       3536/81.,   -928/81.,  -1072/81.,   4040/81.,   3536/81.,  -1072/81.,
       512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
       5120/81.,   -640/81.,    512/81.,    -64/81.,   3968/81.,   4544/81.,
       32768/81.,  -4096/81.,  -4096/81.,    512/81.,    512/81.,   5120/81.,
       3968/81.,   -640/81.,  -1792/81.,   3968/81.,    -64/81.,   -640/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
       -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       3536/81.,   -928/81.,   -424/81.,   -496/81.,   -352/81.,    224/81.,
       -928/81.,   1088/81.,   -352/81.,   4112/81.,   -928/81.,   -352/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       4112/81.,   -496/81.,   -352/81.,   -424/81.,     80/81.,    800/81.},
   // Row    61 : MG ordering  = [ 3, 1, 5, 6, 2, 7, 4 ]
   {   -568/81.,    152/81.,   -496/81.,   -424/81.,  -1072/81.,   -352/81.,
       8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
       80/81.,    800/81.,   -640/81.,   5120/81.,     80/81.,   -640/81.,
       4040/81.,   4112/81.,   -568/81.,   4544/81.,   -496/81.,   3968/81.,
       -496/81.,   -424/81.,    224/81.,   -352/81.,   3536/81.,   -928/81.,
       80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
       -496/81.,   4112/81.,     80/81.,    800/81.,   -352/81.,   -424/81.,
       -928/81.,   1088/81.,   -352/81.,   4112/81.,   -928/81.,   -352/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
       -640/81.,   3968/81.,    -64/81.,   -640/81.,  -1792/81.,   3968/81.,
       -4096/81.,  32768/81.,    512/81.,   5120/81.,  -4096/81.,    512/81.,
       -640/81.,   5120/81.,   3968/81.,   4544/81.,    512/81.,    -64/81.,
       8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
       80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
       512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
       80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
       -1072/81.,   -352/81.,   -496/81.,   3968/81.,    224/81.,  -1792/81.,
       3536/81.,   -928/81.,  -1072/81.,   4040/81.,   3536/81.,  -1072/81.,
       80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
       -496/81.,   3968/81.,   -424/81.,    152/81.,   -640/81.,     80/81.},
   // Row    62 : MG ordering  = [ 3, 1, 5, 6, 4, 2, 7 ]
   {   3968/81.,   -496/81.,  -1792/81.,    224/81.,  -1072/81.,   -352/81.,
       -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
       3968/81.,   -496/81.,   -640/81.,     80/81.,   -424/81.,    152/81.,
       3536/81.,   -928/81.,  -1072/81.,   4040/81.,   3536/81.,  -1072/81.,
       4544/81.,   -568/81.,   3968/81.,   -496/81.,   4040/81.,   4112/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       -640/81.,     80/81.,   5120/81.,   -640/81.,    800/81.,     80/81.,
       -1072/81.,   -352/81.,    152/81.,   -568/81.,   -424/81.,   -496/81.,
       5120/81.,   -640/81.,    512/81.,    -64/81.,   3968/81.,   4544/81.,
       512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
       -4096/81.,    512/81.,  32768/81.,  -4096/81.,   5120/81.,    512/81.,
       -1792/81.,   3968/81.,   3968/81.,   -640/81.,   -640/81.,    -64/81.,
       -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       -928/81.,   1088/81.,   -352/81.,   4112/81.,   -928/81.,   -352/81.,
       3536/81.,   -928/81.,   -424/81.,   -496/81.,   -352/81.,    224/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       -352/81.,   -424/81.,   4112/81.,   -496/81.,    800/81.,     80/81.},
   // Row    63 : MG ordering  = [ 3, 1, 5, 6, 4, 7, 2 ]
   {   -496/81.,   -424/81.,    224/81.,   -352/81.,   3536/81.,   -928/81.,
       80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
       -496/81.,   4112/81.,     80/81.,    800/81.,   -352/81.,   -424/81.,
       -928/81.,   1088/81.,   -352/81.,   4112/81.,   -928/81.,   -352/81.,
       -568/81.,    152/81.,   -496/81.,   -424/81.,  -1072/81.,   -352/81.,
       8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
       80/81.,    800/81.,   -640/81.,   5120/81.,     80/81.,   -640/81.,
       4040/81.,   4112/81.,   -568/81.,   4544/81.,   -496/81.,   3968/81.,
       -640/81.,   3968/81.,    -64/81.,   -640/81.,  -1792/81.,   3968/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
       512/81.,   5120/81.,  -4096/81.,  32768/81.,    512/81.,  -4096/81.,
       3968/81.,   4544/81.,   -640/81.,   5120/81.,    -64/81.,    512/81.,
       80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
       8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       3536/81.,   -928/81.,  -1072/81.,   4040/81.,   3536/81.,  -1072/81.,
       -1072/81.,   -352/81.,   -496/81.,   3968/81.,    224/81.,  -1792/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       -424/81.,    152/81.,   -496/81.,   3968/81.,     80/81.,   -640/81.},
   // Row    64 : MG ordering  = [ 3, 1, 5, 6, 7, 2, 4 ]
   {   4040/81.,  -1072/81.,  -1072/81.,   3536/81.,   3536/81.,   -928/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       152/81.,   -424/81.,     80/81.,   -640/81.,   -496/81.,   3968/81.,
       -1072/81.,   -352/81.,   -496/81.,   3968/81.,    224/81.,  -1792/81.,
       4112/81.,   -352/81.,   -352/81.,   -928/81.,   -928/81.,   1088/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       -424/81.,   -352/81.,    800/81.,     80/81.,   4112/81.,   -496/81.,
       3536/81.,   -928/81.,   -424/81.,   -496/81.,   -352/81.,    224/81.,
       4544/81.,   3968/81.,    -64/81.,    512/81.,   -640/81.,   5120/81.,
       3968/81.,  -1792/81.,   -640/81.,    -64/81.,   3968/81.,   -640/81.,
       512/81.,  -4096/81.,   5120/81.,    512/81.,  32768/81.,  -4096/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       4040/81.,   4112/81.,   -568/81.,   4544/81.,   -496/81.,   3968/81.,
       -1072/81.,   -352/81.,    152/81.,   -568/81.,   -424/81.,   -496/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       80/81.,   -640/81.,    800/81.,     80/81.,   5120/81.,   -640/81.},
   // Row    65 : MG ordering  = [ 3, 1, 5, 6, 7, 4, 2 ]
   {   4112/81.,   -352/81.,   -352/81.,   -928/81.,   -928/81.,   1088/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       -424/81.,   -352/81.,    800/81.,     80/81.,   4112/81.,   -496/81.,
       3536/81.,   -928/81.,   -424/81.,   -496/81.,   -352/81.,    224/81.,
       4040/81.,  -1072/81.,  -1072/81.,   3536/81.,   3536/81.,   -928/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       152/81.,   -424/81.,     80/81.,   -640/81.,   -496/81.,   3968/81.,
       -1072/81.,   -352/81.,   -496/81.,   3968/81.,    224/81.,  -1792/81.,
       3968/81.,  -1792/81.,   -640/81.,    -64/81.,   3968/81.,   -640/81.,
       4544/81.,   3968/81.,    -64/81.,    512/81.,   -640/81.,   5120/81.,
       5120/81.,    512/81.,    512/81.,  -4096/81.,  -4096/81.,  32768/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       -1072/81.,   -352/81.,    152/81.,   -568/81.,   -424/81.,   -496/81.,
       4040/81.,   4112/81.,   -568/81.,   4544/81.,   -496/81.,   3968/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       800/81.,     80/81.,     80/81.,   -640/81.,   -640/81.,   5120/81.},
   // Row    66 : MG ordering  = [ 3, 1, 5, 7, 2, 4, 6 ]
   {   -568/81.,   4544/81.,   4040/81.,   4112/81.,   3968/81.,   -496/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       -1072/81.,   -352/81.,    152/81.,   -568/81.,   -424/81.,   -496/81.,
       -640/81.,     80/81.,   5120/81.,   -640/81.,    800/81.,     80/81.,
       -496/81.,   3968/81.,  -1072/81.,   -352/81.,  -1792/81.,    224/81.,
       80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
       3536/81.,   -928/81.,  -1072/81.,   4040/81.,   3536/81.,  -1072/81.,
       3968/81.,   -496/81.,   -640/81.,     80/81.,   -424/81.,    152/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       -640/81.,   5120/81.,   3968/81.,   4544/81.,    512/81.,    -64/81.,
       3968/81.,   -640/81.,  -1792/81.,   3968/81.,    -64/81.,   -640/81.,
       32768/81.,  -4096/81.,  -4096/81.,    512/81.,    512/81.,   5120/81.,
       3536/81.,   -928/81.,   -424/81.,   -496/81.,   -352/81.,    224/81.,
       -928/81.,   1088/81.,   -352/81.,   4112/81.,   -928/81.,   -352/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       4112/81.,   -496/81.,   -352/81.,   -424/81.,     80/81.,    800/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
       -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.},
   // Row    67 : MG ordering  = [ 3, 1, 5, 7, 2, 6, 4 ]
   {    152/81.,   -568/81.,  -1072/81.,   -352/81.,   -496/81.,   -424/81.,
        80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
        4040/81.,   4112/81.,   -568/81.,   4544/81.,   -496/81.,   3968/81.,
        80/81.,    800/81.,   -640/81.,   5120/81.,     80/81.,   -640/81.,
        -424/81.,   -496/81.,   3536/81.,   -928/81.,    224/81.,   -352/81.,
        -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
        -928/81.,   1088/81.,   -352/81.,   4112/81.,   -928/81.,   -352/81.,
        -496/81.,   4112/81.,     80/81.,    800/81.,   -352/81.,   -424/81.,
        -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
        3968/81.,   -640/81.,  -1792/81.,   3968/81.,    -64/81.,   -640/81.,
        -640/81.,   5120/81.,   3968/81.,   4544/81.,    512/81.,    -64/81.,
        -4096/81.,  32768/81.,    512/81.,   5120/81.,  -4096/81.,    512/81.,
        -1072/81.,   -352/81.,   -496/81.,   3968/81.,    224/81.,  -1792/81.,
        3536/81.,   -928/81.,  -1072/81.,   4040/81.,   3536/81.,  -1072/81.,
        80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
        -496/81.,   3968/81.,   -424/81.,    152/81.,   -640/81.,     80/81.,
        8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
        80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
        512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
        80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.},
   // Row    68 : MG ordering  = [ 3, 1, 5, 7, 4, 2, 6 ]
   {   -496/81.,   3968/81.,  -1072/81.,   -352/81.,  -1792/81.,    224/81.,
       80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
       3536/81.,   -928/81.,  -1072/81.,   4040/81.,   3536/81.,  -1072/81.,
       3968/81.,   -496/81.,   -640/81.,     80/81.,   -424/81.,    152/81.,
       -568/81.,   4544/81.,   4040/81.,   4112/81.,   3968/81.,   -496/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       -1072/81.,   -352/81.,    152/81.,   -568/81.,   -424/81.,   -496/81.,
       -640/81.,     80/81.,   5120/81.,   -640/81.,    800/81.,     80/81.,
       -640/81.,   5120/81.,   3968/81.,   4544/81.,    512/81.,    -64/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       -1792/81.,   3968/81.,   3968/81.,   -640/81.,   -640/81.,    -64/81.,
       -4096/81.,    512/81.,  32768/81.,  -4096/81.,   5120/81.,    512/81.,
       -928/81.,   1088/81.,   -352/81.,   4112/81.,   -928/81.,   -352/81.,
       3536/81.,   -928/81.,   -424/81.,   -496/81.,   -352/81.,    224/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       -352/81.,   -424/81.,   4112/81.,   -496/81.,    800/81.,     80/81.,
       -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.},
   // Row    69 : MG ordering  = [ 3, 1, 5, 7, 4, 6, 2 ]
   {   -424/81.,   -496/81.,   3536/81.,   -928/81.,    224/81.,   -352/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       -928/81.,   1088/81.,   -352/81.,   4112/81.,   -928/81.,   -352/81.,
       -496/81.,   4112/81.,     80/81.,    800/81.,   -352/81.,   -424/81.,
       152/81.,   -568/81.,  -1072/81.,   -352/81.,   -496/81.,   -424/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       4040/81.,   4112/81.,   -568/81.,   4544/81.,   -496/81.,   3968/81.,
       80/81.,    800/81.,   -640/81.,   5120/81.,     80/81.,   -640/81.,
       3968/81.,   -640/81.,  -1792/81.,   3968/81.,    -64/81.,   -640/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       3968/81.,   4544/81.,   -640/81.,   5120/81.,    -64/81.,    512/81.,
       512/81.,   5120/81.,  -4096/81.,  32768/81.,    512/81.,  -4096/81.,
       3536/81.,   -928/81.,  -1072/81.,   4040/81.,   3536/81.,  -1072/81.,
       -1072/81.,   -352/81.,   -496/81.,   3968/81.,    224/81.,  -1792/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       -424/81.,    152/81.,   -496/81.,   3968/81.,     80/81.,   -640/81.,
       80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
       8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.},
   // Row    70 : MG ordering  = [ 3, 1, 5, 7, 6, 2, 4 ]
   {  -1072/81.,   4040/81.,   3536/81.,   -928/81.,  -1072/81.,   3536/81.,
      -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
      -1072/81.,   -352/81.,   -496/81.,   3968/81.,    224/81.,  -1792/81.,
      152/81.,   -424/81.,     80/81.,   -640/81.,   -496/81.,   3968/81.,
      -352/81.,   4112/81.,   -928/81.,   1088/81.,   -352/81.,   -928/81.,
      224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
      3536/81.,   -928/81.,   -424/81.,   -496/81.,   -352/81.,    224/81.,
      -424/81.,   -352/81.,    800/81.,     80/81.,   4112/81.,   -496/81.,
      3968/81.,   4544/81.,   -640/81.,   5120/81.,    -64/81.,    512/81.,
      -1792/81.,   3968/81.,   3968/81.,   -640/81.,   -640/81.,    -64/81.,
      -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
      512/81.,  -4096/81.,   5120/81.,    512/81.,  32768/81.,  -4096/81.,
      4040/81.,   4112/81.,   -568/81.,   4544/81.,   -496/81.,   3968/81.,
      -1072/81.,   -352/81.,    152/81.,   -568/81.,   -424/81.,   -496/81.,
      8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
      80/81.,   -640/81.,    800/81.,     80/81.,   5120/81.,   -640/81.,
      -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
      -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
      -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
      8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.},
   // Row    71 : MG ordering  = [ 3, 1, 5, 7, 6, 4, 2 ]
   {   -352/81.,   4112/81.,   -928/81.,   1088/81.,   -352/81.,   -928/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       3536/81.,   -928/81.,   -424/81.,   -496/81.,   -352/81.,    224/81.,
       -424/81.,   -352/81.,    800/81.,     80/81.,   4112/81.,   -496/81.,
       -1072/81.,   4040/81.,   3536/81.,   -928/81.,  -1072/81.,   3536/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       -1072/81.,   -352/81.,   -496/81.,   3968/81.,    224/81.,  -1792/81.,
       152/81.,   -424/81.,     80/81.,   -640/81.,   -496/81.,   3968/81.,
       -1792/81.,   3968/81.,   3968/81.,   -640/81.,   -640/81.,    -64/81.,
       3968/81.,   4544/81.,   -640/81.,   5120/81.,    -64/81.,    512/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       5120/81.,    512/81.,    512/81.,  -4096/81.,  -4096/81.,  32768/81.,
       -1072/81.,   -352/81.,    152/81.,   -568/81.,   -424/81.,   -496/81.,
       4040/81.,   4112/81.,   -568/81.,   4544/81.,   -496/81.,   3968/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       800/81.,     80/81.,     80/81.,   -640/81.,   -640/81.,   5120/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.},
   // Row    72 : MG ordering  = [ 3, 1, 6, 2, 4, 5, 7 ]
   {    -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
        -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
        -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
        -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
        -640/81.,     80/81.,   5120/81.,   -640/81.,    800/81.,     80/81.,
        3968/81.,   -496/81.,   4544/81.,   -568/81.,   4112/81.,   4040/81.,
        512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
        -424/81.,   -496/81.,   -352/81.,  -1072/81.,   -568/81.,    152/81.,
        3968/81.,   -496/81.,   -640/81.,     80/81.,   -424/81.,    152/81.,
        -1792/81.,    224/81.,   3968/81.,   -496/81.,   -352/81.,  -1072/81.,
        -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
        3536/81.,  -1072/81.,   -928/81.,   3536/81.,   4040/81.,  -1072/81.,
        32768/81.,  -4096/81.,  -4096/81.,    512/81.,    512/81.,   5120/81.,
        -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
        512/81.,    -64/81.,   5120/81.,   -640/81.,   4544/81.,   3968/81.,
        -64/81.,   -640/81.,   -640/81.,   3968/81.,   3968/81.,  -1792/81.,
        4112/81.,   -496/81.,   -352/81.,   -424/81.,     80/81.,    800/81.,
        -352/81.,    224/81.,   -928/81.,   3536/81.,   -496/81.,   -424/81.,
        -928/81.,   -352/81.,   1088/81.,   -928/81.,   4112/81.,   -352/81.,
        8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.},
   // Row    73 : MG ordering  = [ 3, 1, 6, 2, 4, 7, 5 ]
   {    8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
        80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
        512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
        80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
        80/81.,    800/81.,   -640/81.,   5120/81.,     80/81.,   -640/81.,
        -496/81.,   -424/81.,   -568/81.,    152/81.,   -352/81.,  -1072/81.,
        -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
        -496/81.,   3968/81.,   4112/81.,   4040/81.,   4544/81.,   -568/81.,
        -496/81.,   4112/81.,     80/81.,    800/81.,   -352/81.,   -424/81.,
        224/81.,   -352/81.,   -496/81.,   -424/81.,   -928/81.,   3536/81.,
        8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
        -928/81.,   -352/81.,   1088/81.,   -928/81.,   4112/81.,   -352/81.,
        -4096/81.,  32768/81.,    512/81.,   5120/81.,  -4096/81.,    512/81.,
        512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
        -64/81.,   -640/81.,   -640/81.,   3968/81.,   3968/81.,  -1792/81.,
        512/81.,    -64/81.,   5120/81.,   -640/81.,   4544/81.,   3968/81.,
        -496/81.,   3968/81.,   -424/81.,    152/81.,   -640/81.,     80/81.,
        224/81.,  -1792/81.,   -352/81.,  -1072/81.,   3968/81.,   -496/81.,
        3536/81.,  -1072/81.,   -928/81.,   3536/81.,   4040/81.,  -1072/81.,
        -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.},
   // Row    74 : MG ordering  = [ 3, 1, 6, 2, 5, 4, 7 ]
   {   -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       3968/81.,   -496/81.,   -640/81.,     80/81.,   -424/81.,    152/81.,
       -1792/81.,    224/81.,   3968/81.,   -496/81.,   -352/81.,  -1072/81.,
       -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
       3536/81.,  -1072/81.,   -928/81.,   3536/81.,   4040/81.,  -1072/81.,
       -640/81.,     80/81.,   5120/81.,   -640/81.,    800/81.,     80/81.,
       3968/81.,   -496/81.,   4544/81.,   -568/81.,   4112/81.,   4040/81.,
       512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
       -424/81.,   -496/81.,   -352/81.,  -1072/81.,   -568/81.,    152/81.,
       -4096/81.,    512/81.,  32768/81.,  -4096/81.,   5120/81.,    512/81.,
       512/81.,    -64/81.,   5120/81.,   -640/81.,   4544/81.,   3968/81.,
       -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
       -640/81.,    -64/81.,   3968/81.,  -1792/81.,   -640/81.,   3968/81.,
       -352/81.,   -424/81.,   4112/81.,   -496/81.,    800/81.,     80/81.,
       -928/81.,   -352/81.,   1088/81.,   -928/81.,   4112/81.,   -352/81.,
       -352/81.,    224/81.,   -928/81.,   3536/81.,   -496/81.,   -424/81.,
       80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.},
   // Row    75 : MG ordering  = [ 3, 1, 6, 2, 5, 7, 4 ]
   {     80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
         8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
         -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
         -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
         -496/81.,   4112/81.,     80/81.,    800/81.,   -352/81.,   -424/81.,
         224/81.,   -352/81.,   -496/81.,   -424/81.,   -928/81.,   3536/81.,
         8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
         -928/81.,   -352/81.,   1088/81.,   -928/81.,   4112/81.,   -352/81.,
         80/81.,    800/81.,   -640/81.,   5120/81.,     80/81.,   -640/81.,
         -496/81.,   -424/81.,   -568/81.,    152/81.,   -352/81.,  -1072/81.,
         -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
         -496/81.,   3968/81.,   4112/81.,   4040/81.,   4544/81.,   -568/81.,
         512/81.,   5120/81.,  -4096/81.,  32768/81.,    512/81.,  -4096/81.,
         -64/81.,   -640/81.,   -640/81.,   3968/81.,   3968/81.,  -1792/81.,
         512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
         -64/81.,    512/81.,   4544/81.,   3968/81.,   5120/81.,   -640/81.,
         -424/81.,    152/81.,   -496/81.,   3968/81.,     80/81.,   -640/81.,
         3536/81.,  -1072/81.,   -928/81.,   3536/81.,   4040/81.,  -1072/81.,
         224/81.,  -1792/81.,   -352/81.,  -1072/81.,   3968/81.,   -496/81.,
         8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.},
   // Row    76 : MG ordering  = [ 3, 1, 6, 2, 7, 4, 5 ]
   {   -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       152/81.,   -424/81.,     80/81.,   -640/81.,   -496/81.,   3968/81.,
       -1072/81.,   3536/81.,   4040/81.,  -1072/81.,   -928/81.,   3536/81.,
       8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
       224/81.,  -1792/81.,   -352/81.,  -1072/81.,   3968/81.,   -496/81.,
       -424/81.,   -352/81.,    800/81.,     80/81.,   4112/81.,   -496/81.,
       -352/81.,   -928/81.,   4112/81.,   -352/81.,   1088/81.,   -928/81.,
       80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
       -352/81.,    224/81.,   -928/81.,   3536/81.,   -496/81.,   -424/81.,
       512/81.,  -4096/81.,   5120/81.,    512/81.,  32768/81.,  -4096/81.,
       -64/81.,    512/81.,   4544/81.,   3968/81.,   5120/81.,   -640/81.,
       -640/81.,    -64/81.,   3968/81.,  -1792/81.,   -640/81.,   3968/81.,
       -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
       80/81.,   -640/81.,    800/81.,     80/81.,   5120/81.,   -640/81.,
       -496/81.,   3968/81.,   4112/81.,   4040/81.,   4544/81.,   -568/81.,
       -424/81.,   -496/81.,   -352/81.,  -1072/81.,   -568/81.,    152/81.,
       512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.},
   // Row    77 : MG ordering  = [ 3, 1, 6, 2, 7, 5, 4 ]
   {   -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       -424/81.,   -352/81.,    800/81.,     80/81.,   4112/81.,   -496/81.,
       -352/81.,   -928/81.,   4112/81.,   -352/81.,   1088/81.,   -928/81.,
       80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
       -352/81.,    224/81.,   -928/81.,   3536/81.,   -496/81.,   -424/81.,
       152/81.,   -424/81.,     80/81.,   -640/81.,   -496/81.,   3968/81.,
       -1072/81.,   3536/81.,   4040/81.,  -1072/81.,   -928/81.,   3536/81.,
       8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
       224/81.,  -1792/81.,   -352/81.,  -1072/81.,   3968/81.,   -496/81.,
       5120/81.,    512/81.,    512/81.,  -4096/81.,  -4096/81.,  32768/81.,
       -640/81.,    -64/81.,   3968/81.,  -1792/81.,   -640/81.,   3968/81.,
       -64/81.,    512/81.,   4544/81.,   3968/81.,   5120/81.,   -640/81.,
       512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
       800/81.,     80/81.,     80/81.,   -640/81.,   -640/81.,   5120/81.,
       -424/81.,   -496/81.,   -352/81.,  -1072/81.,   -568/81.,    152/81.,
       -496/81.,   3968/81.,   4112/81.,   4040/81.,   4544/81.,   -568/81.,
       -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.},
   // Row    78 : MG ordering  = [ 3, 1, 6, 4, 2, 5, 7 ]
   {   -640/81.,     80/81.,   5120/81.,   -640/81.,    800/81.,     80/81.,
       3968/81.,   -496/81.,   4544/81.,   -568/81.,   4112/81.,   4040/81.,
       512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
       -424/81.,   -496/81.,   -352/81.,  -1072/81.,   -568/81.,    152/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
       -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       -1792/81.,    224/81.,   3968/81.,   -496/81.,   -352/81.,  -1072/81.,
       3968/81.,   -496/81.,   -640/81.,     80/81.,   -424/81.,    152/81.,
       -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
       -928/81.,   3536/81.,   3536/81.,  -1072/81.,  -1072/81.,   4040/81.,
       -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
       32768/81.,  -4096/81.,  -4096/81.,    512/81.,    512/81.,   5120/81.,
       5120/81.,   -640/81.,    512/81.,    -64/81.,   3968/81.,   4544/81.,
       -640/81.,   3968/81.,    -64/81.,   -640/81.,  -1792/81.,   3968/81.,
       -352/81.,    224/81.,   -928/81.,   3536/81.,   -496/81.,   -424/81.,
       4112/81.,   -496/81.,   -352/81.,   -424/81.,     80/81.,    800/81.,
       1088/81.,   -928/81.,   -928/81.,   -352/81.,   -352/81.,   4112/81.,
       80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.},
   // Row    79 : MG ordering  = [ 3, 1, 6, 4, 2, 7, 5 ]
   {     80/81.,    800/81.,   -640/81.,   5120/81.,     80/81.,   -640/81.,
         -496/81.,   -424/81.,   -568/81.,    152/81.,   -352/81.,  -1072/81.,
         -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
         -496/81.,   3968/81.,   4112/81.,   4040/81.,   4544/81.,   -568/81.,
         8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
         80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
         512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
         80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
         224/81.,   -352/81.,   -496/81.,   -424/81.,   -928/81.,   3536/81.,
         -496/81.,   4112/81.,     80/81.,    800/81.,   -352/81.,   -424/81.,
         80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
         1088/81.,   -928/81.,   -928/81.,   -352/81.,   -352/81.,   4112/81.,
         512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
         -4096/81.,  32768/81.,    512/81.,   5120/81.,  -4096/81.,    512/81.,
         -640/81.,   3968/81.,    -64/81.,   -640/81.,  -1792/81.,   3968/81.,
         5120/81.,   -640/81.,    512/81.,    -64/81.,   3968/81.,   4544/81.,
         224/81.,  -1792/81.,   -352/81.,  -1072/81.,   3968/81.,   -496/81.,
         -496/81.,   3968/81.,   -424/81.,    152/81.,   -640/81.,     80/81.,
         -928/81.,   3536/81.,   3536/81.,  -1072/81.,  -1072/81.,   4040/81.,
         -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.},
   // Row    80 : MG ordering  = [ 3, 1, 6, 4, 5, 2, 7 ]
   {   3968/81.,   -496/81.,   -640/81.,     80/81.,   -424/81.,    152/81.,
       -1792/81.,    224/81.,   3968/81.,   -496/81.,   -352/81.,  -1072/81.,
       -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
       3536/81.,  -1072/81.,   -928/81.,   3536/81.,   4040/81.,  -1072/81.,
       -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       3968/81.,   -496/81.,   4544/81.,   -568/81.,   4112/81.,   4040/81.,
       -640/81.,     80/81.,   5120/81.,   -640/81.,    800/81.,     80/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       -352/81.,  -1072/81.,   -424/81.,   -496/81.,    152/81.,   -568/81.,
       512/81.,    -64/81.,   5120/81.,   -640/81.,   4544/81.,   3968/81.,
       -4096/81.,    512/81.,  32768/81.,  -4096/81.,   5120/81.,    512/81.,
       512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
       3968/81.,  -1792/81.,   -640/81.,    -64/81.,   3968/81.,   -640/81.,
       -928/81.,   -352/81.,   1088/81.,   -928/81.,   4112/81.,   -352/81.,
       -352/81.,   -424/81.,   4112/81.,   -496/81.,    800/81.,     80/81.,
       -928/81.,   3536/81.,   -352/81.,    224/81.,   -424/81.,   -496/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.},
   // Row    81 : MG ordering  = [ 3, 1, 6, 4, 5, 7, 2 ]
   {   -496/81.,   4112/81.,     80/81.,    800/81.,   -352/81.,   -424/81.,
       224/81.,   -352/81.,   -496/81.,   -424/81.,   -928/81.,   3536/81.,
       8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
       -928/81.,   -352/81.,   1088/81.,   -928/81.,   4112/81.,   -352/81.,
       80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
       8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       -496/81.,   -424/81.,   -568/81.,    152/81.,   -352/81.,  -1072/81.,
       80/81.,    800/81.,   -640/81.,   5120/81.,     80/81.,   -640/81.,
       8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
       4112/81.,   4040/81.,   -496/81.,   3968/81.,   -568/81.,   4544/81.,
       -64/81.,   -640/81.,   -640/81.,   3968/81.,   3968/81.,  -1792/81.,
       512/81.,   5120/81.,  -4096/81.,  32768/81.,    512/81.,  -4096/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
       4544/81.,   3968/81.,    -64/81.,    512/81.,   -640/81.,   5120/81.,
       3536/81.,  -1072/81.,   -928/81.,   3536/81.,   4040/81.,  -1072/81.,
       -424/81.,    152/81.,   -496/81.,   3968/81.,     80/81.,   -640/81.,
       -352/81.,  -1072/81.,    224/81.,  -1792/81.,   -496/81.,   3968/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.},
   // Row    82 : MG ordering  = [ 3, 1, 6, 4, 7, 2, 5 ]
   {    152/81.,   -424/81.,     80/81.,   -640/81.,   -496/81.,   3968/81.,
        -1072/81.,   3536/81.,   4040/81.,  -1072/81.,   -928/81.,   3536/81.,
        8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
        224/81.,  -1792/81.,   -352/81.,  -1072/81.,   3968/81.,   -496/81.,
        -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
        -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
        -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
        8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
        -352/81.,   -928/81.,   4112/81.,   -352/81.,   1088/81.,   -928/81.,
        -424/81.,   -352/81.,    800/81.,     80/81.,   4112/81.,   -496/81.,
        -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
        -928/81.,   3536/81.,   -352/81.,    224/81.,   -424/81.,   -496/81.,
        -64/81.,    512/81.,   4544/81.,   3968/81.,   5120/81.,   -640/81.,
        512/81.,  -4096/81.,   5120/81.,    512/81.,  32768/81.,  -4096/81.,
        3968/81.,  -1792/81.,   -640/81.,    -64/81.,   3968/81.,   -640/81.,
        512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
        -496/81.,   3968/81.,   4112/81.,   4040/81.,   4544/81.,   -568/81.,
        80/81.,   -640/81.,    800/81.,     80/81.,   5120/81.,   -640/81.,
        -352/81.,  -1072/81.,   -424/81.,   -496/81.,    152/81.,   -568/81.,
        -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.},
   // Row    83 : MG ordering  = [ 3, 1, 6, 4, 7, 5, 2 ]
   {   -424/81.,   -352/81.,    800/81.,     80/81.,   4112/81.,   -496/81.,
       -352/81.,   -928/81.,   4112/81.,   -352/81.,   1088/81.,   -928/81.,
       80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
       -352/81.,    224/81.,   -928/81.,   3536/81.,   -496/81.,   -424/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       -1072/81.,   3536/81.,   4040/81.,  -1072/81.,   -928/81.,   3536/81.,
       152/81.,   -424/81.,     80/81.,   -640/81.,   -496/81.,   3968/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       -352/81.,  -1072/81.,    224/81.,  -1792/81.,   -496/81.,   3968/81.,
       -640/81.,    -64/81.,   3968/81.,  -1792/81.,   -640/81.,   3968/81.,
       5120/81.,    512/81.,    512/81.,  -4096/81.,  -4096/81.,  32768/81.,
       4544/81.,   3968/81.,    -64/81.,    512/81.,   -640/81.,   5120/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
       -424/81.,   -496/81.,   -352/81.,  -1072/81.,   -568/81.,    152/81.,
       800/81.,     80/81.,     80/81.,   -640/81.,   -640/81.,   5120/81.,
       4112/81.,   4040/81.,   -496/81.,   3968/81.,   -568/81.,   4544/81.,
       8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.},
   // Row    84 : MG ordering  = [ 3, 1, 6, 5, 2, 4, 7 ]
   {   3968/81.,   -496/81.,   4544/81.,   -568/81.,   4112/81.,   4040/81.,
       -640/81.,     80/81.,   5120/81.,   -640/81.,    800/81.,     80/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       -352/81.,  -1072/81.,   -424/81.,   -496/81.,    152/81.,   -568/81.,
       -1792/81.,    224/81.,   3968/81.,   -496/81.,   -352/81.,  -1072/81.,
       3968/81.,   -496/81.,   -640/81.,     80/81.,   -424/81.,    152/81.,
       -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
       -928/81.,   3536/81.,   3536/81.,  -1072/81.,  -1072/81.,   4040/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
       -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
       5120/81.,   -640/81.,    512/81.,    -64/81.,   3968/81.,   4544/81.,
       32768/81.,  -4096/81.,  -4096/81.,    512/81.,    512/81.,   5120/81.,
       3968/81.,   -640/81.,  -1792/81.,   3968/81.,    -64/81.,   -640/81.,
       -928/81.,   3536/81.,   -352/81.,    224/81.,   -424/81.,   -496/81.,
       1088/81.,   -928/81.,   -928/81.,   -352/81.,   -352/81.,   4112/81.,
       4112/81.,   -496/81.,   -352/81.,   -424/81.,     80/81.,    800/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.},
   // Row    85 : MG ordering  = [ 3, 1, 6, 5, 2, 7, 4 ]
   {   -496/81.,   -424/81.,   -568/81.,    152/81.,   -352/81.,  -1072/81.,
       80/81.,    800/81.,   -640/81.,   5120/81.,     80/81.,   -640/81.,
       8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
       4112/81.,   4040/81.,   -496/81.,   3968/81.,   -568/81.,   4544/81.,
       224/81.,   -352/81.,   -496/81.,   -424/81.,   -928/81.,   3536/81.,
       -496/81.,   4112/81.,     80/81.,    800/81.,   -352/81.,   -424/81.,
       80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
       1088/81.,   -928/81.,   -928/81.,   -352/81.,   -352/81.,   4112/81.,
       8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
       80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
       512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
       80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
       -640/81.,   3968/81.,    -64/81.,   -640/81.,  -1792/81.,   3968/81.,
       -4096/81.,  32768/81.,    512/81.,   5120/81.,  -4096/81.,    512/81.,
       -640/81.,   5120/81.,   3968/81.,   4544/81.,    512/81.,    -64/81.,
       -352/81.,  -1072/81.,    224/81.,  -1792/81.,   -496/81.,   3968/81.,
       -928/81.,   3536/81.,   3536/81.,  -1072/81.,  -1072/81.,   4040/81.,
       -496/81.,   3968/81.,   -424/81.,    152/81.,   -640/81.,     80/81.,
       80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.},
   // Row    86 : MG ordering  = [ 3, 1, 6, 5, 4, 2, 7 ]
   {  -1792/81.,    224/81.,   3968/81.,   -496/81.,   -352/81.,  -1072/81.,
      3968/81.,   -496/81.,   -640/81.,     80/81.,   -424/81.,    152/81.,
      -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
      -928/81.,   3536/81.,   3536/81.,  -1072/81.,  -1072/81.,   4040/81.,
      3968/81.,   -496/81.,   4544/81.,   -568/81.,   4112/81.,   4040/81.,
      -640/81.,     80/81.,   5120/81.,   -640/81.,    800/81.,     80/81.,
      -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
      -352/81.,  -1072/81.,   -424/81.,   -496/81.,    152/81.,   -568/81.,
      -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
      -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
      512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
      224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
      5120/81.,   -640/81.,    512/81.,    -64/81.,   3968/81.,   4544/81.,
      512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
      -4096/81.,    512/81.,  32768/81.,  -4096/81.,   5120/81.,    512/81.,
      -1792/81.,   3968/81.,   3968/81.,   -640/81.,   -640/81.,    -64/81.,
      1088/81.,   -928/81.,   -928/81.,   -352/81.,   -352/81.,   4112/81.,
      -928/81.,   3536/81.,   -352/81.,    224/81.,   -424/81.,   -496/81.,
      -352/81.,   -424/81.,   4112/81.,   -496/81.,    800/81.,     80/81.,
      224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.},
   // Row    87 : MG ordering  = [ 3, 1, 6, 5, 4, 7, 2 ]
   {    224/81.,   -352/81.,   -496/81.,   -424/81.,   -928/81.,   3536/81.,
        -496/81.,   4112/81.,     80/81.,    800/81.,   -352/81.,   -424/81.,
        80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
        1088/81.,   -928/81.,   -928/81.,   -352/81.,   -352/81.,   4112/81.,
        -496/81.,   -424/81.,   -568/81.,    152/81.,   -352/81.,  -1072/81.,
        80/81.,    800/81.,   -640/81.,   5120/81.,     80/81.,   -640/81.,
        8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
        4112/81.,   4040/81.,   -496/81.,   3968/81.,   -568/81.,   4544/81.,
        80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
        8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
        -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
        -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
        -640/81.,   3968/81.,    -64/81.,   -640/81.,  -1792/81.,   3968/81.,
        -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
        512/81.,   5120/81.,  -4096/81.,  32768/81.,    512/81.,  -4096/81.,
        3968/81.,   4544/81.,   -640/81.,   5120/81.,    -64/81.,    512/81.,
        -928/81.,   3536/81.,   3536/81.,  -1072/81.,  -1072/81.,   4040/81.,
        -352/81.,  -1072/81.,    224/81.,  -1792/81.,   -496/81.,   3968/81.,
        -424/81.,    152/81.,   -496/81.,   3968/81.,     80/81.,   -640/81.,
        -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.},
   // Row    88 : MG ordering  = [ 3, 1, 6, 5, 7, 2, 4 ]
   {  -1072/81.,   3536/81.,   4040/81.,  -1072/81.,   -928/81.,   3536/81.,
      152/81.,   -424/81.,     80/81.,   -640/81.,   -496/81.,   3968/81.,
      -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
      -352/81.,  -1072/81.,    224/81.,  -1792/81.,   -496/81.,   3968/81.,
      -352/81.,   -928/81.,   4112/81.,   -352/81.,   1088/81.,   -928/81.,
      -424/81.,   -352/81.,    800/81.,     80/81.,   4112/81.,   -496/81.,
      -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
      -928/81.,   3536/81.,   -352/81.,    224/81.,   -424/81.,   -496/81.,
      -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
      -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
      -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
      8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
      4544/81.,   3968/81.,    -64/81.,    512/81.,   -640/81.,   5120/81.,
      3968/81.,  -1792/81.,   -640/81.,    -64/81.,   3968/81.,   -640/81.,
      512/81.,  -4096/81.,   5120/81.,    512/81.,  32768/81.,  -4096/81.,
      -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
      4112/81.,   4040/81.,   -496/81.,   3968/81.,   -568/81.,   4544/81.,
      -352/81.,  -1072/81.,   -424/81.,   -496/81.,    152/81.,   -568/81.,
      80/81.,   -640/81.,    800/81.,     80/81.,   5120/81.,   -640/81.,
      8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.},
   // Row    89 : MG ordering  = [ 3, 1, 6, 5, 7, 4, 2 ]
   {   -352/81.,   -928/81.,   4112/81.,   -352/81.,   1088/81.,   -928/81.,
       -424/81.,   -352/81.,    800/81.,     80/81.,   4112/81.,   -496/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       -928/81.,   3536/81.,   -352/81.,    224/81.,   -424/81.,   -496/81.,
       -1072/81.,   3536/81.,   4040/81.,  -1072/81.,   -928/81.,   3536/81.,
       152/81.,   -424/81.,     80/81.,   -640/81.,   -496/81.,   3968/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       -352/81.,  -1072/81.,    224/81.,  -1792/81.,   -496/81.,   3968/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       3968/81.,  -1792/81.,   -640/81.,    -64/81.,   3968/81.,   -640/81.,
       4544/81.,   3968/81.,    -64/81.,    512/81.,   -640/81.,   5120/81.,
       5120/81.,    512/81.,    512/81.,  -4096/81.,  -4096/81.,  32768/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       -352/81.,  -1072/81.,   -424/81.,   -496/81.,    152/81.,   -568/81.,
       4112/81.,   4040/81.,   -496/81.,   3968/81.,   -568/81.,   4544/81.,
       800/81.,     80/81.,     80/81.,   -640/81.,   -640/81.,   5120/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.},
   // Row    90 : MG ordering  = [ 3, 1, 6, 7, 2, 4, 5 ]
   {   4040/81.,   4112/81.,   -568/81.,   4544/81.,   -496/81.,   3968/81.,
       -1072/81.,   -352/81.,    152/81.,   -568/81.,   -424/81.,   -496/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       80/81.,   -640/81.,    800/81.,     80/81.,   5120/81.,   -640/81.,
       -1072/81.,   -352/81.,   -496/81.,   3968/81.,    224/81.,  -1792/81.,
       3536/81.,   -928/81.,  -1072/81.,   4040/81.,   3536/81.,  -1072/81.,
       80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
       -496/81.,   3968/81.,   -424/81.,    152/81.,   -640/81.,     80/81.,
       3536/81.,   -928/81.,   -424/81.,   -496/81.,   -352/81.,    224/81.,
       -928/81.,   1088/81.,   -352/81.,   4112/81.,   -928/81.,   -352/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       4112/81.,   -496/81.,   -352/81.,   -424/81.,     80/81.,    800/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       -640/81.,   5120/81.,   3968/81.,   4544/81.,    512/81.,    -64/81.,
       3968/81.,   -640/81.,  -1792/81.,   3968/81.,    -64/81.,   -640/81.,
       32768/81.,  -4096/81.,  -4096/81.,    512/81.,    512/81.,   5120/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.},
   // Row    91 : MG ordering  = [ 3, 1, 6, 7, 2, 5, 4 ]
   {  -1072/81.,   -352/81.,    152/81.,   -568/81.,   -424/81.,   -496/81.,
      4040/81.,   4112/81.,   -568/81.,   4544/81.,   -496/81.,   3968/81.,
      80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
      800/81.,     80/81.,     80/81.,   -640/81.,   -640/81.,   5120/81.,
      3536/81.,   -928/81.,   -424/81.,   -496/81.,   -352/81.,    224/81.,
      -928/81.,   1088/81.,   -352/81.,   4112/81.,   -928/81.,   -352/81.,
      -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
      4112/81.,   -496/81.,   -352/81.,   -424/81.,     80/81.,    800/81.,
      -1072/81.,   -352/81.,   -496/81.,   3968/81.,    224/81.,  -1792/81.,
      3536/81.,   -928/81.,  -1072/81.,   4040/81.,   3536/81.,  -1072/81.,
      80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
      -496/81.,   3968/81.,   -424/81.,    152/81.,   -640/81.,     80/81.,
      -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
      3968/81.,   -640/81.,  -1792/81.,   3968/81.,    -64/81.,   -640/81.,
      -640/81.,   5120/81.,   3968/81.,   4544/81.,    512/81.,    -64/81.,
      -4096/81.,  32768/81.,    512/81.,   5120/81.,  -4096/81.,    512/81.,
      80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
      -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
      80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
      512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.},
   // Row    92 : MG ordering  = [ 3, 1, 6, 7, 4, 2, 5 ]
   {  -1072/81.,   -352/81.,   -496/81.,   3968/81.,    224/81.,  -1792/81.,
      3536/81.,   -928/81.,  -1072/81.,   4040/81.,   3536/81.,  -1072/81.,
      80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
      -496/81.,   3968/81.,   -424/81.,    152/81.,   -640/81.,     80/81.,
      4040/81.,   4112/81.,   -568/81.,   4544/81.,   -496/81.,   3968/81.,
      -1072/81.,   -352/81.,    152/81.,   -568/81.,   -424/81.,   -496/81.,
      8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
      80/81.,   -640/81.,    800/81.,     80/81.,   5120/81.,   -640/81.,
      -928/81.,   1088/81.,   -352/81.,   4112/81.,   -928/81.,   -352/81.,
      3536/81.,   -928/81.,   -424/81.,   -496/81.,   -352/81.,    224/81.,
      224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
      -352/81.,   -424/81.,   4112/81.,   -496/81.,    800/81.,     80/81.,
      -640/81.,   5120/81.,   3968/81.,   4544/81.,    512/81.,    -64/81.,
      -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
      -1792/81.,   3968/81.,   3968/81.,   -640/81.,   -640/81.,    -64/81.,
      -4096/81.,    512/81.,  32768/81.,  -4096/81.,   5120/81.,    512/81.,
      80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
      8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
      224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
      512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.},
   // Row    93 : MG ordering  = [ 3, 1, 6, 7, 4, 5, 2 ]
   {   3536/81.,   -928/81.,   -424/81.,   -496/81.,   -352/81.,    224/81.,
       -928/81.,   1088/81.,   -352/81.,   4112/81.,   -928/81.,   -352/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       4112/81.,   -496/81.,   -352/81.,   -424/81.,     80/81.,    800/81.,
       -1072/81.,   -352/81.,    152/81.,   -568/81.,   -424/81.,   -496/81.,
       4040/81.,   4112/81.,   -568/81.,   4544/81.,   -496/81.,   3968/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       800/81.,     80/81.,     80/81.,   -640/81.,   -640/81.,   5120/81.,
       3536/81.,   -928/81.,  -1072/81.,   4040/81.,   3536/81.,  -1072/81.,
       -1072/81.,   -352/81.,   -496/81.,   3968/81.,    224/81.,  -1792/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       -424/81.,    152/81.,   -496/81.,   3968/81.,     80/81.,   -640/81.,
       3968/81.,   -640/81.,  -1792/81.,   3968/81.,    -64/81.,   -640/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       3968/81.,   4544/81.,   -640/81.,   5120/81.,    -64/81.,    512/81.,
       512/81.,   5120/81.,  -4096/81.,  32768/81.,    512/81.,  -4096/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.},
   // Row    94 : MG ordering  = [ 3, 1, 6, 7, 5, 2, 4 ]
   {   3536/81.,   -928/81.,  -1072/81.,   4040/81.,   3536/81.,  -1072/81.,
       -1072/81.,   -352/81.,   -496/81.,   3968/81.,    224/81.,  -1792/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       -424/81.,    152/81.,   -496/81.,   3968/81.,     80/81.,   -640/81.,
       -928/81.,   1088/81.,   -352/81.,   4112/81.,   -928/81.,   -352/81.,
       3536/81.,   -928/81.,   -424/81.,   -496/81.,   -352/81.,    224/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       -352/81.,   -424/81.,   4112/81.,   -496/81.,    800/81.,     80/81.,
       4040/81.,   4112/81.,   -568/81.,   4544/81.,   -496/81.,   3968/81.,
       -1072/81.,   -352/81.,    152/81.,   -568/81.,   -424/81.,   -496/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       80/81.,   -640/81.,    800/81.,     80/81.,   5120/81.,   -640/81.,
       3968/81.,   4544/81.,   -640/81.,   5120/81.,    -64/81.,    512/81.,
       -1792/81.,   3968/81.,   3968/81.,   -640/81.,   -640/81.,    -64/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       512/81.,  -4096/81.,   5120/81.,    512/81.,  32768/81.,  -4096/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.},
   // Row    95 : MG ordering  = [ 3, 1, 6, 7, 5, 4, 2 ]
   {   -928/81.,   1088/81.,   -352/81.,   4112/81.,   -928/81.,   -352/81.,
       3536/81.,   -928/81.,   -424/81.,   -496/81.,   -352/81.,    224/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       -352/81.,   -424/81.,   4112/81.,   -496/81.,    800/81.,     80/81.,
       3536/81.,   -928/81.,  -1072/81.,   4040/81.,   3536/81.,  -1072/81.,
       -1072/81.,   -352/81.,   -496/81.,   3968/81.,    224/81.,  -1792/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       -424/81.,    152/81.,   -496/81.,   3968/81.,     80/81.,   -640/81.,
       -1072/81.,   -352/81.,    152/81.,   -568/81.,   -424/81.,   -496/81.,
       4040/81.,   4112/81.,   -568/81.,   4544/81.,   -496/81.,   3968/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       800/81.,     80/81.,     80/81.,   -640/81.,   -640/81.,   5120/81.,
       -1792/81.,   3968/81.,   3968/81.,   -640/81.,   -640/81.,    -64/81.,
       3968/81.,   4544/81.,   -640/81.,   5120/81.,    -64/81.,    512/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       5120/81.,    512/81.,    512/81.,  -4096/81.,  -4096/81.,  32768/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.},
   // Row    96 : MG ordering  = [ 3, 1, 7, 2, 4, 5, 6 ]
   {    8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
        80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
        -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
        -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
        80/81.,   -640/81.,    800/81.,     80/81.,   5120/81.,   -640/81.,
        -496/81.,   3968/81.,   4112/81.,   4040/81.,   4544/81.,   -568/81.,
        -424/81.,   -496/81.,   -352/81.,  -1072/81.,   -568/81.,    152/81.,
        512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
        -496/81.,   3968/81.,   -424/81.,    152/81.,   -640/81.,     80/81.,
        224/81.,  -1792/81.,   -352/81.,  -1072/81.,   3968/81.,   -496/81.,
        3536/81.,  -1072/81.,   -928/81.,   3536/81.,   4040/81.,  -1072/81.,
        -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
        4112/81.,   -496/81.,   -352/81.,   -424/81.,     80/81.,    800/81.,
        -352/81.,    224/81.,   -928/81.,   3536/81.,   -496/81.,   -424/81.,
        -928/81.,   -352/81.,   1088/81.,   -928/81.,   4112/81.,   -352/81.,
        8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
        32768/81.,  -4096/81.,  -4096/81.,    512/81.,    512/81.,   5120/81.,
        -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
        512/81.,    -64/81.,   5120/81.,   -640/81.,   4544/81.,   3968/81.,
        -64/81.,   -640/81.,   -640/81.,   3968/81.,   3968/81.,  -1792/81.},
   // Row    97 : MG ordering  = [ 3, 1, 7, 2, 4, 6, 5 ]
   {     80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
         -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
         80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
         512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
         800/81.,     80/81.,     80/81.,   -640/81.,   -640/81.,   5120/81.,
         -424/81.,   -496/81.,   -352/81.,  -1072/81.,   -568/81.,    152/81.,
         -496/81.,   3968/81.,   4112/81.,   4040/81.,   4544/81.,   -568/81.,
         -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
         4112/81.,   -496/81.,   -352/81.,   -424/81.,     80/81.,    800/81.,
         -352/81.,    224/81.,   -928/81.,   3536/81.,   -496/81.,   -424/81.,
         -928/81.,   -352/81.,   1088/81.,   -928/81.,   4112/81.,   -352/81.,
         8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
         -496/81.,   3968/81.,   -424/81.,    152/81.,   -640/81.,     80/81.,
         224/81.,  -1792/81.,   -352/81.,  -1072/81.,   3968/81.,   -496/81.,
         3536/81.,  -1072/81.,   -928/81.,   3536/81.,   4040/81.,  -1072/81.,
         -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
         -4096/81.,  32768/81.,    512/81.,   5120/81.,  -4096/81.,    512/81.,
         512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
         -64/81.,   -640/81.,   -640/81.,   3968/81.,   3968/81.,  -1792/81.,
         512/81.,    -64/81.,   5120/81.,   -640/81.,   4544/81.,   3968/81.},
   // Row    98 : MG ordering  = [ 3, 1, 7, 2, 5, 4, 6 ]
   {     80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
         8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
         224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
         512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
         -496/81.,   3968/81.,   -424/81.,    152/81.,   -640/81.,     80/81.,
         224/81.,  -1792/81.,   -352/81.,  -1072/81.,   3968/81.,   -496/81.,
         3536/81.,  -1072/81.,   -928/81.,   3536/81.,   4040/81.,  -1072/81.,
         -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
         80/81.,   -640/81.,    800/81.,     80/81.,   5120/81.,   -640/81.,
         -496/81.,   3968/81.,   4112/81.,   4040/81.,   4544/81.,   -568/81.,
         -424/81.,   -496/81.,   -352/81.,  -1072/81.,   -568/81.,    152/81.,
         512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
         -352/81.,   -424/81.,   4112/81.,   -496/81.,    800/81.,     80/81.,
         -928/81.,   -352/81.,   1088/81.,   -928/81.,   4112/81.,   -352/81.,
         -352/81.,    224/81.,   -928/81.,   3536/81.,   -496/81.,   -424/81.,
         80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
         -4096/81.,    512/81.,  32768/81.,  -4096/81.,   5120/81.,    512/81.,
         512/81.,    -64/81.,   5120/81.,   -640/81.,   4544/81.,   3968/81.,
         -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
         -640/81.,    -64/81.,   3968/81.,  -1792/81.,   -640/81.,   3968/81.},
   // Row    99 : MG ordering  = [ 3, 1, 7, 2, 5, 6, 4 ]
   {   -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
       4112/81.,   -496/81.,   -352/81.,   -424/81.,     80/81.,    800/81.,
       -352/81.,    224/81.,   -928/81.,   3536/81.,   -496/81.,   -424/81.,
       -928/81.,   -352/81.,   1088/81.,   -928/81.,   4112/81.,   -352/81.,
       8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
       800/81.,     80/81.,     80/81.,   -640/81.,   -640/81.,   5120/81.,
       -424/81.,   -496/81.,   -352/81.,  -1072/81.,   -568/81.,    152/81.,
       -496/81.,   3968/81.,   4112/81.,   4040/81.,   4544/81.,   -568/81.,
       -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
       -424/81.,    152/81.,   -496/81.,   3968/81.,     80/81.,   -640/81.,
       3536/81.,  -1072/81.,   -928/81.,   3536/81.,   4040/81.,  -1072/81.,
       224/81.,  -1792/81.,   -352/81.,  -1072/81.,   3968/81.,   -496/81.,
       8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
       512/81.,   5120/81.,  -4096/81.,  32768/81.,    512/81.,  -4096/81.,
       -64/81.,   -640/81.,   -640/81.,   3968/81.,   3968/81.,  -1792/81.,
       512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
       -64/81.,    512/81.,   4544/81.,   3968/81.,   5120/81.,   -640/81.},
   // Row   100 : MG ordering  = [ 3, 1, 7, 2, 6, 4, 5 ]
   {   -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       -424/81.,    152/81.,   -496/81.,   3968/81.,     80/81.,   -640/81.,
       3536/81.,  -1072/81.,   -928/81.,   3536/81.,   4040/81.,  -1072/81.,
       224/81.,  -1792/81.,   -352/81.,  -1072/81.,   3968/81.,   -496/81.,
       8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
       -352/81.,   -424/81.,   4112/81.,   -496/81.,    800/81.,     80/81.,
       -928/81.,   -352/81.,   1088/81.,   -928/81.,   4112/81.,   -352/81.,
       -352/81.,    224/81.,   -928/81.,   3536/81.,   -496/81.,   -424/81.,
       80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
       80/81.,   -640/81.,    800/81.,     80/81.,   5120/81.,   -640/81.,
       -496/81.,   3968/81.,   4112/81.,   4040/81.,   4544/81.,   -568/81.,
       -424/81.,   -496/81.,   -352/81.,  -1072/81.,   -568/81.,    152/81.,
       512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
       512/81.,  -4096/81.,   5120/81.,    512/81.,  32768/81.,  -4096/81.,
       -64/81.,    512/81.,   4544/81.,   3968/81.,   5120/81.,   -640/81.,
       -640/81.,    -64/81.,   3968/81.,  -1792/81.,   -640/81.,   3968/81.,
       -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.},
   // Row   101 : MG ordering  = [ 3, 1, 7, 2, 6, 5, 4 ]
   {    224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
        -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
        80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
        -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
        -352/81.,   -424/81.,   4112/81.,   -496/81.,    800/81.,     80/81.,
        -928/81.,   -352/81.,   1088/81.,   -928/81.,   4112/81.,   -352/81.,
        -352/81.,    224/81.,   -928/81.,   3536/81.,   -496/81.,   -424/81.,
        80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
        -424/81.,    152/81.,   -496/81.,   3968/81.,     80/81.,   -640/81.,
        3536/81.,  -1072/81.,   -928/81.,   3536/81.,   4040/81.,  -1072/81.,
        224/81.,  -1792/81.,   -352/81.,  -1072/81.,   3968/81.,   -496/81.,
        8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
        800/81.,     80/81.,     80/81.,   -640/81.,   -640/81.,   5120/81.,
        -424/81.,   -496/81.,   -352/81.,  -1072/81.,   -568/81.,    152/81.,
        -496/81.,   3968/81.,   4112/81.,   4040/81.,   4544/81.,   -568/81.,
        -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
        5120/81.,    512/81.,    512/81.,  -4096/81.,  -4096/81.,  32768/81.,
        -640/81.,    -64/81.,   3968/81.,  -1792/81.,   -640/81.,   3968/81.,
        -64/81.,    512/81.,   4544/81.,   3968/81.,   5120/81.,   -640/81.,
        512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.},
   // Row   102 : MG ordering  = [ 3, 1, 7, 4, 2, 5, 6 ]
   {     80/81.,   -640/81.,    800/81.,     80/81.,   5120/81.,   -640/81.,
         -496/81.,   3968/81.,   4112/81.,   4040/81.,   4544/81.,   -568/81.,
         -424/81.,   -496/81.,   -352/81.,  -1072/81.,   -568/81.,    152/81.,
         512/81.,    -64/81.,    -64/81.,      8/81.,      8/81.,     80/81.,
         8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
         80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
         -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
         -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
         224/81.,  -1792/81.,   -352/81.,  -1072/81.,   3968/81.,   -496/81.,
         -496/81.,   3968/81.,   -424/81.,    152/81.,   -640/81.,     80/81.,
         -928/81.,   3536/81.,   3536/81.,  -1072/81.,  -1072/81.,   4040/81.,
         -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
         -352/81.,    224/81.,   -928/81.,   3536/81.,   -496/81.,   -424/81.,
         4112/81.,   -496/81.,   -352/81.,   -424/81.,     80/81.,    800/81.,
         1088/81.,   -928/81.,   -928/81.,   -352/81.,   -352/81.,   4112/81.,
         80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
         -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
         32768/81.,  -4096/81.,  -4096/81.,    512/81.,    512/81.,   5120/81.,
         5120/81.,   -640/81.,    512/81.,    -64/81.,   3968/81.,   4544/81.,
         -640/81.,   3968/81.,    -64/81.,   -640/81.,  -1792/81.,   3968/81.},
   // Row   103 : MG ordering  = [ 3, 1, 7, 4, 2, 6, 5 ]
   {    800/81.,     80/81.,     80/81.,   -640/81.,   -640/81.,   5120/81.,
        -424/81.,   -496/81.,   -352/81.,  -1072/81.,   -568/81.,    152/81.,
        -496/81.,   3968/81.,   4112/81.,   4040/81.,   4544/81.,   -568/81.,
        -64/81.,    512/81.,      8/81.,     80/81.,    -64/81.,      8/81.,
        80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
        -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
        80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
        512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
        -352/81.,    224/81.,   -928/81.,   3536/81.,   -496/81.,   -424/81.,
        4112/81.,   -496/81.,   -352/81.,   -424/81.,     80/81.,    800/81.,
        1088/81.,   -928/81.,   -928/81.,   -352/81.,   -352/81.,   4112/81.,
        80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
        224/81.,  -1792/81.,   -352/81.,  -1072/81.,   3968/81.,   -496/81.,
        -496/81.,   3968/81.,   -424/81.,    152/81.,   -640/81.,     80/81.,
        -928/81.,   3536/81.,   3536/81.,  -1072/81.,  -1072/81.,   4040/81.,
        -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
        512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
        -4096/81.,  32768/81.,    512/81.,   5120/81.,  -4096/81.,    512/81.,
        -640/81.,   3968/81.,    -64/81.,   -640/81.,  -1792/81.,   3968/81.,
        5120/81.,   -640/81.,    512/81.,    -64/81.,   3968/81.,   4544/81.},
   // Row   104 : MG ordering  = [ 3, 1, 7, 4, 5, 2, 6 ]
   {   -496/81.,   3968/81.,   -424/81.,    152/81.,   -640/81.,     80/81.,
       224/81.,  -1792/81.,   -352/81.,  -1072/81.,   3968/81.,   -496/81.,
       3536/81.,  -1072/81.,   -928/81.,   3536/81.,   4040/81.,  -1072/81.,
       -64/81.,      8/81.,   -640/81.,     80/81.,   -568/81.,   -496/81.,
       80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
       -496/81.,   3968/81.,   4112/81.,   4040/81.,   4544/81.,   -568/81.,
       80/81.,   -640/81.,    800/81.,     80/81.,   5120/81.,   -640/81.,
       -352/81.,  -1072/81.,   -424/81.,   -496/81.,    152/81.,   -568/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       -928/81.,   -352/81.,   1088/81.,   -928/81.,   4112/81.,   -352/81.,
       -352/81.,   -424/81.,   4112/81.,   -496/81.,    800/81.,     80/81.,
       -928/81.,   3536/81.,   -352/81.,    224/81.,   -424/81.,   -496/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       512/81.,    -64/81.,   5120/81.,   -640/81.,   4544/81.,   3968/81.,
       -4096/81.,    512/81.,  32768/81.,  -4096/81.,   5120/81.,    512/81.,
       512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
       3968/81.,  -1792/81.,   -640/81.,    -64/81.,   3968/81.,   -640/81.},
   // Row   105 : MG ordering  = [ 3, 1, 7, 4, 5, 6, 2 ]
   {   4112/81.,   -496/81.,   -352/81.,   -424/81.,     80/81.,    800/81.,
       -352/81.,    224/81.,   -928/81.,   3536/81.,   -496/81.,   -424/81.,
       -928/81.,   -352/81.,   1088/81.,   -928/81.,   4112/81.,   -352/81.,
       8/81.,     80/81.,     80/81.,   -496/81.,   -496/81.,    224/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
       -424/81.,   -496/81.,   -352/81.,  -1072/81.,   -568/81.,    152/81.,
       800/81.,     80/81.,     80/81.,   -640/81.,   -640/81.,   5120/81.,
       4112/81.,   4040/81.,   -496/81.,   3968/81.,   -568/81.,   4544/81.,
       8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
       3536/81.,  -1072/81.,   -928/81.,   3536/81.,   4040/81.,  -1072/81.,
       -424/81.,    152/81.,   -496/81.,   3968/81.,     80/81.,   -640/81.,
       -352/81.,  -1072/81.,    224/81.,  -1792/81.,   -496/81.,   3968/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       -64/81.,   -640/81.,   -640/81.,   3968/81.,   3968/81.,  -1792/81.,
       512/81.,   5120/81.,  -4096/81.,  32768/81.,    512/81.,  -4096/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
       4544/81.,   3968/81.,    -64/81.,    512/81.,   -640/81.,   5120/81.},
   // Row   106 : MG ordering  = [ 3, 1, 7, 4, 6, 2, 5 ]
   {   -424/81.,    152/81.,   -496/81.,   3968/81.,     80/81.,   -640/81.,
       3536/81.,  -1072/81.,   -928/81.,   3536/81.,   4040/81.,  -1072/81.,
       224/81.,  -1792/81.,   -352/81.,  -1072/81.,   3968/81.,   -496/81.,
       8/81.,    -64/81.,   -568/81.,   -496/81.,   -640/81.,     80/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       -928/81.,   -352/81.,   1088/81.,   -928/81.,   4112/81.,   -352/81.,
       -352/81.,   -424/81.,   4112/81.,   -496/81.,    800/81.,     80/81.,
       -928/81.,   3536/81.,   -352/81.,    224/81.,   -424/81.,   -496/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       -496/81.,   3968/81.,   4112/81.,   4040/81.,   4544/81.,   -568/81.,
       80/81.,   -640/81.,    800/81.,     80/81.,   5120/81.,   -640/81.,
       -352/81.,  -1072/81.,   -424/81.,   -496/81.,    152/81.,   -568/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       -64/81.,    512/81.,   4544/81.,   3968/81.,   5120/81.,   -640/81.,
       512/81.,  -4096/81.,   5120/81.,    512/81.,  32768/81.,  -4096/81.,
       3968/81.,  -1792/81.,   -640/81.,    -64/81.,   3968/81.,   -640/81.,
       512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.},
   // Row   107 : MG ordering  = [ 3, 1, 7, 4, 6, 5, 2 ]
   {   -352/81.,   -424/81.,   4112/81.,   -496/81.,    800/81.,     80/81.,
       -928/81.,   -352/81.,   1088/81.,   -928/81.,   4112/81.,   -352/81.,
       -352/81.,    224/81.,   -928/81.,   3536/81.,   -496/81.,   -424/81.,
       80/81.,      8/81.,   -496/81.,    224/81.,     80/81.,   -496/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       3536/81.,  -1072/81.,   -928/81.,   3536/81.,   4040/81.,  -1072/81.,
       -424/81.,    152/81.,   -496/81.,   3968/81.,     80/81.,   -640/81.,
       -352/81.,  -1072/81.,    224/81.,  -1792/81.,   -496/81.,   3968/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       -424/81.,   -496/81.,   -352/81.,  -1072/81.,   -568/81.,    152/81.,
       800/81.,     80/81.,     80/81.,   -640/81.,   -640/81.,   5120/81.,
       4112/81.,   4040/81.,   -496/81.,   3968/81.,   -568/81.,   4544/81.,
       8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
       -640/81.,    -64/81.,   3968/81.,  -1792/81.,   -640/81.,   3968/81.,
       5120/81.,    512/81.,    512/81.,  -4096/81.,  -4096/81.,  32768/81.,
       4544/81.,   3968/81.,    -64/81.,    512/81.,   -640/81.,   5120/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.},
   // Row   108 : MG ordering  = [ 3, 1, 7, 5, 2, 4, 6 ]
   {   -496/81.,   3968/81.,   4112/81.,   4040/81.,   4544/81.,   -568/81.,
       80/81.,   -640/81.,    800/81.,     80/81.,   5120/81.,   -640/81.,
       -352/81.,  -1072/81.,   -424/81.,   -496/81.,    152/81.,   -568/81.,
       -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
       224/81.,  -1792/81.,   -352/81.,  -1072/81.,   3968/81.,   -496/81.,
       -496/81.,   3968/81.,   -424/81.,    152/81.,   -640/81.,     80/81.,
       -928/81.,   3536/81.,   3536/81.,  -1072/81.,  -1072/81.,   4040/81.,
       -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
       -928/81.,   3536/81.,   -352/81.,    224/81.,   -424/81.,   -496/81.,
       1088/81.,   -928/81.,   -928/81.,   -352/81.,   -352/81.,   4112/81.,
       4112/81.,   -496/81.,   -352/81.,   -424/81.,     80/81.,    800/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
       5120/81.,   -640/81.,    512/81.,    -64/81.,   3968/81.,   4544/81.,
       32768/81.,  -4096/81.,  -4096/81.,    512/81.,    512/81.,   5120/81.,
       3968/81.,   -640/81.,  -1792/81.,   3968/81.,    -64/81.,   -640/81.},
   // Row   109 : MG ordering  = [ 3, 1, 7, 5, 2, 6, 4 ]
   {   -424/81.,   -496/81.,   -352/81.,  -1072/81.,   -568/81.,    152/81.,
       800/81.,     80/81.,     80/81.,   -640/81.,   -640/81.,   5120/81.,
       4112/81.,   4040/81.,   -496/81.,   3968/81.,   -568/81.,   4544/81.,
       8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
       -352/81.,    224/81.,   -928/81.,   3536/81.,   -496/81.,   -424/81.,
       4112/81.,   -496/81.,   -352/81.,   -424/81.,     80/81.,    800/81.,
       1088/81.,   -928/81.,   -928/81.,   -352/81.,   -352/81.,   4112/81.,
       80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
       512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
       -352/81.,  -1072/81.,    224/81.,  -1792/81.,   -496/81.,   3968/81.,
       -928/81.,   3536/81.,   3536/81.,  -1072/81.,  -1072/81.,   4040/81.,
       -496/81.,   3968/81.,   -424/81.,    152/81.,   -640/81.,     80/81.,
       80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
       -640/81.,   3968/81.,    -64/81.,   -640/81.,  -1792/81.,   3968/81.,
       -4096/81.,  32768/81.,    512/81.,   5120/81.,  -4096/81.,    512/81.,
       -640/81.,   5120/81.,   3968/81.,   4544/81.,    512/81.,    -64/81.},
   // Row   110 : MG ordering  = [ 3, 1, 7, 5, 4, 2, 6 ]
   {    224/81.,  -1792/81.,   -352/81.,  -1072/81.,   3968/81.,   -496/81.,
        -496/81.,   3968/81.,   -424/81.,    152/81.,   -640/81.,     80/81.,
        -928/81.,   3536/81.,   3536/81.,  -1072/81.,  -1072/81.,   4040/81.,
        -640/81.,     80/81.,    -64/81.,      8/81.,   -496/81.,   -568/81.,
        -496/81.,   3968/81.,   4112/81.,   4040/81.,   4544/81.,   -568/81.,
        80/81.,   -640/81.,    800/81.,     80/81.,   5120/81.,   -640/81.,
        -352/81.,  -1072/81.,   -424/81.,   -496/81.,    152/81.,   -568/81.,
        -64/81.,      8/81.,    512/81.,    -64/81.,     80/81.,      8/81.,
        80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
        8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
        224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
        512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
        1088/81.,   -928/81.,   -928/81.,   -352/81.,   -352/81.,   4112/81.,
        -928/81.,   3536/81.,   -352/81.,    224/81.,   -424/81.,   -496/81.,
        -352/81.,   -424/81.,   4112/81.,   -496/81.,    800/81.,     80/81.,
        224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
        5120/81.,   -640/81.,    512/81.,    -64/81.,   3968/81.,   4544/81.,
        512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
        -4096/81.,    512/81.,  32768/81.,  -4096/81.,   5120/81.,    512/81.,
        -1792/81.,   3968/81.,   3968/81.,   -640/81.,   -640/81.,    -64/81.},
   // Row   111 : MG ordering  = [ 3, 1, 7, 5, 4, 6, 2 ]
   {   -352/81.,    224/81.,   -928/81.,   3536/81.,   -496/81.,   -424/81.,
       4112/81.,   -496/81.,   -352/81.,   -424/81.,     80/81.,    800/81.,
       1088/81.,   -928/81.,   -928/81.,   -352/81.,   -352/81.,   4112/81.,
       80/81.,   -496/81.,      8/81.,     80/81.,    224/81.,   -496/81.,
       -424/81.,   -496/81.,   -352/81.,  -1072/81.,   -568/81.,    152/81.,
       800/81.,     80/81.,     80/81.,   -640/81.,   -640/81.,   5120/81.,
       4112/81.,   4040/81.,   -496/81.,   3968/81.,   -568/81.,   4544/81.,
       8/81.,     80/81.,    -64/81.,    512/81.,      8/81.,    -64/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
       -928/81.,   3536/81.,   3536/81.,  -1072/81.,  -1072/81.,   4040/81.,
       -352/81.,  -1072/81.,    224/81.,  -1792/81.,   -496/81.,   3968/81.,
       -424/81.,    152/81.,   -496/81.,   3968/81.,     80/81.,   -640/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       -640/81.,   3968/81.,    -64/81.,   -640/81.,  -1792/81.,   3968/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
       512/81.,   5120/81.,  -4096/81.,  32768/81.,    512/81.,  -4096/81.,
       3968/81.,   4544/81.,   -640/81.,   5120/81.,    -64/81.,    512/81.},
   // Row   112 : MG ordering  = [ 3, 1, 7, 5, 6, 2, 4 ]
   {   3536/81.,  -1072/81.,   -928/81.,   3536/81.,   4040/81.,  -1072/81.,
       -424/81.,    152/81.,   -496/81.,   3968/81.,     80/81.,   -640/81.,
       -352/81.,  -1072/81.,    224/81.,  -1792/81.,   -496/81.,   3968/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       -928/81.,   -352/81.,   1088/81.,   -928/81.,   4112/81.,   -352/81.,
       -352/81.,   -424/81.,   4112/81.,   -496/81.,    800/81.,     80/81.,
       -928/81.,   3536/81.,   -352/81.,    224/81.,   -424/81.,   -496/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       4112/81.,   4040/81.,   -496/81.,   3968/81.,   -568/81.,   4544/81.,
       -352/81.,  -1072/81.,   -424/81.,   -496/81.,    152/81.,   -568/81.,
       80/81.,   -640/81.,    800/81.,     80/81.,   5120/81.,   -640/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       4544/81.,   3968/81.,    -64/81.,    512/81.,   -640/81.,   5120/81.,
       3968/81.,  -1792/81.,   -640/81.,    -64/81.,   3968/81.,   -640/81.,
       512/81.,  -4096/81.,   5120/81.,    512/81.,  32768/81.,  -4096/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.},
   // Row   113 : MG ordering  = [ 3, 1, 7, 5, 6, 4, 2 ]
   {   -928/81.,   -352/81.,   1088/81.,   -928/81.,   4112/81.,   -352/81.,
       -352/81.,   -424/81.,   4112/81.,   -496/81.,    800/81.,     80/81.,
       -928/81.,   3536/81.,   -352/81.,    224/81.,   -424/81.,   -496/81.,
       -496/81.,    224/81.,     80/81.,      8/81.,   -496/81.,     80/81.,
       3536/81.,  -1072/81.,   -928/81.,   3536/81.,   4040/81.,  -1072/81.,
       -424/81.,    152/81.,   -496/81.,   3968/81.,     80/81.,   -640/81.,
       -352/81.,  -1072/81.,    224/81.,  -1792/81.,   -496/81.,   3968/81.,
       -568/81.,   -496/81.,      8/81.,    -64/81.,     80/81.,   -640/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       -352/81.,  -1072/81.,   -424/81.,   -496/81.,    152/81.,   -568/81.,
       4112/81.,   4040/81.,   -496/81.,   3968/81.,   -568/81.,   4544/81.,
       800/81.,     80/81.,     80/81.,   -640/81.,   -640/81.,   5120/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       3968/81.,  -1792/81.,   -640/81.,    -64/81.,   3968/81.,   -640/81.,
       4544/81.,   3968/81.,    -64/81.,    512/81.,   -640/81.,   5120/81.,
       5120/81.,    512/81.,    512/81.,  -4096/81.,  -4096/81.,  32768/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.},
   // Row   114 : MG ordering  = [ 3, 1, 7, 6, 2, 4, 5 ]
   {   4112/81.,   4040/81.,   -496/81.,   3968/81.,   -568/81.,   4544/81.,
       -352/81.,  -1072/81.,   -424/81.,   -496/81.,    152/81.,   -568/81.,
       80/81.,   -640/81.,    800/81.,     80/81.,   5120/81.,   -640/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       -352/81.,  -1072/81.,    224/81.,  -1792/81.,   -496/81.,   3968/81.,
       -928/81.,   3536/81.,   3536/81.,  -1072/81.,  -1072/81.,   4040/81.,
       -496/81.,   3968/81.,   -424/81.,    152/81.,   -640/81.,     80/81.,
       80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
       -928/81.,   3536/81.,   -352/81.,    224/81.,   -424/81.,   -496/81.,
       1088/81.,   -928/81.,   -928/81.,   -352/81.,   -352/81.,   4112/81.,
       4112/81.,   -496/81.,   -352/81.,   -424/81.,     80/81.,    800/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       -4096/81.,    512/81.,    512/81.,    -64/81.,    -64/81.,   -640/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       -640/81.,   5120/81.,   3968/81.,   4544/81.,    512/81.,    -64/81.,
       3968/81.,   -640/81.,  -1792/81.,   3968/81.,    -64/81.,   -640/81.,
       32768/81.,  -4096/81.,  -4096/81.,    512/81.,    512/81.,   5120/81.},
   // Row   115 : MG ordering  = [ 3, 1, 7, 6, 2, 5, 4 ]
   {   -352/81.,  -1072/81.,   -424/81.,   -496/81.,    152/81.,   -568/81.,
       4112/81.,   4040/81.,   -496/81.,   3968/81.,   -568/81.,   4544/81.,
       800/81.,     80/81.,     80/81.,   -640/81.,   -640/81.,   5120/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       -928/81.,   3536/81.,   -352/81.,    224/81.,   -424/81.,   -496/81.,
       1088/81.,   -928/81.,   -928/81.,   -352/81.,   -352/81.,   4112/81.,
       4112/81.,   -496/81.,   -352/81.,   -424/81.,     80/81.,    800/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       -352/81.,  -1072/81.,    224/81.,  -1792/81.,   -496/81.,   3968/81.,
       -928/81.,   3536/81.,   3536/81.,  -1072/81.,  -1072/81.,   4040/81.,
       -496/81.,   3968/81.,   -424/81.,    152/81.,   -640/81.,     80/81.,
       80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
       512/81.,  -4096/81.,    -64/81.,   -640/81.,    512/81.,    -64/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       3968/81.,   -640/81.,  -1792/81.,   3968/81.,    -64/81.,   -640/81.,
       -640/81.,   5120/81.,   3968/81.,   4544/81.,    512/81.,    -64/81.,
       -4096/81.,  32768/81.,    512/81.,   5120/81.,  -4096/81.,    512/81.},
   // Row   116 : MG ordering  = [ 3, 1, 7, 6, 4, 2, 5 ]
   {   -352/81.,  -1072/81.,    224/81.,  -1792/81.,   -496/81.,   3968/81.,
       -928/81.,   3536/81.,   3536/81.,  -1072/81.,  -1072/81.,   4040/81.,
       -496/81.,   3968/81.,   -424/81.,    152/81.,   -640/81.,     80/81.,
       80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
       4112/81.,   4040/81.,   -496/81.,   3968/81.,   -568/81.,   4544/81.,
       -352/81.,  -1072/81.,   -424/81.,   -496/81.,    152/81.,   -568/81.,
       80/81.,   -640/81.,    800/81.,     80/81.,   5120/81.,   -640/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       1088/81.,   -928/81.,   -928/81.,   -352/81.,   -352/81.,   4112/81.,
       -928/81.,   3536/81.,   -352/81.,    224/81.,   -424/81.,   -496/81.,
       -352/81.,   -424/81.,   4112/81.,   -496/81.,    800/81.,     80/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       80/81.,   -640/81.,   -496/81.,   -568/81.,    -64/81.,      8/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       512/81.,    -64/81.,  -4096/81.,    512/81.,   -640/81.,    -64/81.,
       -640/81.,   5120/81.,   3968/81.,   4544/81.,    512/81.,    -64/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       -1792/81.,   3968/81.,   3968/81.,   -640/81.,   -640/81.,    -64/81.,
       -4096/81.,    512/81.,  32768/81.,  -4096/81.,   5120/81.,    512/81.},
   // Row   117 : MG ordering  = [ 3, 1, 7, 6, 4, 5, 2 ]
   {   -928/81.,   3536/81.,   -352/81.,    224/81.,   -424/81.,   -496/81.,
       1088/81.,   -928/81.,   -928/81.,   -352/81.,   -352/81.,   4112/81.,
       4112/81.,   -496/81.,   -352/81.,   -424/81.,     80/81.,    800/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       -352/81.,  -1072/81.,   -424/81.,   -496/81.,    152/81.,   -568/81.,
       4112/81.,   4040/81.,   -496/81.,   3968/81.,   -568/81.,   4544/81.,
       800/81.,     80/81.,     80/81.,   -640/81.,   -640/81.,   5120/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       -928/81.,   3536/81.,   3536/81.,  -1072/81.,  -1072/81.,   4040/81.,
       -352/81.,  -1072/81.,    224/81.,  -1792/81.,   -496/81.,   3968/81.,
       -424/81.,    152/81.,   -496/81.,   3968/81.,     80/81.,   -640/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       -496/81.,     80/81.,    224/81.,   -496/81.,      8/81.,     80/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       -64/81.,   -640/81.,    512/81.,  -4096/81.,    -64/81.,    512/81.,
       3968/81.,   -640/81.,  -1792/81.,   3968/81.,    -64/81.,   -640/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       3968/81.,   4544/81.,   -640/81.,   5120/81.,    -64/81.,    512/81.,
       512/81.,   5120/81.,  -4096/81.,  32768/81.,    512/81.,  -4096/81.},
   // Row   118 : MG ordering  = [ 3, 1, 7, 6, 5, 2, 4 ]
   {   -928/81.,   3536/81.,   3536/81.,  -1072/81.,  -1072/81.,   4040/81.,
       -352/81.,  -1072/81.,    224/81.,  -1792/81.,   -496/81.,   3968/81.,
       -424/81.,    152/81.,   -496/81.,   3968/81.,     80/81.,   -640/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       1088/81.,   -928/81.,   -928/81.,   -352/81.,   -352/81.,   4112/81.,
       -928/81.,   3536/81.,   -352/81.,    224/81.,   -424/81.,   -496/81.,
       -352/81.,   -424/81.,   4112/81.,   -496/81.,    800/81.,     80/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       4112/81.,   4040/81.,   -496/81.,   3968/81.,   -568/81.,   4544/81.,
       -352/81.,  -1072/81.,   -424/81.,   -496/81.,    152/81.,   -568/81.,
       80/81.,   -640/81.,    800/81.,     80/81.,   5120/81.,   -640/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       8/81.,    -64/81.,     80/81.,      8/81.,    512/81.,    -64/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       3968/81.,   4544/81.,   -640/81.,   5120/81.,    -64/81.,    512/81.,
       -1792/81.,   3968/81.,   3968/81.,   -640/81.,   -640/81.,    -64/81.,
       -64/81.,    512/81.,   -640/81.,    -64/81.,  -4096/81.,    512/81.,
       512/81.,  -4096/81.,   5120/81.,    512/81.,  32768/81.,  -4096/81.},
   // Row   119 : MG ordering  = [ 3, 1, 7, 6, 5, 4, 2 ]
   {   1088/81.,   -928/81.,   -928/81.,   -352/81.,   -352/81.,   4112/81.,
       -928/81.,   3536/81.,   -352/81.,    224/81.,   -424/81.,   -496/81.,
       -352/81.,   -424/81.,   4112/81.,   -496/81.,    800/81.,     80/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       -928/81.,   3536/81.,   3536/81.,  -1072/81.,  -1072/81.,   4040/81.,
       -352/81.,  -1072/81.,    224/81.,  -1792/81.,   -496/81.,   3968/81.,
       -424/81.,    152/81.,   -496/81.,   3968/81.,     80/81.,   -640/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       -352/81.,  -1072/81.,   -424/81.,   -496/81.,    152/81.,   -568/81.,
       4112/81.,   4040/81.,   -496/81.,   3968/81.,   -568/81.,   4544/81.,
       800/81.,     80/81.,     80/81.,   -640/81.,   -640/81.,   5120/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       224/81.,   -496/81.,   -496/81.,     80/81.,     80/81.,      8/81.,
       -496/81.,   -568/81.,     80/81.,   -640/81.,      8/81.,    -64/81.,
       80/81.,      8/81.,      8/81.,    -64/81.,    -64/81.,    512/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       -1792/81.,   3968/81.,   3968/81.,   -640/81.,   -640/81.,    -64/81.,
       3968/81.,   4544/81.,   -640/81.,   5120/81.,    -64/81.,    512/81.,
       -640/81.,    -64/81.,    -64/81.,    512/81.,    512/81.,  -4096/81.,
       5120/81.,    512/81.,    512/81.,  -4096/81.,  -4096/81.,  32768/81.}
  };

//-----------------------------------------------------------------------//

// A base class which calculates spinor inner products <ij> and [ij]
class SpinorProducts {

 public:
  // Calculate sqrt(k^+) = sqrt(k[0] + k[3])
  Pythia8::complex sqrt_k_plus (Vec4 k);

  // Calculate sqrt(k^-) = sqrt(k[0] - k[3])
  Pythia8::complex sqrt_k_minus (Vec4 k);

  // Calculate e^{i*phi} = \frac{k[1] + k[2]} {sqrt(k^+ * k^i)}
  void phasePos (Vec4 k_i, Vec4 k_j);

  // Calculate e^{-i*phi} = \frac{k[1] - k[2]} {sqrt(k^+ * k^i)}
  void phaseNeg (Vec4 k_i, Vec4 k_j);

  // Calculate <ij>
  Pythia8::complex spinorProdPos (Vec4 k_i, Vec4 k_j);

  // Calculate [ij]
  Pythia8::complex spinorProdNeg (Vec4 k_i, Vec4 k_j);

 protected:
  Pythia8::complex phaseDenom_i;
  Pythia8::complex phaseDenom_j;
  Pythia8::complex phase_i;
  Pythia8::complex phase_j;
};

//-----------------------------------------------------------------------//


class MHVamplitudes : SpinorProducts {

 public:

  // Constructor
  MHVamplitudes () {
    Error      = true;
    isInitPtr  = false;
    isInit     = false;
    pos1Sav    = -1;
    pos2Sav    = -1;
    colorDepth = -1;
  };

  // Initialise pointers
  void initPtr(Couplings* couplingsPtrIn, ParticleData* particleDataPtrIn,
    Settings* settingsPtrIn) {
    couplingsPtr     = couplingsPtrIn;
    particleDataPtr  = particleDataPtrIn;
    settingsPtr      = settingsPtrIn;
    isInitPtr        = true;
  }

  // Initialise
  bool init();

  // Initialise spinor bag
  bool calcSpinorProds(vector<Vec4> k, vector<int> h, vector<int> ID, bool isOrdered = false);

  // Calculate colour amp squared
  double ME2 ();

  // allGlue: Which amplitude from ampBag do I need
  int getAmpNum(int nIn1, int nIn2, int nIn3, int nIn4, int nIn5);

  // Calculate factorials
  long factorial (int nIn) {
    return (nIn == 1) ? 1 : factorial(nIn - 1) * nIn;
  }

  // Get type of helicity amplitude
  int helAmpType(vector<int> hin);

  //Proposed function. Leaving it here in case it is to be used later
  /*  // Get type of amplitude from IDs
      int getAmpID(vector<int> idsIn);*/

  // Get helicity-dependent part of MHV amplitudes (for polarising a process)
  vector<double> getMHVhelAmps(vector<Vec4> k, vector<int> ID, int ampType);

  // Print spBag
  void printSPbag() {
    for (int i = 0; i < nPartons; i++) {
      for (int j = 0; j < nPartons; j++) {
        cout << spBag[i][j] << ", ";
      }
      cout << endl;
    }
  }

  // Reset all amplitudes, spinor products to 0
  void resetBags() {
    for (int i = 0; i < nPartonsMax; i++)
      for (int j = 0; j < nPartonsMax; j++) spBag[i][j] = 0;
    for (int i = 0; i < nAmpsMax; i++) ampBag[i] = 0;
  }


  // Order particles if not already ordered by Vincia::MEcorrections
  vector<int> orderParticles(vector<int> h, vector<int> ID);

  // Set the positions of the incoming guys
  void setIIpos(int pos1In, int pos2In) {
    pos1Sav = min(pos1In,pos2In);
    pos2Sav = max(pos1In,pos2In);
  }

  // Set color depth
  void setColorDepth(int colorDepthIn) {colorDepth = colorDepthIn;}

  // Get color depth
  int getColorDepth() {return colorDepth;}

  // Set verbose level
  void setVerbose(int verboseIn) {verbose = verboseIn;}

 private:

  // Is initialised?
  bool isInitPtr;
  bool isInit;

  // Pointers to PYTHIA8 objects
  Couplings*    couplingsPtr;
  ParticleData* particleDataPtr;
  Settings*     settingsPtr;

  // Colour Mode (0: leading colour, 1: Vincia colour, 2: full colour)
  int colorDepth;

  // Verbose level
  int verbose;

  // Positions of incoming particles
  int pos1Sav;
  int pos2Sav;

  // Something's gone wrong?
  bool Error;
  bool fermHelConserved;
  bool nonChiralW;

  // CKM matrix elements and W coupling
  double Vud, Vus, Vub;
  double Vcd, Vcs, Vcb;
  double gWfl, gWflSq;
  double gWfr, gWfrSq;

  // Z and photon coupling
  double ee2, ee, efd;
  double efl, efu, ey;
  double ez, alpha, cw;
  double sin2w, sw;
  double gZul, gZur, gZdl;
  double gZdr, gZll, gZlr;
  double gAl, gAu, gAd;

  // Masses, Widths and propagators
  double  mW, mWSq, widthW;
  double  mZ, mZSq, widthZ;
  Pythia8::complex propW;
  Pythia8::complex propZ;

  // Type of process
  bool allGlue, twoQuark, fourQuark;
  bool isQCD, isWeak;
  bool mhvMinus, mhvPlus;
  bool sameFlav;

  // Place setters
  int l1, lBar1;
  int diffGlue1, diffGlue2;
  int q1, q2, qBar1, qBar2;

  // How many gluons, partons?
  int nGlue;
  int nPartons;

  // Store spinor products, amplitudes in array
  Pythia8::complex ampBag[nAmpsMax];
  Pythia8::complex spBag[nPartonsMax][nPartonsMax];

  // Values that are best calculated in calcSpinorProds, but used in ME2
  Pythia8::complex num;
  Pythia8::complex prefact;
  Pythia8::complex prefact2;
  double  aveColIn;

};

} // End namespace Vincia

#endif
